/*
 *  Copyright (C) 2007 Torbjörn Svangård  <trptorbjorn@gmail.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author(s) : Torbjörn Svangård  <trptorbjorn@gmail.com>
 *              Daniel    Bessler    <danielb@kiba-dock.org>
*/

#include "../config.h"
#include <stdlib.h>
#include <string.h>
#include <libintl.h>
#include <gtk/gtk.h>
#include <locale.h>

#include <libxml/parser.h>

#include "kiba-settings.h"
#include "settings-dbus.h"
#include "kiba-prefs.h"
#include "kiba-prefs-schema.h"
#include "widgets.h"

static GtkTooltips  *tooltips = NULL;

static gboolean type_is_valid (const char *type) {
	const char *kiba_pref_types[] = {
		"char", "int", "double", "bool", "list"
	};
	
	int i;
	
	for (i = 0; i < G_N_ELEMENTS (kiba_pref_types); i++) {
		if (strcmp (kiba_pref_types[i], type) == 0)
			return TRUE;
	}
	
	fprintf (stderr, "Invalid Type '%s'\n"
			 "Please use one of the following types: ",
			 type);
	for (i = 0; i < G_N_ELEMENTS (kiba_pref_types); i++) {
		fprintf (stderr, "%s", kiba_pref_types[i]);
		if (i + 1 < G_N_ELEMENTS (kiba_pref_types))
			fprintf (stderr, ", ");
		else
			fprintf (stderr, "\n");
	}
	
	return FALSE;
}

GtkTooltips* kiba_settings_get_tooltips (void) {
	if (!tooltips)
		tooltips = gtk_tooltips_new();
	
	return tooltips;
}

int main (int argc, char *argv[]) {
	KibaSchema *schema_data;
	gchar *config_path = NULL;
	gchar *kiba_user_dir;
	gint i;
	
	gtk_set_locale();
	gtk_disable_setlocale();
	
	setlocale (LC_NUMERIC,  "C");  /* use decimal point instead of comma */
	setlocale (LC_CTYPE,    "");
	setlocale (LC_MESSAGES, "");
	
#ifdef ENABLE_NLS
	bindtextdomain (PACKAGE, DATADIR"/locale");
	textdomain (PACKAGE);
	bind_textdomain_codeset (PACKAGE, "UTF-8");	
#endif
	
	gtk_init(&argc, &argv);
	
	i = 0;
	while (i < argc) {
		if (g_strrstr (argv[i], "--config-file") != NULL) {
			config_path = g_strdup (argv[i+1]);
			if (!g_file_test (config_path, G_FILE_TEST_EXISTS)) {
				g_free (config_path);
				config_path = NULL;
			}
			break;
		}
		i++;
	}
	
	/* create kiba home directory */
	kiba_user_dir = g_strdup_printf ("%s/.kiba-dock", g_get_home_dir ());
	if (!g_file_test (kiba_user_dir, G_FILE_TEST_IS_DIR)) {
		if (g_mkdir (kiba_user_dir, 7664) != 0)
			exit (0);
	}
	
	/* initiall preferences */
	if (!config_path) {
		config_path = g_strdup_printf ("%s/config.xml", kiba_user_dir);
	}
	kiba_prefs_init (config_path);
	g_free (config_path);
	g_free (kiba_user_dir);
	
	i = 0;
	while (i < argc) {
		if (strcmp (argv[i], "--help") == 0 ||
		    strcmp (argv[i], "-h") == 0) {
			fprintf (stderr,
				 "Usage: kiba-settings [--get-'$type' '$name']"
						  "[--set-'$type' '$name' '$value']"
						  "[--config-file config_file]"
						  "...\n");
			kiba_prefs_fini();
			return 0;
		}
		/* get Value from settings file */
		else if (g_strrstr (argv[i], "--get-")) {
			char    *value;
			
			if (!type_is_valid (argv[i]+6)) {
				kiba_prefs_fini();
				exit (0);
			}
			
			if (i+1 >= argc) {
				fprintf (stderr, "No Setting Name defined\n");
				kiba_prefs_fini();
				exit (0);
			}
			value = kiba_prefs_get_option_from_file (argv[i+1], argv[i]+6);
			if (!value)
				fprintf (stderr, "'%s' not found\n", argv[i+1]);
			else
				fprintf (stderr, "%s\n", value);
			
			kiba_prefs_fini();
			exit (0);
		}
		/* write Value to settings */
		else if (g_strrstr (argv[i], "--set-")) {
			if (!type_is_valid (argv[i]+6)) {
				kiba_prefs_fini();
				exit (0);
			}
			
			if (i+1 >= argc) {
				fprintf (stderr, "No Setting Name defined\n");
				kiba_prefs_fini();
				exit (0);
			}
			
			if (i+2 >= argc) {
				fprintf (stderr, "No Setting Value defined\n");
				kiba_prefs_fini();
				exit (0);
			}
			
			if (!kiba_prefs_write_to_config_file (argv[i+1],
					argv[i]+6,
					argv[i+2]))
				fprintf (stderr, "Failed to set '%s' to '%s'\n",
						 argv[i+1], argv[i+2]);
			
			kiba_prefs_fini();
			exit (0);
		}
		i++;
	}
	
	/* initiall libxml */
	LIBXML_TEST_VERSION
	
	/* Load kiba schemas, so that we can acces the data for creating the widgets */
	schema_data = kiba_prefs_schema_get();
	if (g_list_length(schema_data->sections) == 0) {
		fprintf (stderr, "No valid .schema file(s) found\n");
		return EXIT_FAILURE;
	}
	kiba_prefs_schema_set_options_from_config(schema_data);
	
	kiba_settings_dbus_init();
	
	/* create the widgets
	   here is a small schema of the Layout:
		VBox {  //<< Main Container
			Frame { //<< Top Container
				HBox {
					Frame { HBox { VButtonBox }},
					VBox {
						Label, //<< Active Section/Description
						Notebook { HBox { SettingWidgets }}}}},
			HBox { Button }} //<< Quit Button
	*/
	kiba_settings_create_window (schema_data);
	
	gtk_main();
	
	kiba_prefs_schema_fini (schema_data);
	kiba_prefs_fini();
	
	return EXIT_SUCCESS;
}

