/*
 *  Copyright (C) 2007 Torbjörn Svangård  <trptorbjorn@gmail.com>
 *  Copyright (C) 2007 Daniel Bessler  <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author(s) : Torbjörn Svangård  <trptorbjorn@gmail.com>
 *              Daniel    Bessler    <danielb@kiba-dock.org>
 */

#include "../config.h"

#include <stdlib.h>
#include <string.h>
#include <libintl.h>
#include <gtk/gtk.h>

#include "kiba-settings.h"
#include "settings-dbus.h"
#include "kiba-gradient-chooser.h"
#include "kiba-prefs.h"
#include "kiba-prefs-schema.h"
#include "widgets.h"

#define _(string) gettext (string)

#define EDITABLE_RANGE  "editable"
#define TEXT_RANGE      "text"
#define ICON_RANGE      "pngfile"
#define RGBA_RANGE      "rgba"
#define ENCRYPTED_RANGE "encrypt"

#define CAIRO_CLOCK_THEME_DIR  "/usr/share/cairo-clock/themes/"
#define CLOCK_DIR              "/clock"
#define SYSINFO_DIR            "/sysinfo"
#define STATUS_PUSH(bar, txt)  gtk_statusbar_push (bar, gtk_statusbar_get_context_id(bar, "info"), txt)

#define KIBA_DATA_DIR      DATADIR "/kiba-dock"
#define SYSINFO_THEME_DIR  KIBA_DATA_DIR "/icons/sysinfo_themes"

static GtkWidget *status_bar = NULL;
static gint status_pop_id = 0;

enum {
	TITLE_COLUMN,
	N_COLUMNS
};


void status_bar_push_message (const char *message) {
	STATUS_PUSH (GTK_STATUSBAR(status_bar), message);
}

/**
Helper Functions
**/

static GdkPixbuf* get_pixbuf_from_file (const char *file) {
	GdkPixbuf   *icon = NULL;
	GtkIconInfo *info = NULL;
	
	if (file) {
		icon = gdk_pixbuf_new_from_file_at_size (file, 48, -1, NULL);
		if (icon)
			return icon;
		
		info = gtk_icon_theme_lookup_icon (gtk_icon_theme_get_default (), file, 48, 0);
		if (info != NULL) {
			icon= gdk_pixbuf_new_from_file_at_size (gtk_icon_info_get_filename (info), 48, -1, NULL);
			gtk_icon_info_free (info);
			info = NULL;
		}
		if (icon)
			return icon;
	}
	
	info = gtk_icon_theme_lookup_icon (gtk_icon_theme_get_default (),
				"stock_missing-image", 48, 0);
	if (info != NULL) {
		icon= gdk_pixbuf_new_from_file_at_size (gtk_icon_info_get_filename (info),
				48, -1, NULL);
		gtk_icon_info_free (info);
	}
	
	return icon;
}

static GdkPixbuf* get_pixbuf_from_icon_list (GSList *icon_list) {
	GSList *l;
	GdkPixbuf *ret = NULL;
	
	for (l = icon_list; l != NULL; l = l->next) {
		char *icon_file = l->data;
		
		ret = get_pixbuf_from_file (icon_file);
		if (ret)
			break;
	}
	
	return ret;
}

static char *fixName(char *nameToFix) {
	char *ret;
	
	ret = g_strdup(g_ascii_strdown(nameToFix, strlen(nameToFix)));
	g_strdelimit(ret, "_", ' ');
	/*ret[0] = g_ascii_toupper(ret[0]);*/
	
	return ret;
}

/**
Callbacks
**/

typedef struct _ResetClosure {
	GtkWidget *widget;
	SchemaMember *member;
}ResetClosure;

static void combobox_set_default (GtkWidget  *combobox,
				  const char *default_value) {
	GtkTreeModel *model;
	GtkTreeIter   iter;
	gboolean      valid;
	
	model = gtk_combo_box_get_model (GTK_COMBO_BOX (combobox));
	valid = gtk_tree_model_get_iter_first (model, &iter);
	
	while (valid) {
		gchar *str_data;
	
		gtk_tree_model_get (model, &iter, 0, &str_data, -1);
		if (strcmp (default_value, str_data) == 0) {
			gtk_combo_box_set_active_iter (
				GTK_COMBO_BOX (combobox), &iter);
			break;
		}
		g_free (str_data);
		
		valid = gtk_tree_model_iter_next (model, &iter);
	}
}

static void reset_to_default (GtkButton *button,
			      gpointer   user_data) {
	ResetClosure *closure = user_data;
	
	const char *setting_name = gtk_widget_get_name (closure->widget);
	g_return_if_fail (setting_name != NULL);
	
	if (GTK_IS_CHECK_BUTTON (closure->widget))
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (closure->widget),
			strcmp (closure->member->default_value, "TRUE") == 0);
	else if (GTK_IS_COLOR_BUTTON (closure->widget)) {
		GdkColor col;
		unsigned int r,g,b, a;
		
		sscanf (closure->member->default_value,
				"#%02X%02X%02X%02X",
				&r, &g, &b, &a);
		col.red    = (float) r * 256;
		col.green  = (float) g * 256;
		col.blue   = (float) b * 256;
		gtk_color_button_set_color(GTK_COLOR_BUTTON (closure->widget), &col);
	
		if (gtk_color_button_get_use_alpha (GTK_COLOR_BUTTON (closure->widget)))
			gtk_color_button_set_alpha  (GTK_COLOR_BUTTON (closure->widget), a * 256);
	}
	else if (GTK_IS_SCALE (closure->widget))
		gtk_range_set_value (GTK_RANGE (closure->widget),
			atof (closure->member->default_value));
	else if (GTK_IS_SPIN_BUTTON (closure->widget))
		gtk_spin_button_set_value (GTK_SPIN_BUTTON (closure->widget),
			atof (closure->member->default_value));
	else if (GTK_IS_ENTRY (closure->widget))
		gtk_entry_set_text (GTK_ENTRY (closure->widget),
			closure->member->default_value);
	else if (GTK_IS_FILE_CHOOSER (closure->widget))
		gtk_file_chooser_set_filename (GTK_FILE_CHOOSER (closure->widget),
					       closure->member->default_value);
	else if (GTK_IS_COMBO_BOX (closure->widget)) {
		combobox_set_default (closure->widget,
			closure->member->default_value);
	}
	else if (KIBA_IS_GRADIENT_CHOOSER (closure->widget)) {
		GSList *items, *l;
		kiba_gradient_chooser_clear(KIBA_GRADIENT_CHOOSER(closure->widget));
		items = kiba_prefs_get_list_from_string (closure->member->default_value);
		for (l = items; l != NULL; l = l->next)
			kiba_gradient_chooser_add_color(KIBA_GRADIENT_CHOOSER(closure->widget), (char*) l->data);
// 		g_signal_connect(G_OBJECT(widget), "changed",
// 			G_CALLBACK(__gradient_chooser_changed), option);
//TODO
	}
//TODO list widgets
	else if (GTK_IS_HBOX (closure->widget)) {
		GList     *childs;
		GtkWidget *combobox;
		
		childs    = gtk_container_get_children (GTK_CONTAINER (closure->widget));
		combobox  = childs->data;
		combobox_set_default (combobox, closure->member->default_value);
		g_list_free (childs);
	}
	
	kiba_prefs_write_to_config_file (closure->member->widget_name,
				 closure->member->type,
				 closure->member->default_value);
	
	if (strcmp (closure->member->type, "int") == 0)
		kiba_settings_dbus_set_int (closure->member->widget_name,
			atoi(closure->member->default_value));
	else if (strcmp (closure->member->type, "double") == 0)
		kiba_settings_dbus_set_float (closure->member->widget_name,
			atof(closure->member->default_value));
	else if (strcmp (closure->member->type, "char") == 0)
		kiba_settings_dbus_set_char (closure->member->widget_name,
			closure->member->default_value);
	else if (strcmp (closure->member->type, "bool") == 0)
		kiba_settings_dbus_set_bool (closure->member->widget_name,
			(gboolean) (strcmp(closure->member->default_value, "TRUE") == 0));
	else if (strcmp (closure->member->type, "list") == 0)
		kiba_settings_dbus_set_list (closure->member->widget_name,
			closure->member->default_value);
	status_bar_push_message("Config File Saved");
}

static void set_spinbutton_value_from_range (GtkRange *range,
					     gpointer  user_data) {
	GtkWidget *spinbutton = user_data;
	
	gtk_spin_button_set_value (GTK_SPIN_BUTTON (spinbutton),
		gtk_range_get_value (range));
}

static void set_range_value_from_spinbutton (GtkSpinButton *spinbutton,
					     gpointer       user_data) {
	GtkWidget *range = user_data;
	
	gtk_range_set_value (GTK_RANGE (range),
		gtk_spin_button_get_value (spinbutton));
}

static void checkbutton_cb (GtkButton *check_button,
		   	    gpointer   user_data) {
	SchemaMember *data = user_data;
	const char *name;
	const char *value;
	
	name = gtk_widget_get_name (GTK_WIDGET (check_button));
	value = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(check_button)) ? "TRUE" : "FALSE";
	
	if (data->value) g_free(data->value);
	data->value = g_strdup(value);
	kiba_prefs_write_to_config_file (name, "bool", value);
	kiba_settings_dbus_set_bool (name, gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(check_button)));
	
	status_bar_push_message("Config File Saved");
}

static void __gradient_chooser_changed (KibaGradientChooser *chooser,
                                        gpointer data) {
	SchemaMember *settings = data;
	GList *colors = kiba_gradient_chooser_get_colors(chooser);
	char *text;
	
	const char* name = gtk_widget_get_name (GTK_WIDGET (chooser));
	
	if (settings->value) g_free(settings->value);
	settings->value = kiba_prefs_list_to_string(colors);
	kiba_prefs_write_to_config_file (name, "list", settings->value);
	kiba_settings_dbus_set_list (name, settings->value);
	
	status_bar_push_message("Config File Saved");
}

static void double_range_cb (GtkRange *range,
			     gpointer  user_data) {
	SchemaMember *data = user_data;
	char *value;
	
	const char* name = gtk_widget_get_name (GTK_WIDGET (range));
	
	/* save two decimals */
	value = g_strdup_printf("%0.2f", gtk_range_get_value (range));
	
	if (data->value) g_free(data->value);
	data->value = g_strdup(value);
	kiba_prefs_write_to_config_file (name, "double", value);
	kiba_settings_dbus_set_float (name, gtk_range_get_value (range));
	status_bar_push_message("Config File Saved");
	
	g_free(value);
}

static void rgb_cb (GtkColorButton *widget,
		    gpointer        user_data) {
	SchemaMember *data = user_data;
	GdkColor col;
	char *value;
	
	gtk_color_button_get_color(widget, &col);
	value = g_strdup_printf("#%02X%02X%02X",
				col.red / 256,
				col.green / 256,
				col.blue / 256);
	const char* name = gtk_widget_get_name(GTK_WIDGET(widget));
	
	if (data->value) g_free(data->value);
	data->value = g_strdup(value);
	kiba_prefs_write_to_config_file (name, "char", value);
	kiba_settings_dbus_set_char (name, value);
	status_bar_push_message("Config File Saved");
	
	g_free(value);
}

static void rgba_cb (GtkColorButton *widget,
		     gpointer        user_data) {
	SchemaMember *data = user_data;
	GdkColor col;
	int   alpha;
	char *value;
	
	gtk_color_button_get_color(widget, &col);
	alpha = gtk_color_button_get_alpha(widget);
	value = g_strdup_printf("#%02X%02X%02X%02X",
				col.red / 256,
				col.green / 256,
				col.blue / 256,
				alpha / 256);
	const char* name = gtk_widget_get_name(GTK_WIDGET(widget));
	
	if (data->value) g_free(data->value);
	data->value = g_strdup(value);
	kiba_prefs_write_to_config_file (name, "char", value);
	kiba_settings_dbus_set_char (name, value);
	status_bar_push_message("Config File Saved");
	
	g_free(value);
}

static void combobox_cb (GtkComboBox *combobox,
			 gpointer     user_data) {
	SchemaMember *data = user_data;
	const char* name = gtk_widget_get_name (GTK_WIDGET(combobox));
	const char* value = gtk_combo_box_get_active_text (combobox);
	
	if (data->value) g_free(data->value);
	data->value = g_strdup(value);
	kiba_prefs_write_to_config_file (name, "char", value);
	kiba_settings_dbus_set_char (name, value);
	status_bar_push_message("Config File Saved");
}

static void font_cb (GtkFontButton *font_button,
		     gpointer       user_data) {
	SchemaMember *data = user_data;
	const char* name  = gtk_widget_get_name (GTK_WIDGET(font_button));
	const char* value = gtk_font_button_get_font_name (font_button);
	
	if (data->value) g_free(data->value);
	data->value = g_strdup(value);
	kiba_prefs_write_to_config_file (name, "char", value);
	kiba_settings_dbus_set_char (name, value);
	status_bar_push_message("Config File Saved");
}

static void entry_cb (GtkEntry *entry,
		      gpointer  user_data) {
	SchemaMember *data = user_data;
	
	const char *name  = gtk_widget_get_name(GTK_WIDGET(entry));
	const char* value = gtk_entry_get_text (entry);
	
	if (data->value) g_free(data->value);
	data->value = g_strdup(value);
	kiba_prefs_write_to_config_file (name, "char", value);
	kiba_settings_dbus_set_char (name, value);
	status_bar_push_message("Config File Saved");
}

static void filechooser_cb (GtkFileChooser *chooser,
			    gpointer        user_data) {
	SchemaMember *data = user_data;
	const char *name  = gtk_widget_get_name (GTK_WIDGET(chooser));
	const char* value = gtk_file_chooser_get_filename (chooser);
	if (!value || strlen (value) < 2 || strcmp (value, "(null)") == 0 ||
	    g_str_has_suffix (value, ".desktop") || strcmp(value, data->value) == 0)
		return;
	
	if (data->value) g_free(data->value);
	data->value = g_strdup(value);
	kiba_prefs_write_to_config_file (name, "char", value);
	kiba_settings_dbus_set_char (name, value);
	status_bar_push_message("Config File Saved");
}

static void spinbutton_cb (GtkSpinButton *button,
			   gpointer       user_data) {
	SchemaMember *data = user_data;
	float value;
	char *char_value;
	
	const char *name  = gtk_widget_get_name (GTK_WIDGET (button));
	
	value = gtk_spin_button_get_value (button);
	char_value = g_strdup_printf ("%g", value);
	
	if (data->value) g_free(data->value);
	data->value = g_strdup(char_value);
	kiba_prefs_write_to_config_file (name, "int", char_value);
	kiba_settings_dbus_set_int (name, value);
	status_bar_push_message("Config File Saved");
	
	g_free (char_value);
}

void update_preview_cb (GtkFileChooser *file_chooser,
			gpointer        data) {
	GtkWidget *preview;
	GdkPixbuf *pixbuf;
	
	preview = GTK_WIDGET(data);
	const char* filename = gtk_file_chooser_get_preview_filename(file_chooser);
	
	if (filename != NULL) {
		pixbuf = gdk_pixbuf_new_from_file_at_size(filename, 128, 128, NULL);
		
		gtk_image_set_from_pixbuf (GTK_IMAGE(preview), pixbuf);
		
		gtk_file_chooser_set_preview_widget_active(file_chooser, pixbuf != NULL);
		if (pixbuf)
			gdk_pixbuf_unref(pixbuf);
	}
}

static void add_dialog_destroy_cb (GtkButton *check_button,
		   		   gpointer   user_data) {
	GtkWidget *dialog = user_data;
	
	gtk_widget_destroy (dialog);
}

static void add_dialog_accept_cb (GtkButton *button,
		   		  gpointer   user_data) {
	GtkTreeView *view = user_data;
	SchemaMember *settings;
	GtkTreeStore *store;
	GtkTreeIter iter;
	GtkWidget *entry;
	const char *text;
	char *buf;
	
	entry = g_object_get_data (G_OBJECT(button), "entry-widget");
	settings = g_object_get_data (G_OBJECT(view), "Pref-Member");
	text = gtk_entry_get_text (GTK_ENTRY(entry));
	
	if (!text || strlen(text) < 3)
		return;
	
	buf = g_strdup (settings->value);
	buf[strlen(buf)-1] = '\0';
	if (strcmp (buf, "[]") == 0)
		settings->value = g_strdup_printf ("[%s]", text);
	else
		settings->value = g_strdup_printf ("%s,%s]", buf, text);
	kiba_prefs_write_to_config_file (settings->widget_name, "list", settings->value);
	kiba_settings_dbus_set_list (settings->widget_name, settings->value);
	g_free (buf);
	
	store = GTK_TREE_STORE(gtk_tree_view_get_model(view));
	gtk_tree_store_append (store, &iter, NULL);  /* Acquire an iterator */
	
	gtk_tree_store_set (store, &iter, TITLE_COLUMN, text, -1);
}

static void cb_add_item_to_list (GtkWidget  *button,
				 gpointer    data) {
	GtkTreeView *view = GTK_TREE_VIEW(data);
	GtkWidget *dialog, *entry, *button_box, *box, *dialog_button;
	
	dialog = gtk_dialog_new();
	
	box = gtk_vbox_new(TRUE, 5);
	button_box = gtk_hbox_new(FALSE, 5);
	
	entry = gtk_entry_new ();
	gtk_widget_show (entry);
	gtk_box_pack_start (GTK_BOX (box), entry, TRUE, TRUE, 0);
	gtk_widget_show (box);
	
	dialog_button = gtk_button_new_from_stock(GTK_STOCK_OK);
	g_object_set_data (G_OBJECT(dialog_button), "entry-widget", entry);
	gtk_box_pack_start (GTK_BOX (button_box), dialog_button, FALSE, FALSE, 0);
	gtk_widget_show (dialog_button);
	g_signal_connect (G_OBJECT(dialog_button), "clicked",
		G_CALLBACK(add_dialog_accept_cb), view);
	g_signal_connect_after (G_OBJECT(dialog_button), "clicked",
		G_CALLBACK(add_dialog_destroy_cb), dialog);
	dialog_button = gtk_button_new_from_stock(GTK_STOCK_CANCEL);
	gtk_box_pack_start (GTK_BOX (button_box), dialog_button, FALSE, FALSE, 0);
	gtk_widget_show (dialog_button);
	g_signal_connect (G_OBJECT(dialog_button), "clicked",
		G_CALLBACK(add_dialog_destroy_cb), dialog);

	gtk_box_pack_start (GTK_BOX (box), button_box, TRUE, TRUE, 0);
	gtk_widget_show (button_box);

	gtk_dialog_add_action_widget (GTK_DIALOG(dialog), box, 13);
	gtk_widget_show (dialog);
}

static void cb_remove_item_from_list_real (GtkTreeModel  *model,
					   GtkTreePath   *path,
					   GtkTreeIter   *iter,
					   gpointer       data) {
	GtkTreeView *view = GTK_TREE_VIEW(data);
	SchemaMember *settings;
	GSList *list, *l;
	gchar *text, *string = NULL;
	gint count;
	
	settings = g_object_get_data (G_OBJECT(view), "Pref-Member");
	list = kiba_prefs_get_list_from_string(settings->value);
	
	gtk_tree_model_get (model, iter, TITLE_COLUMN, &text, -1);
	
	count = 0;
	for (l = list; l != NULL; l = l->next) {
		char *item = l->data;
		char *buf = NULL;
		
		if (strcmp (item, text) == 0)
			continue;
		if (string) {
			buf = g_strdup (string);
			g_free (string);
			string = NULL;
		}
		
		if (!count)
			string = g_strdup_printf ("[%s,", item);
		else
			string = g_strdup_printf ("%s%s,", buf, item);
		
		count++;
	}
	if (string)
		string[strlen(string)-1] = ']';
	else
		string = g_strdup ("[]");
	settings->value = string;
	kiba_prefs_write_to_config_file (settings->widget_name, "list", string);
	kiba_settings_dbus_set_list (settings->widget_name, string);
	status_bar_push_message("Config File Saved");
	
	gtk_tree_model_row_deleted (model, path);
}

static void cb_remove_item_from_list (GtkWidget *button,
				      gpointer   data) {
	GtkTreeView *view = GTK_TREE_VIEW(data);
	GtkTreeSelection  *selection;
	
	selection = gtk_tree_view_get_selection (view);
	gtk_tree_selection_selected_foreach (selection,
		cb_remove_item_from_list_real, view);
}

static gint status_pop (gpointer data) {
	GtkStatusbar *bar = data;
	
	gtk_statusbar_pop (bar, gtk_statusbar_get_context_id(bar, "info"));
	
	status_pop_id = 0;
	
	return FALSE;
}

static void status_pushed (GtkStatusbar *statusbar,
			   guint         context_id,
			   gchar        *text,
			   gpointer      user_data) {
	if (status_pop_id) {
		gtk_statusbar_pop (statusbar, gtk_statusbar_get_context_id(statusbar, "info"));
		g_source_remove (status_pop_id);
	}
	
	status_pop_id = g_timeout_add (2*1000, status_pop, statusbar);
}

/**
Widget Creation
**/

static GtkWidget* get_hbox_from_widget (GtkWidget *widget) {
	GtkWidget *box;
	
	box = gtk_hbox_new (FALSE, 0);
	
	gtk_box_pack_start (GTK_BOX (box),
			    widget,
			    FALSE, FALSE, 0);
	
	return box;
}

static GtkWidget *create_main_window_widget(char *title,
					    int   x,
					    int   y) {
	GtkWidget *widget;
	
	widget = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(widget), title);
	gtk_widget_set_usize(widget, x, y);
	g_signal_connect (G_OBJECT (widget),"delete_event",
		G_CALLBACK(gtk_main_quit), NULL);
	
	return widget;
}

static GtkWidget *create_double_hscale_widget (SchemaMember *data) {
	GtkWidget *widget;
	GtkObject *adjustment;
	char *low = NULL,*high = NULL;
	
	/* data.the[i].range+1  if the first char is - (negative) */
	high = g_strstr_len(data->range+1, strlen(data->range), "-") + 1;
	low = g_strndup(data->range, strlen(data->range) - strlen(high) - 1);
	adjustment = gtk_adjustment_new(atof(data->value),
					atof(low),
					atof(high),
					0.1, 0.0, 0.0);
	g_free(low);
	
	widget = gtk_hscale_new(GTK_ADJUSTMENT(adjustment));
	gtk_scale_set_draw_value (GTK_SCALE(widget), FALSE);
	gtk_scale_set_digits (GTK_SCALE(widget), 2);
	gtk_widget_set_name(widget, data->widget_name);
	
	gtk_tooltips_set_tip(kiba_settings_get_tooltips(), widget, data->description, NULL);
	
 	g_signal_connect (G_OBJECT (widget), "value-changed",
		G_CALLBACK (double_range_cb), data);
	
	return widget;
}

static GtkWidget *create_label_widget (const char *text) {
	GtkWidget *widget;
	
	widget = gtk_label_new(text);
	gtk_misc_set_alignment (GTK_MISC(widget), 0.0, 0.5);
	
	gtk_widget_set_name(widget, text);
	
	return widget;
}

static GtkWidget *create_colour_widget (SchemaMember *data) {
	GtkWidget *widget;
	GdkColor col;
	unsigned int r,g,b;
	
	widget = gtk_color_button_new();
	
	sscanf(data->value, "#%02X%02X%02X", &r, &g, &b);
	col.red    = (float) r * 256;
	col.green  = (float) g * 256;
	col.blue   = (float) b * 256;
	gtk_color_button_set_color((GtkColorButton *)widget, &col);
	
 	g_signal_connect (G_OBJECT (widget), "color-set",
		G_CALLBACK (rgb_cb), data);
	
	gtk_widget_set_name(widget, data->widget_name);
	gtk_tooltips_set_tip(kiba_settings_get_tooltips(), widget, data->description, NULL);
	
	return widget;
}

static GtkWidget *create_rgba_widget (SchemaMember *data) {
	GtkWidget *widget;
	GdkColor col;
	unsigned int r,g,b, alpha;
	
	widget = gtk_color_button_new();
	
	sscanf(data->value, "#%02X%02X%02X%02X", &r, &g, &b, &alpha);
	col.red    = (float) r * 256;
	col.green  = (float) g * 256;
	col.blue   = (float) b * 256;
	gtk_color_button_set_color(GTK_COLOR_BUTTON (widget), &col);
	
	gtk_color_button_set_use_alpha (GTK_COLOR_BUTTON (widget), TRUE);
	gtk_color_button_set_alpha  (GTK_COLOR_BUTTON (widget), alpha * 256);
	
 	g_signal_connect (G_OBJECT (widget), "color-set",
		G_CALLBACK (rgba_cb), data);
	
	gtk_widget_set_name(widget, data->widget_name);
	gtk_tooltips_set_tip(kiba_settings_get_tooltips(), widget, data->description, NULL);
	
	return widget;
}

static void get_clock_themes (GtkWidget *widget, const char *theme_directory) {
	GDir    *dir;
	GError  *error = NULL;
	char    *buf;
	
	dir = g_dir_open (theme_directory, 0, &error);
	if (error != NULL) {
		g_message ("Failed to open Theme Directory %s\notification: %s\n",
					theme_directory, error->message);
		g_error_free (error);
		return;
	}
	
	while ((buf = (char *) g_dir_read_name (dir)) != NULL)
		gtk_combo_box_append_text (GTK_COMBO_BOX(widget), buf);
}

static GtkWidget *create_combobox_widget (SchemaMember *data) {
	GtkWidget *combobox;
	
	combobox = gtk_combo_box_new_text();
	/* append the string from the configfile first to the combobox*/
	if (strcmp(data->value, ICON_RANGE) != 0 && !g_str_has_suffix (data->value, ".png") &&
	   !g_str_has_suffix (data->value, ".svg"))
		gtk_combo_box_append_text(GTK_COMBO_BOX(combobox), data->value);
	
	/* if a range exist */
	if (data->range) {
		guint f;
		gchar **strings = g_strsplit(data->range, ",", -1);
	
		/* append all strings to the combobox... but not the one that is set in config file */
		for (f = 0; f <  g_strv_length(strings); f++)
			if (strcmp(data->value, strings[f]) != 0 &&
			    strcmp(strings[f], ICON_RANGE) != 0 &&
			   !g_str_has_suffix (strings[f], ".png") &&
			   !g_str_has_suffix (strings[f], ".svg"))
				gtk_combo_box_append_text(GTK_COMBO_BOX(combobox), strings[f]);
	
		g_strfreev(strings);
	}
	else if (strcmp(data->widget_name, "clock_theme") == 0) {
		char *theme_directory = NULL;
	
		theme_directory = g_strdup_printf ("%s/.kiba-dock" CLOCK_DIR, g_get_home_dir ());
		if (g_file_test (theme_directory, G_FILE_TEST_IS_DIR))
			get_clock_themes (combobox, theme_directory);
		g_free (theme_directory);
		
		if (g_file_test (CAIRO_CLOCK_THEME_DIR, G_FILE_TEST_IS_DIR))
			get_clock_themes (combobox, CAIRO_CLOCK_THEME_DIR);
	}
	else if (strcmp(data->widget_name, "sysinfo_theme") == 0) {
		char *theme_directory = NULL;
	
		theme_directory = g_strdup_printf ("%s/.kiba-dock" SYSINFO_DIR "/themes", g_get_home_dir ());
		if (g_file_test (theme_directory, G_FILE_TEST_IS_DIR))
			get_clock_themes (combobox, theme_directory);
		g_free (theme_directory);
		
		if (g_file_test (SYSINFO_THEME_DIR, G_FILE_TEST_IS_DIR))
			get_clock_themes (combobox, SYSINFO_THEME_DIR);
	}
	
	gtk_combo_box_set_active(GTK_COMBO_BOX(combobox), 0);

	g_signal_connect (G_OBJECT (combobox), "changed",
		G_CALLBACK (combobox_cb), data);
	
	gtk_widget_set_name(combobox, data->widget_name);
	
	return combobox;
}

static GtkWidget *create_checkbox_widget (SchemaMember *data) {
	GtkWidget *widget;
	
	widget = gtk_check_button_new();
	
	gtk_widget_set_name(widget, data->widget_name);
	gtk_tooltips_set_tip(kiba_settings_get_tooltips(), widget, data->description, NULL);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(widget),
				(g_str_has_prefix(data->value, "TRUE")) ? TRUE : FALSE);
 	g_signal_connect (G_OBJECT (widget), "clicked",
		G_CALLBACK (checkbutton_cb), data);
	
	return widget;
}

static GtkWidget *create_filechooser_widget(SchemaMember *data) {
	GtkWidget *widget, *filechooser;
	GtkWidget *preview;
	char      *name;
	
	widget = gtk_event_box_new ();
	
	name = fixName(data->widget_name);
	filechooser = gtk_file_chooser_button_new(name, GTK_FILE_CHOOSER_ACTION_OPEN);
	g_free(name);
	
	gtk_widget_set_name (filechooser, data->widget_name);
	if (g_strrstr (data->value, ".png") ||  g_strrstr (data->value, ".svg"))
		gtk_file_chooser_set_filename (GTK_FILE_CHOOSER (filechooser), data->value);
	else if (g_str_has_suffix (data->widget_name, "_background") &&
		 g_file_test ("/usr/share/nautilus/patterns", G_FILE_TEST_IS_DIR)) {
		gtk_file_chooser_set_current_folder (
			GTK_FILE_CHOOSER(filechooser),
			"/usr/share/nautilus/patterns");
		gtk_file_chooser_add_shortcut_folder (
			GTK_FILE_CHOOSER(filechooser),
			"/usr/share/nautilus/patterns", NULL);
	}
	g_signal_connect (G_OBJECT (filechooser), "selection-changed",
		G_CALLBACK (filechooser_cb), data);
	
	preview = gtk_image_new ();
	gtk_file_chooser_set_preview_widget (GTK_FILE_CHOOSER(filechooser), preview);
	g_signal_connect (GTK_FILE_CHOOSER(filechooser), "update-preview",
			G_CALLBACK(update_preview_cb), preview);
	
	/* fix for tooltips not showing up,
	   remove the eventbox when http://bugzilla.gnome.org/show_bug.cgi?id=303717 is closed */
	gtk_container_add (GTK_CONTAINER (widget), filechooser);
	gtk_tooltips_set_tip(kiba_settings_get_tooltips(),
		widget, data->description, NULL);
	
	return widget;
}

static GtkWidget *create_font_widget (SchemaMember *data) {
	GtkWidget *widget;
	char *name, *font_description;
	
	name = fixName(data->widget_name);
	widget = gtk_font_button_new();
	g_free(name);
	
	gtk_widget_set_name(widget, data->widget_name);
	gtk_tooltips_set_tip(kiba_settings_get_tooltips(), widget, data->description, NULL);
	
	font_description = kiba_prefs_get_option_from_file (data->widget_name, "char");
	gtk_font_button_set_font_name (GTK_FONT_BUTTON (widget), font_description);
	
 	g_signal_connect (G_OBJECT (widget), "font-set",
		G_CALLBACK (font_cb), data);
	
	return widget;
}

static GtkWidget *get_float_spinbutton_widget (const char* key,
					       const char* value,
					       const char *range) {
	GtkWidget *spinbutton;
	float min = 0.0, max = 100.0, current;
	
	current = atof (value);
	
	if (range && strlen (range) > 2) {
		char *min_string  = NULL,
		     *max_string = NULL;
		
		max_string = g_strstr_len (range+1,
				strlen(range), "-") + 1;
		min_string = g_strndup (range,
				strlen(range) - strlen(max_string) - 1);
		
		min = atof (min_string);
		max = atof (max_string);
		
		g_free(min_string);
	}
	
	spinbutton = gtk_spin_button_new_with_range (min, max, max/100);
	gtk_spin_button_set_value (GTK_SPIN_BUTTON (spinbutton), current);
	gtk_spin_button_set_digits (GTK_SPIN_BUTTON (spinbutton), 2);
	gtk_entry_set_width_chars (GTK_ENTRY (spinbutton), 7);
	gtk_widget_set_name (spinbutton, key);
	
	return spinbutton;
}

static GtkWidget *create_list_widget (SchemaMember *data) {
	GtkTreeStore *store;
	GtkWidget *tree;
	GtkTreeViewColumn *column;
	GtkCellRenderer *renderer;
	GSList *items, *l;
	
	items = kiba_prefs_get_list_from_string (data->value);
	g_return_val_if_fail (items != NULL, NULL);
	
	/* Create a model. */
	store = gtk_tree_store_new (N_COLUMNS, G_TYPE_STRING);
	
	/* fill the model with data */
	for (l = items; l != NULL; l = l->next) {
		char *item = l->data;
		GtkTreeIter iter;
		
		gtk_tree_store_append (store, &iter, NULL);  /* Acquire an iterator */
		
		gtk_tree_store_set (store, &iter, TITLE_COLUMN, item, -1);
	}
	
	/* Create a view */
	tree = gtk_tree_view_new_with_model (GTK_TREE_MODEL (store));
	// gtk_tree_view_set_enable_search (GTK_TREE_VIEW (tree), TRUE);
	// gtk_tree_view_set_enable_tree_lines (GTK_TREE_VIEW (tree), TRUE);
	// gtk_tree_view_set_headers_visible (GTK_TREE_VIEW (tree), FALSE);
	g_object_unref (G_OBJECT (store));
	
	/* Create a cell render */
	renderer = gtk_cell_renderer_text_new ();
	
	/* Create a column, associating the "text" attribute of the
	* cell_renderer to the first column of the model */
	column = gtk_tree_view_column_new_with_attributes (data->description, renderer,
							"text", TITLE_COLUMN,
							NULL);
	gtk_tree_view_append_column (GTK_TREE_VIEW (tree), column);
	
	return tree;
}


static GtkWidget *get_spinbutton_widget (SchemaMember *data) {
	GtkWidget *spinbutton;
	gint min = 0, max = 100, current;
	
	current = atoi (data->value);
	
	if (data->range && strlen (data->range) > 2) {
		char *min_string  = NULL,
		     *max_string = NULL;
		
		max_string = g_strstr_len (data->range+1,
				strlen(data->range), "-") + 1;
		min_string = g_strndup (data->range,
				strlen(data->range) - strlen(max_string) - 1);
		
		min = atoi (min_string);
		max = atoi (max_string);
		
		g_free(min_string);
	}
	
	spinbutton = gtk_spin_button_new_with_range (min, max, 1);
	gtk_spin_button_set_digits (GTK_SPIN_BUTTON (spinbutton), 0);
	gtk_spin_button_set_value (GTK_SPIN_BUTTON (spinbutton), current);
	gtk_spin_button_set_digits (GTK_SPIN_BUTTON (spinbutton), 0);
	gtk_widget_set_name (spinbutton, data->widget_name);
 	g_signal_connect (G_OBJECT (spinbutton), "value-changed",
		G_CALLBACK (spinbutton_cb), data);
	
	return spinbutton;
}

static GtkWidget *fill_subsection(SchemaSubSection *subsection) {
	GtkWidget *container;
	GtkWidget *widget;
	GtkWidget *scrolledwindow;
	int i=-1, c=0;
	GList *l;
	
	scrolledwindow = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolledwindow),
					GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	
	/* insert a table inside the scrollable window */
	container = gtk_table_new(g_list_length(subsection->options), 3, FALSE);
	gtk_table_set_col_spacings (GTK_TABLE(container), 5);
	gtk_table_set_row_spacings (GTK_TABLE(container), 5);
	
	for(l = subsection->options; l != NULL; l = l->next) {
		SchemaMember *option = l->data;
		i++;
		if (strcmp(option->type, "bool") == 0) {
			
			widget = create_label_widget(option->title);
			gtk_table_attach (GTK_TABLE(container), widget,
				1, 2 , c, 1+c, GTK_FILL, GTK_FILL, 0, 0);
			
			widget = create_checkbox_widget(option);
			gtk_table_attach (GTK_TABLE(container), get_hbox_from_widget (widget),
				2, 3, c, 1+c, GTK_FILL, GTK_FILL, 0, 0);
		}
		else if (strcmp(option->type, "int") == 0) {
			/* skip position widgets for now */
			if (g_str_has_suffix (option->widget_name, "_position"))
				continue;
	
			widget = create_label_widget(option->title);
			gtk_table_attach (GTK_TABLE(container), widget,
				1, 2 , c, 1+c, GTK_FILL, GTK_FILL, 0, 0);
	
			widget = get_spinbutton_widget(option);
			gtk_tooltips_set_tip(kiba_settings_get_tooltips(),
				widget, option->description, NULL);
			gtk_table_attach (GTK_TABLE(container), get_hbox_from_widget (widget),
				2, 3, c, 1+c, GTK_FILL, GTK_FILL, 0, 0);
		}
		else if (strcmp(option->type, "double") == 0 &&
			 strcmp(option->range, TEXT_RANGE) != 0 &&
			 strcmp(option->range, "") != 0 ) {
			GtkWidget *box, *spinbutton;
			
			widget = create_label_widget(option->title);
			gtk_table_attach (GTK_TABLE(container), widget,
				1, 2 , c, 1+c, GTK_FILL, GTK_FILL, 0, 0);
			
			box = gtk_hbox_new (FALSE, 0);
			widget = create_double_hscale_widget(option);
			gtk_box_pack_start (GTK_BOX (box),
					widget,
					TRUE, TRUE, 0);
			spinbutton = get_float_spinbutton_widget (option->title,
					option->value,
					option->range);
			gtk_box_pack_start (GTK_BOX (box),
					spinbutton,
					FALSE, FALSE, 0);
			
			g_signal_connect (G_OBJECT(widget), "value-changed",
				G_CALLBACK (set_spinbutton_value_from_range), spinbutton);
			g_signal_connect (G_OBJECT(spinbutton), "value-changed",
				G_CALLBACK (set_range_value_from_spinbutton), widget);
			
			gtk_table_attach (GTK_TABLE(container), box,
				2, 3, c, 1+c, GTK_FILL | GTK_EXPAND, GTK_FILL, 0, 0);
		}
		else if (strcmp(option->type, "list") == 0) {
			if (option->range &&
			    strcmp (option->range, RGBA_RANGE) == 0) {
				GSList *items, *l;
				widget = kiba_gradient_chooser_new();
				gtk_widget_set_name(widget, option->widget_name);
				items = kiba_prefs_get_list_from_string (option->value);
				for (l = items; l != NULL; l = l->next)
					kiba_gradient_chooser_add_color(KIBA_GRADIENT_CHOOSER(widget), (char*) l->data);
				g_signal_connect(G_OBJECT(widget), "changed",
					G_CALLBACK(__gradient_chooser_changed), option);
				gtk_table_attach(GTK_TABLE(container), widget,
					2, 3, c, 1+c, GTK_FILL,GTK_FILL, 0, 0);
				GtkWidget *label_widget = create_label_widget(option->title);
				gtk_table_attach (GTK_TABLE(container), label_widget,
					1, 2 , c, 1+c, GTK_FILL, GTK_FILL, 0, 0);
			}
			else {
				GtkWidget *button;
				GtkWidget *vbox;
				GtkWidget *image;
				GtkWidget *list_box;
				
				list_box = gtk_hbox_new (FALSE, 0);
				widget = create_list_widget (option);
				g_object_set_data  (G_OBJECT(widget), "Pref-Member", option);
				gtk_box_pack_start (GTK_BOX(list_box), widget, TRUE, TRUE, 0);
				
				vbox = gtk_vbox_new (TRUE, 0);
				button = gtk_button_new ();
				image = gtk_image_new_from_stock (GTK_STOCK_ADD, GTK_ICON_SIZE_BUTTON);
				gtk_button_set_image (GTK_BUTTON (button), image);
				g_signal_connect (G_OBJECT(button), "clicked",
					G_CALLBACK (cb_add_item_to_list), widget);
				gtk_box_pack_start (GTK_BOX(vbox), button, FALSE, FALSE, 0);
				
				button = gtk_button_new ();
				image = gtk_image_new_from_stock (GTK_STOCK_REMOVE, GTK_ICON_SIZE_BUTTON);
				gtk_button_set_image (GTK_BUTTON (button), image);
				g_signal_connect (G_OBJECT(button), "clicked",
					G_CALLBACK (cb_remove_item_from_list), widget);
				gtk_box_pack_end (GTK_BOX(vbox), button, FALSE, FALSE, 0);
	
				gtk_box_pack_start (GTK_BOX(list_box), vbox, FALSE, FALSE, 0);
	
				gtk_table_attach(GTK_TABLE(container), list_box,
					1, 3 , c, 1+c,
					GTK_FILL | GTK_EXPAND, GTK_FILL,
					0, 0);
			}
		}
		else if (strcmp(option->type, "char") == 0) {
			if (option->range &&
			    strcmp (option->range, RGBA_RANGE) == 0)
				widget = create_rgba_widget(option);
			else if (option->value[0] == '#')
				widget = create_colour_widget(option);
			/* combobox + filechooser */
			else if (option->range &&
				 g_strrstr (option->range, ICON_RANGE)) {
				GtkWidget *file_chooser;
				int        fnord = 1;
				
				widget = gtk_hbox_new (FALSE, 5);
				
				if (strcmp (option->range, ICON_RANGE) != 0) {
					GtkWidget *combobox;
					
					combobox = create_combobox_widget(option);
					gtk_box_pack_start (GTK_BOX (widget),
							combobox,
							FALSE, FALSE, 0);
					fnord++;
				}
				
				file_chooser = create_filechooser_widget(option);
				gtk_box_pack_start (GTK_BOX (widget),
						file_chooser,
						FALSE, FALSE, 0);
			}
			else if (option->range &&
				 g_strrstr (option->range, TEXT_RANGE))
				widget = create_font_widget(option);
			else if (option->range &&
				 strcmp (option->range, EDITABLE_RANGE) == 0) {
				widget = get_entry_widget (option->widget_name,
						option->value,
						option->description);
				g_signal_connect (G_OBJECT (widget), "activate",
					G_CALLBACK (entry_cb), option);
			}
			else
				widget = create_combobox_widget (option);
			
			GtkWidget *label_widget;
			
			gtk_table_attach(GTK_TABLE(container), get_hbox_from_widget (widget),
				2, 3, c, 1+c, GTK_FILL,GTK_FILL, 0, 0);
			label_widget = create_label_widget(option->title);
			gtk_table_attach (GTK_TABLE(container), label_widget,
				1, 2 , c, 1+c, GTK_FILL, GTK_FILL, 0, 0);
		}
		else
			continue;
		
		GtkWidget *image,
			  *button;
		
		button = gtk_button_new ();
		image = gtk_image_new_from_stock (GTK_STOCK_CLEAR,
						GTK_ICON_SIZE_BUTTON);
		gtk_button_set_image (GTK_BUTTON (button), image);
		gtk_tooltips_set_tip(kiba_settings_get_tooltips(),
			button, _("Reset to default"), NULL);
		gtk_table_attach (GTK_TABLE(container), button,
			0, 1 , c, 1+c, 0, 0, 0, 0);
		
		ResetClosure *closure = g_new0 (ResetClosure, 1);
		closure->widget = widget;
		closure->member = option;
		g_signal_connect (G_OBJECT (button), "clicked",
			G_CALLBACK (reset_to_default), closure);
		
		c++;
	}
	
	gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(scrolledwindow),
						container);
	
	return scrolledwindow;
}

static gboolean subsection_exist(const gchar *section,
			         GtkNotebook *notebook) {
	gint i;
	
	/* check all pages in notebook if section name exist */
	for (i = 0; i < gtk_notebook_get_n_pages(notebook); i++) {
		if (strcmp (section, gtk_notebook_get_tab_label_text(notebook,
		    gtk_notebook_get_nth_page(notebook, i))) == 0)
			return TRUE;
	}
	
	return FALSE;
}

static gboolean section_exist (const gchar *section,
			       GtkWidget   *button_box) {
	GList *buttons, *l;
	
	buttons = gtk_container_get_children (GTK_CONTAINER (button_box));
	
	for (l = buttons; l != NULL; l = l->next) {
		GtkWidget *box = l->data;
		GList     *box_childs, *li;
		
		box_childs = gtk_container_get_children (GTK_CONTAINER (box));
		
		for (li = box_childs; li != NULL; li = li->next) {
			GtkWidget *child = li->data;
			
			if (!GTK_IS_LABEL (child))
				continue;
			
			const gchar* label_text = gtk_label_get_text (GTK_LABEL (child));
			
			if (strcmp (g_ascii_strdown (label_text, strlen(label_text)),
				    g_ascii_strdown (section, strlen(section))) == 0)
				return TRUE;
			
			break;
		}
	}
	
	return FALSE;
}

static void clear_notebook (GtkWidget *notebook) {
	gint i, num_pages;
	
	num_pages = gtk_notebook_get_n_pages (GTK_NOTEBOOK (notebook));
	
	for (i = 0; i < num_pages; i++)
		gtk_notebook_remove_page (GTK_NOTEBOOK (notebook), -1);
}

static void activate_section (SchemaSection  *section,
			      GtkWidget      *setting_notebook,
			      GtkWidget      *section_label,
			      GtkWidget      *section_description,
			      GtkWidget      *button_label) {
	gchar *markup;
	GList *l;
	
	clear_notebook (setting_notebook);
	
	gtk_label_set_use_markup (GTK_LABEL (button_label), TRUE);
	
	/* set title text */
	markup = g_markup_printf_escaped ("<big><b>%s</b></big>", section->section_name);
	gtk_label_set_markup (GTK_LABEL (section_label), markup);
	g_free (markup);
	
	/* set description text */
	gtk_label_set_text (GTK_LABEL (section_description), section->description);
	
	/* set button label markup */
	markup = g_markup_printf_escaped ("<span style=\"italic\"><b>%s</b></span>",
			gtk_label_get_text (GTK_LABEL (button_label)));
	gtk_label_set_markup (GTK_LABEL (button_label), markup);
	g_free (markup);
	
	for (l = section->subsections; l != NULL; l = l->next) {
		SchemaSubSection *subsection = l->data;
		if (!subsection_exist(subsection->subsection_name, GTK_NOTEBOOK (setting_notebook))) {
			gtk_notebook_append_page(GTK_NOTEBOOK(setting_notebook),
				fill_subsection(subsection),
				gtk_label_new(subsection->subsection_name));
		}
	}
}

typedef struct _SectionPressClosure {
	SchemaSection  *section;
	GdkPixbuf      *section_pixbuf;
	GtkWidget      *notebook;
	GtkWidget      *title;
	GtkWidget      *title_box;
	GtkWidget      *title_image;
	GtkWidget      *title_description;
	GtkWidget      *box_image;
	GtkWidget      *label;
	GList          *labels;
} SectionPressClosure;

static void section_button_clicked (GtkWidget *button,
				    gpointer   data) {
	SectionPressClosure *closure = data;
	GList *l;
	
	/* remove markup from previous selected label */
	for (l = closure->labels; l != NULL; l = l->next) {
		GtkWidget *label = l->data;
		
		if (!GTK_IS_LABEL (label))
			continue;
		
		if (gtk_label_get_use_markup (GTK_LABEL (label)))
			gtk_label_set_text (GTK_LABEL (label),
				    gtk_label_get_text (GTK_LABEL (label)));
	}
	
	/* set title image */
	gtk_image_set_from_pixbuf (GTK_IMAGE (closure->title_image), closure->section_pixbuf);
	/* highlight button_pixbuf */
/*
	GdkPixbuf *highlight_pixbuf = NULL;
	highlight_pixbuf = gdk_pixbuf_add_alpha (closure->section_pixbuf, FALSE, 255, 255, 255);
	gtk_image_set_from_pixbuf (GTK_IMAGE (closure->box_image), highlight_pixbuf);
*/
		
	activate_section (closure->section, closure->notebook,
			closure->title, closure->title_description, closure->label);
	
	gtk_widget_grab_focus (closure->notebook);
	gtk_widget_show_all (closure->notebook);
}

/**
Public Functions
**/

GtkWidget* get_entry_widget (const char* id,
			     const char* value,
			     const char* description) {
	GtkWidget *entry;
	
// 	widget = gtk_event_box_new ();
	entry = gtk_entry_new ();
	
	if (value)
		gtk_entry_set_text (GTK_ENTRY (entry), value);
	
	gtk_widget_set_name(entry, id);
	
	/* fix for tooltips not showing up,
	   remove the eventbox when http://bugzilla.gnome.org/show_bug.cgi?id=303717 is closed */
// 	gtk_container_add (GTK_CONTAINER (widget), entry);
// 	if (description)
// 		gtk_tooltips_set_tip (kiba_settings_get_tooltips(),
// 			widget, description, NULL);
	
// 	return widget;
	return entry;
}

void kiba_settings_create_window (KibaSchema *schema_data) {
	GtkWidget *window,
		  *main_container,
		  *top_container,
		  *section_frame,
		  *setting_box,
		  *section_label,
		  *section_image,
		  *section_description,
		  *section_scrolled_window,
		  *section_button_box,
		  *setting_notebook,
		  *bottom_container,
		  *quit_button,
		  *title_hbox,
		  *title_vbox,
		  *main_frame;
	GdkPixbuf *kiba_icon;
	GList *labels = NULL, *l;
	
	window = create_main_window_widget(WINDOW_LABEL, 650, 400);
	
	kiba_icon = gdk_pixbuf_new_from_inline (-1, app_fallback_icon, FALSE, NULL);
	gtk_window_set_icon (GTK_WINDOW (window), kiba_icon);
	g_object_unref (kiba_icon);
	
	main_container = gtk_vbox_new (FALSE, 0);
	gtk_container_set_border_width (GTK_CONTAINER (main_container), 5);
	gtk_container_add(GTK_CONTAINER(window), main_container);
	
	main_frame = gtk_frame_new (NULL);
	gtk_frame_set_shadow_type (GTK_FRAME (main_frame), GTK_SHADOW_ETCHED_IN);
	
	top_container = gtk_hbox_new (FALSE, 0);
	
	/* create Section container Widgets */
	section_frame = gtk_frame_new (_("Sections"));
	gtk_frame_set_shadow_type (GTK_FRAME (section_frame), GTK_SHADOW_ETCHED_IN);
	gtk_container_set_border_width (GTK_CONTAINER (section_frame), 5);
	
	section_scrolled_window = gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (section_scrolled_window),
			GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
	section_button_box = gtk_vbutton_box_new ();
	gtk_button_box_set_layout (GTK_BUTTON_BOX (section_button_box), GTK_BUTTONBOX_START);
	gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (section_scrolled_window),
					       section_button_box);
	gtk_container_add(GTK_CONTAINER(section_frame), section_scrolled_window);
	gtk_box_pack_start (GTK_BOX (top_container),
			    section_frame,
			    FALSE, FALSE, 0);
	
	/* create Settings container widgets */
	setting_box = gtk_vbox_new (FALSE, 5);
	gtk_container_set_border_width (GTK_CONTAINER (setting_box), 5);
	title_hbox = gtk_hbox_new (FALSE, 5);

	section_image = gtk_image_new_from_stock (GTK_STOCK_MISSING_IMAGE,
						GTK_ICON_SIZE_DIALOG);
	gtk_box_pack_start (GTK_BOX (title_hbox),
			    section_image,
			    FALSE, FALSE, 0);
	
	title_vbox = gtk_vbox_new (FALSE, 0);
	section_label = gtk_label_new (NULL);
	gtk_misc_set_alignment (GTK_MISC (section_label), 0.0, 0.5);
	gtk_box_pack_start (GTK_BOX (title_vbox),
			    section_label,
			    FALSE, FALSE, 0);
	section_description = gtk_label_new (NULL);
	gtk_misc_set_alignment (GTK_MISC (section_label), 0.0, 0.5);
	gtk_label_set_line_wrap (GTK_LABEL(section_description), TRUE);
	gtk_box_pack_start (GTK_BOX (title_vbox),
			    section_description,
			    FALSE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX (title_hbox),
			    title_vbox,
			    FALSE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX (setting_box),
			    title_hbox,
			    FALSE, FALSE, 0);
	setting_notebook = gtk_notebook_new();
	gtk_notebook_set_tab_pos (GTK_NOTEBOOK(setting_notebook), GTK_POS_TOP);
	gtk_notebook_set_scrollable (GTK_NOTEBOOK(setting_notebook), TRUE);
	gtk_box_pack_start (GTK_BOX (setting_box),
			    setting_notebook,
			    TRUE, TRUE, 0);
	gtk_box_pack_start (GTK_BOX (top_container),
			    setting_box,
			    TRUE, TRUE, 0);
	
	gtk_container_add (GTK_CONTAINER(main_frame), top_container);
	
	/* create Bottom Container (Quit Button) */
	bottom_container = gtk_hbox_new (FALSE, 0);
	gtk_container_set_border_width (GTK_CONTAINER (bottom_container), 5);
	quit_button = gtk_button_new_from_stock (GTK_STOCK_QUIT);
	g_signal_connect (G_OBJECT (quit_button), "released", gtk_main_quit, NULL);
	gtk_box_pack_end (GTK_BOX (bottom_container),
			    quit_button,
			    FALSE, FALSE, 0);
	status_bar = gtk_statusbar_new();
	gtk_container_set_border_width (GTK_CONTAINER(status_bar), 0);
	gtk_statusbar_set_has_resize_grip (GTK_STATUSBAR(status_bar), FALSE);
	g_signal_connect_after (G_OBJECT (status_bar),
		"text-pushed", G_CALLBACK(status_pushed), NULL);
	gtk_box_pack_start (GTK_BOX (bottom_container),
			    status_bar,
			    TRUE, TRUE, 0);
	
	/* Fill Button Box with Sections */
	for (l = schema_data->sections; l != NULL; l = l->next) {
		SchemaSection *section = l->data;
		
		if (!section_exist(section->section_name, section_button_box)) {
			GtkWidget *box_hbox, *button, *button_label, *button_image;
			GdkPixbuf *icon = NULL;
			
			button = gtk_button_new();
			gtk_button_set_relief (GTK_BUTTON (button), GTK_RELIEF_NONE);
			gtk_button_set_focus_on_click (GTK_BUTTON (button), FALSE);
			
			box_hbox = gtk_hbox_new (FALSE, 0);
			gtk_container_add (GTK_CONTAINER (button), box_hbox);
			icon = get_pixbuf_from_icon_list (section->icon_list);
			button_image = gtk_image_new_from_pixbuf (icon);
			gtk_box_pack_start (GTK_BOX (box_hbox),
					button_image,
					FALSE, FALSE, 0);
			
			button_label = gtk_label_new(section->section_name);
			gtk_label_set_use_markup (GTK_LABEL (button_label), FALSE);
			labels = g_list_append (labels, button_label);
			gtk_box_pack_start (GTK_BOX (box_hbox),
					button_label,
					FALSE, FALSE, 5);
			
			SectionPressClosure *closure = g_new0 (SectionPressClosure, 1);
			closure->section = section;
			closure->notebook = setting_notebook;
			closure->title = section_label;
			closure->section_pixbuf = icon;
			closure->label = button_label;
			closure->title_description = section_description;
			closure->title_image = section_image;
			closure->labels = labels;
			closure->title_box = title_hbox;
			closure->box_image = button_image;
			
			g_signal_connect ((gpointer) button, "clicked",
				G_CALLBACK (section_button_clicked), closure);
			
			/* activate the first section */
			if (l->prev == NULL) {
				gtk_image_set_from_pixbuf (GTK_IMAGE (section_image), icon);
				
				activate_section (section,
						  setting_notebook,
						  section_label,
						  section_description,
						  button_label);
			}
				
			
			gtk_container_add (GTK_CONTAINER (section_button_box), button);
		}
	}
	
	/* grab focus on the notebook, since we dont wanna have focus on the buttonbox */
	gtk_widget_grab_focus (setting_notebook);
	
	/* fill main container */
	gtk_box_pack_start (GTK_BOX (main_container),
			    main_frame,
			    TRUE, TRUE, 0);
	gtk_box_pack_start (GTK_BOX (main_container),
			    bottom_container,
			    FALSE, FALSE, 0);
	
	gtk_widget_show_all (window);
}
