/*
 *  Copyright (C) 2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include "kiba-win.h"
#include "kiba-dock.h"
#include "dnd-window.h"

static GtkWidget *dnd_window = NULL;

static gboolean drag_window_expose_event_cb (GtkWidget      *widget,
					     GdkEventExpose *expose,
					     gpointer        data) {
	KibaObject *object = KIBA_OBJECT(data);
	GtkWidget *dnd_widget = GTK_WIDGET(object);
	cairo_t *cr;
	GtkWidget *child;
	cairo_surface_t *surface = NULL;
	float factor_x = 1.2f, factor_y = 1.2f;
	const KibaLayoutData *d;
	gint window_max_width, window_max_height;
	GList *l;
	
	d = &kiba_layout_data[KIBA_DOCK(dnd_widget->parent)->d.position];
	
	for (l = object->custom_scale_funcs; l != NULL; l = l->next) {
		KibaObjectCustomScaleData *data = l->data;
		
		if (data->scale_widget)
			continue;
		
		float custom_factor_x = 1.0f;
		float custom_factor_y = 1.0f;
		
		data->scale_func (object,
			&custom_factor_x,
			&custom_factor_y,
			data->data);
			
		factor_x *= custom_factor_x;
		factor_y *= custom_factor_y;
	}
	
	window_max_width  = factor_x*dnd_widget->allocation.width;
	window_max_height = factor_y*dnd_widget->allocation.height;
	
	cr = gdk_cairo_create (expose->window);
	if (!cr) return TRUE;
		cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
		cairo_set_source_rgba (cr, 0, 0, 0, 0);
		cairo_paint (cr);
		
		cairo_set_operator (cr, CAIRO_OPERATOR_OVER);
		cairo_translate (cr,
			window_max_width/2 - dnd_widget->allocation.width/2,
			window_max_height/2 - dnd_widget->allocation.height/2);
		if (kiba_object_scale (object, cr)) {
			cairo_save (cr);
				kiba_object_paint_shadow (object, cr);
			cairo_restore (cr);
			cairo_save(cr);
				if (!object->paint)
					kiba_object_paint (object, cr, NULL);
				else
					object->paint (object, cr, object->paint_data);
			cairo_restore (cr);
		}
	cairo_destroy (cr);
	
	child = gtk_bin_get_child(GTK_BIN(widget));
	
	if (child)
		gtk_container_propagate_expose (GTK_CONTAINER (widget), child,  expose);
	
	return TRUE;
}

// void kiba_dnd_window_destroy_get_surface_offset

void kiba_dnd_window_destroy (GtkWidget *window) {
	KibaObject *object = g_object_get_data (G_OBJECT(window), "kiba-object");
	gtk_widget_destroy(window);
	dnd_window = NULL;
}

GtkWidget* kiba_dnd_window_new (KibaObject *object) {
	GdkScreen *screen = gdk_screen_get_default ();
	GtkWidget *ret;
	gint height, width;
	float factor_x = 1.2f, factor_y = 1.2f;
	GList *l;
	
	g_return_val_if_fail (KIBA_IS_OBJECT(object), dnd_window);
	
	if (dnd_window) {
		return NULL;
	}
	
	ret = gtk_window_new (GTK_WINDOW_POPUP);
	gtk_window_set_decorated (GTK_WINDOW(ret), FALSE);
	gtk_window_set_resizable (GTK_WINDOW(ret), FALSE);
	gtk_window_set_skip_taskbar_hint (GTK_WINDOW(ret), TRUE);
	gtk_window_set_type_hint (GTK_WINDOW(ret), GDK_WINDOW_TYPE_HINT_DND);
	
	g_object_set_data (G_OBJECT(ret), "kiba-object", object);
	gtk_widget_set_colormap (ret, gdk_screen_get_rgba_colormap (screen));
	
	width = GTK_WIDGET(object)->allocation.width;
	height = GTK_WIDGET(object)->allocation.height;
	
	for (l = object->custom_scale_funcs; l != NULL; l = l->next) {
		KibaObjectCustomScaleData *data = l->data;
		
		if (data->scale_widget)
			continue;
		
		float custom_factor_x = 1.0f;
		float custom_factor_y = 1.0f;
		
		data->max_func (object,
			&custom_factor_x,
			&custom_factor_y,
			data->data);
			
		factor_x *= custom_factor_x;
		factor_y *= custom_factor_y;
	}
	
	width *= factor_x;
	height *= factor_y;
	
	gtk_widget_set_usize (ret,
		width,
		height);
	
	g_signal_connect (G_OBJECT(ret), "expose-event",
		G_CALLBACK(drag_window_expose_event_cb), object);
	
	gtk_window_present (GTK_WINDOW(ret));
	
	dnd_window = ret;
	
	return ret;
}
