/*
 *  Copyright (C) 2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

/**
 * SECTION:kiba-fixed-object
 * @short_description: A fixed #KibaObject.
 * @see_also:#KibaWin,#KibaDock
 * @stability: Unstable
 *
 * A fixed #KibaObject.
 */

#include "kiba-fixed-object.h"
#include "kiba-types.h"
#include "kiba-dock.h"

G_DEFINE_TYPE (KibaFixedObject, kiba_fixed_object, KIBA_TYPE_OBJECT)

/**
 * kiba_fixed_object_set_size:
 * @fixed_object: The KibaFixedObject instance.
 * @size: The size of the object.
 *
 * Sets the size of the fixed object.
 */
void kiba_fixed_object_set_size (KibaFixedObject *fixed_object,
				 gint             size) {
	KibaObject *object;
	float scale;
	
	g_return_if_fail (KIBA_IS_FIXED_OBJECT(fixed_object));
	
	object = KIBA_OBJECT(fixed_object);
	
	scale = (float) size/object->size.width;
	
	kiba_object_set_requisition_factor (object, scale, scale);
}

static void kiba_fixed_object_size_request (GtkWidget      *widget,
					    GtkRequisition *requisition) {
	KibaObject *object;
	GtkWidget *dock;
	const KibaLayoutData *d;
	GtkWidget *child;
	GList *l;
	float scale_x = 1.0f;
	float scale_y = 1.0f;
	
	g_return_if_fail (KIBA_IS_FIXED_OBJECT(widget));
	
	object = KIBA_OBJECT(widget);
	dock = widget->parent;
	
	if (!dock)
		d = &kiba_layout_data[KIBA_POSITION_BOTTOM_EDGE];
	else
		d = &kiba_layout_data[KIBA_DOCK(dock)->d.position];

	child = gtk_bin_get_child (GTK_BIN(widget));
	/* using the child allocation, if there is a child */
	if (child) {
		gtk_widget_size_request (child, requisition);
		return;
	}
	
	scale_x = object->scale_x;
	scale_y = object->scale_y;
	requisition->width  = object->size.width;
	requisition->height = object->size.height;
	
	for (l = object->custom_scale_funcs; l != NULL; l = l->next) {
		KibaObjectCustomScaleData *data = l->data;
		
		if (!data->scale_widget)
			continue;
		
		float custom_factor_x = 1.0f;
		float custom_factor_y = 1.0f;
		
		data->scale_func (object,
			&custom_factor_x,
			&custom_factor_y,
			data->data);
		
		scale_x *= custom_factor_x;
		scale_y *= custom_factor_y;
	}
	
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL)
		requisition->width  = (int) requisition->width*scale_x;
	else
		requisition->height = (int) requisition->height*scale_y;
}

static void kiba_fixed_object_class_init (KibaFixedObjectClass *class) {
	GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (class);
	
	widget_class->size_request = kiba_fixed_object_size_request;
}

static void kiba_fixed_object_init (KibaFixedObject *fixed_object) {
	KibaObject *object = KIBA_OBJECT(fixed_object);
	KibaDock *dock = kiba_dock_get();
	
//TODO changed callback
	object->size = (GtkRequisition) {dock->d.height, dock->d.height};
}

/**
 * kiba_fixed_object_new:
 *
 * Creates and returns a KibaFixedObject Instance.
 *
 * Returns: The KibaFixedObject object created.
 */
GtkWidget* kiba_fixed_object_new (const char *group, KibaObjectWindowState state) {
	GtkWidget *widget =  g_object_new (KIBA_TYPE_FIXED_OBJECT,
                                 "group", group, NULL);
	KIBA_OBJECT(widget)->win_state = state;
	gtk_container_add(GTK_CONTAINER(kiba_dock_get()), widget);
	return widget;	
}
