/*
 *  Copyright (C) 2006-2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

/**
 * SECTION:kiba-icon-view
 * @short_description: Icon grid.
 * @see_also:#KibaIconViewWin
 * @stability: Unstable
 *
 * Container realizing a icon grid.
 */

#include "../config.h"
#include <gtk/gtk.h>

#include "kiba-icon-view.h"
#include "kiba-cairo-utils.h"
#include "kiba-types.h"
#include "kiba-image.h"

G_DEFINE_TYPE (KibaIconView, kiba_icon_view, GTK_TYPE_BOX)

enum {
	SIG_CHILD_ACTIVATED = 0,
	SIG_CHILD_SELECTED,
	SIG_CHILD_UNSELECTED,
	SIG_KIBA_ICON_VIEW_LAST
};

static guint kiba_icon_view_signals[SIG_KIBA_ICON_VIEW_LAST] = { 0 };

static gboolean kiba_icon_view_expose_cb (GtkWidget      *widget,
				          GdkEventExpose *expose) {
	KibaIconView *icon_view = KIBA_ICON_VIEW(widget);
	cairo_t *cr = NULL;
	gint border_width;
	float radius;
	
	border_width = 5;
	radius = 3.0f;
	
	if (!GTK_WIDGET_REALIZED (widget))
		return FALSE;
	
	if (!GTK_IS_WIDGET(icon_view->selected_child) ||
	    !GTK_WIDGET_MAPPED (icon_view->selected_child) ||
	    icon_view->selected_image)
		return FALSE;
	
	cr = gdk_cairo_create (expose->window);
	if (cairo_status (cr) != CAIRO_STATUS_SUCCESS)
		return FALSE;
	
	cairo_rectangle (cr,
		expose->area.x, expose->area.y,
		expose->area.width, expose->area.height);
	cairo_clip (cr);
	
	cairo_set_operator (cr, CAIRO_OPERATOR_OVER);
	cairo_translate (cr,
		icon_view->selected_child->allocation.x,
		icon_view->selected_child->allocation.y);
	cairo_paint_shape (cr,
		KIBA_ROUNDED, KIBA_ROUNDED, KIBA_ROUNDED, KIBA_ROUNDED,
		border_width/2,
		border_width/2,
		icon_view->selected_child->allocation.width - border_width/2,
		icon_view->selected_child->allocation.height - border_width/2,
		radius);
	cairo_set_source_rgba (cr, 0, 0, 0, 0.8);
	cairo_set_line_width (cr, border_width);
	cairo_stroke_preserve(cr);
	cairo_set_source_rgba (cr, 1, 1, 1, 0.6);
	cairo_fill(cr);
	
	cairo_destroy (cr);
	
	return FALSE;
}

static gboolean kiba_icon_view_expose_event (GtkWidget      *widget,
				             GdkEventExpose *expose) {
	GtkContainer *container = GTK_CONTAINER(widget);
	GList *l;
	gboolean expose_outside_view;
	
	expose_outside_view = FALSE;
	
	if (expose->area.x < widget->allocation.x) {
		expose->area.x = widget->allocation.x;
		expose_outside_view = TRUE;
	}
	if (expose->area.x + expose->area.width > widget->allocation.x + widget->allocation.width) {
		expose->area.height = widget->allocation.x + widget->allocation.width - expose->area.x;
		expose_outside_view = TRUE;
	}
	if (expose->area.y < widget->allocation.y) {
		expose->area.y = widget->allocation.y;
		expose_outside_view = TRUE;
	}
	if (expose->area.y + expose->area.height > widget->allocation.y + widget->allocation.height) {
		expose->area.height = widget->allocation.y + widget->allocation.height - expose->area.y;
		expose_outside_view = TRUE;
	}
	if (expose_outside_view) {
		gdk_region_destroy (expose->region);
		expose->region = gdk_region_rectangle (&expose->area);
	}
	
	for (l = gtk_container_get_children(container); l != NULL; l = l->next) {
		GtkWidget *child = l->data;
		gtk_container_propagate_expose (container, child,  expose);
	}
	
	return FALSE;
}

static KibaIconViewChild* kiba_icon_view_get_child_for_widget (KibaIconView *icon_view,
							       GtkWidget    *widget) {
	GList *l;
	
	for (l = icon_view->children; l != NULL; l = l->next) {
		KibaIconViewChild *child = l->data;
		
		if (child->widget == widget)
			return child;
	}
	
	return NULL;
}

static void select_container_images (KibaIconView  *icon_view,
				     GtkContainer  *container) {
	GList *l;
	
	for (l = gtk_container_get_children(container); l != NULL; l = l->next) {
		GtkWidget *child = l->data;
		
		if (KIBA_IS_IMAGE(child)) {
			kiba_image_select (KIBA_IMAGE(child), TRUE);
			icon_view->selected_image = TRUE;
		}
		else if (GTK_IS_CONTAINER(child)) {
			select_container_images (icon_view, GTK_CONTAINER(child));
		}
	}
}

static void mark_container_images (KibaIconView  *icon_view,
				   GtkContainer  *container) {
	GList *l;
	
	for (l = gtk_container_get_children(container); l != NULL; l = l->next) {
		GtkWidget *child = l->data;
		
		if (KIBA_IS_IMAGE(child)) {
			if (GTK_WIDGET(container) == icon_view->selected_child) {
				kiba_image_select (KIBA_IMAGE(child), TRUE);
				icon_view->selected_image = TRUE;
			}
			else if (kiba_image_is_selected(KIBA_IMAGE(child))) {
				kiba_image_select (KIBA_IMAGE(child), FALSE);
			}
		}
		else if (GTK_IS_CONTAINER(child)) {
			if (GTK_WIDGET(container) == icon_view->selected_child)
				select_container_images (icon_view, GTK_CONTAINER(child));
			else
				mark_container_images (icon_view, GTK_CONTAINER(child));
		}
	}
}

static gint icon_view_child_slide_timeout (gpointer data) {
	KibaIconView *icon_view = data;
	GtkWidget *widget = GTK_WIDGET(icon_view);
	gboolean sliding;
	GList *l;
	
	sliding = FALSE;
	for (l = icon_view->children; l != NULL; l = l->next) {
		KibaIconViewChild *child = l->data;
		GtkAllocation allocation;
		gint slide_step;
		
		allocation = child->widget->allocation;
		
		gint diff = allocation.y - child->allocation.y;
		diff = MAX(diff, -diff);
		if (icon_view->slide_step > 0)
			slide_step = icon_view->slide_step + diff/allocation.height - 1;
		else {
			slide_step = allocation.y - child->allocation.y;
			slide_step = MAX(slide_step, -slide_step);
		}
		
		if (allocation.y + allocation.height + slide_step <= widget->allocation.y ||
		    allocation.y - slide_step >= widget->allocation.height + widget->allocation.y) {
			if (GTK_WIDGET_MAPPED(child->widget)) {
				gtk_widget_hide (child->widget);
			}
		}
		else if (!GTK_WIDGET_MAPPED(child->widget)) {
			gtk_widget_show (child->widget);
		}
		
		if (allocation.y == child->allocation.y &&
		    allocation.x == child->allocation.x)
			continue;
		
		sliding = TRUE;
		
		if (allocation.y > child->allocation.y) {
			allocation.y -= slide_step;
			if (allocation.y < child->allocation.y)
				allocation.y = child->allocation.y;
		}
		if (allocation.x > child->allocation.x) {
			allocation.x -= slide_step;
			if (allocation.x < child->allocation.x)
				allocation.x = child->allocation.x;
		}
		if (allocation.y < child->allocation.y) {
			allocation.y += slide_step;
			if (allocation.y > child->allocation.y)
				allocation.y = child->allocation.y;
		}
		if (allocation.x < child->allocation.x) {
			allocation.x += slide_step;
			if (allocation.x > child->allocation.x)
				allocation.x = child->allocation.x;
		}
		
		gtk_widget_size_allocate (child->widget, &allocation);
	}
	
	if (!sliding) {
		icon_view->slide_id = 0;
		return FALSE;
	}
	
	return TRUE;
}

static void kiba_icon_view_children_size_allocate (GtkWidget *widget) {
	KibaIconView *icon_view = KIBA_ICON_VIEW(widget);
	GtkContainer *container;
	GtkAllocation obj_allocation;
	GList *l;
	gint i;
	gint columns;
	gint spacing;
	gint border_width;
	gint height;
	gboolean child_allocation_changed;
	
	border_width = gtk_container_get_border_width (GTK_CONTAINER(widget));
	spacing = gtk_box_get_spacing (GTK_BOX(widget));
	columns = kiba_icon_view_get_columns (KIBA_ICON_VIEW(widget));
	
	container = GTK_CONTAINER(widget);
	
	child_allocation_changed = FALSE;
	obj_allocation.x = spacing + border_width + widget->allocation.x;
	obj_allocation.y = spacing + border_width + widget->allocation.y;
	
	i = 1;
	height = 0;
	for (l = icon_view->children_before; l != NULL; l = l->next) {
		GtkWidget *child = l->data;
		GtkRequisition req;
		
		gtk_widget_size_request (child, &req);
		
		height = MAX(height, req.height + spacing);
		
		if (i == columns) {
			obj_allocation.y -= height;
			i = 0;
		}
		
		i++;
	}
	if (i > 1)
		obj_allocation.y -= height;
	
	i = 1;
	height = 0;
	for (l = icon_view->children; l != NULL; l = l->next) {
		KibaIconViewChild *view_child = l->data;
		GtkWidget *child = view_child->widget;
		GtkRequisition req;
		
		gtk_widget_size_request (child, &req);
		
		obj_allocation.width  = req.width;
		obj_allocation.height = req.height;
		
		height = MAX(height, req.height + spacing);
		
		view_child->allocation = obj_allocation;
		child->allocation.width = req.width;
		child->allocation.height = req.height;
		
		if (child->allocation.x == -1 &&
		    child->allocation.y == -1) {
			gtk_widget_size_allocate (child, &obj_allocation);
		}
		else if (view_child->allocation.x != child->allocation.x ||
		         view_child->allocation.y != child->allocation.y)
			child_allocation_changed = TRUE;
		
		if (i == columns) {
			obj_allocation.y += height;
			obj_allocation.x = spacing + border_width + widget->allocation.x;
			i = 0;
		}
		else
			obj_allocation.x += obj_allocation.width + spacing;
		
		i++;
	}
	
	if (!icon_view->slide_id && child_allocation_changed)
		icon_view->slide_id = g_timeout_add (40,
			icon_view_child_slide_timeout, icon_view);
}

/**
Properties
**/

/**
 * kiba_icon_view_get_columns:
 * @icon_view: The KibaIconView instance.
 *
 * Returns the number of columns.
 *
 * Returns: The number of columns.
 */
gint kiba_icon_view_get_columns (KibaIconView *icon_view) {
	g_return_val_if_fail (KIBA_IS_ICON_VIEW(icon_view), 0);
	
	return icon_view->columns;
}

/**
 * kiba_icon_view_set_columns:
 * @icon_view: The KibaIconView instance.
 * @columns: The number of columns.
 *
 * Sets the number of columns.
 */
void kiba_icon_view_set_columns (KibaIconView *icon_view,
				 gint          columns) {
	g_return_if_fail (KIBA_IS_ICON_VIEW(icon_view));
	g_return_if_fail (columns > 0);
	g_return_if_fail (columns < 20);
	
	icon_view->columns = columns;
	
	gtk_widget_queue_resize (GTK_WIDGET(icon_view));
}

/**
 * kiba_icon_view_activate_child:
 * @icon_view: The KibaIconView instance.
 * @child: The KibaIconView child to activate.
 *
 * Activates child.
 */
void kiba_icon_view_activate_child (KibaIconView *icon_view,
				    GtkWidget    *child) {
	GList *children;
	
	g_return_if_fail (KIBA_IS_ICON_VIEW(icon_view));
	
	children = gtk_container_get_children(GTK_CONTAINER(icon_view));
	g_return_if_fail (g_list_find(children, child) != NULL);
	
	if (icon_view->selected_child != child) {
		icon_view->selected_child = child;
		gtk_widget_queue_draw (child);
	}
	
	g_signal_emit (icon_view,
		kiba_icon_view_signals[SIG_CHILD_ACTIVATED],
		0, child);
}

/**
 * kiba_icon_view_select_child:
 * @icon_view: The KibaIconView instance.
 * @child: The KibaIconView child to select.
 *
 * Selects a icon view child.
 */
void kiba_icon_view_select_child (KibaIconView *icon_view,
				  GtkWidget    *child) {
	GList *children, *l;
	GtkWidget *old_selected;
	
	g_return_if_fail (KIBA_IS_ICON_VIEW(icon_view));
	
	children = gtk_container_get_children(GTK_CONTAINER(icon_view));
	g_return_if_fail (g_list_find(children, child) != NULL);
	
	if (icon_view->selected_child == child)
		return;
	
	old_selected = icon_view->selected_child;
	icon_view->selected_child = child;
	
	if (GTK_IS_WIDGET(old_selected)) {
		gtk_widget_queue_draw (old_selected);
		
		g_signal_emit (icon_view,
			kiba_icon_view_signals[SIG_CHILD_UNSELECTED],
			0, old_selected);
	}
	gtk_widget_queue_draw (child);
	
	icon_view->selected_image = FALSE;
	mark_container_images(icon_view, GTK_CONTAINER(icon_view));
	
	g_signal_emit (icon_view,
		kiba_icon_view_signals[SIG_CHILD_SELECTED],
		0, child);
}

/**
 * kiba_icon_view_get_selected_child:
 * @icon_view: The KibaIconView instance.
 *
 * Returns the selected child.
 *
 * Returns: The selcted KibaIconView child.
 */
GtkWidget* kiba_icon_view_get_selected_child (KibaIconView *icon_view) {
	g_return_val_if_fail (KIBA_IS_ICON_VIEW(icon_view), NULL);
	
	return icon_view->selected_child;
}

/**
 * kiba_icon_view_show_child:
 * @icon_view: The KibaIconView instance.
 * @child: The KibaIconView child.
 * @state: The show state.
 *
 * Shows/Hides child.
 */
void kiba_icon_view_show_child (KibaIconView          *icon_view,
				GtkWidget             *child,
				KibaIconViewShowState  state) {
	KibaIconViewChild *view_child;
	
	g_return_if_fail (KIBA_IS_ICON_VIEW(icon_view));
	g_return_if_fail (GTK_IS_WIDGET(child));
	
	view_child = kiba_icon_view_get_child_for_widget (icon_view, child);
	
	if (state == KIBA_ICON_VIEW_SHOW) {
		if (!g_list_find (icon_view->active_children, child)) {
			icon_view->active_children = g_list_append (icon_view->active_children, child);
			icon_view->active_icon_count++;
			gtk_widget_queue_resize (GTK_WIDGET(icon_view));
		}
		if (g_list_find (icon_view->children_before, child)) {
			icon_view->children_before = g_list_remove (icon_view->children_before, child);
			icon_view->icons_before_count--;
		}
		if (g_list_find (icon_view->children_after, child)) {
			icon_view->children_after = g_list_remove (icon_view->children_after, child);
			icon_view->icons_after_count--;
		}
	}
	else {
		if (state == KIBA_ICON_VIEW_HIDE_BEFORE &&
		   !g_list_find (icon_view->children_before, child)) {
			icon_view->children_before = g_list_append (icon_view->children_before, child);
			icon_view->icons_before_count++;
		}
		else if (state == KIBA_ICON_VIEW_HIDE_AFTER &&
		        !g_list_find (icon_view->children_after, child)) {
			icon_view->children_after = g_list_insert(icon_view->children_after, child, 0);
			icon_view->icons_after_count++;
		}
		if (g_list_find (icon_view->active_children, child)) {
			icon_view->active_children = g_list_remove (icon_view->active_children, child);
			icon_view->active_icon_count--;
			gtk_widget_queue_resize (GTK_WIDGET(icon_view));
		}
	}
}

/**
 * kiba_icon_view_set_prelight_ontop:
 * @icon_view: The KibaIconView instance.
 * @ontop: TRUE if the prelight should be ontop, otherwise FALSE.
 *
 * Sets wether the prelight is ontop.
 */
void kiba_icon_view_set_prelight_ontop (KibaIconView *icon_view,
					gboolean      ontop) {
	GList *children;
	
	g_return_if_fail (KIBA_IS_ICON_VIEW(icon_view));
	
	if (icon_view->prelight_ontop == ontop)
		return;
	
	icon_view->prelight_ontop = ontop;
	
	if (!icon_view->prelightable)
		return;
	
	g_signal_handlers_disconnect_by_func (G_OBJECT(icon_view),
		G_CALLBACK(kiba_icon_view_expose_cb), NULL);
	
	if (icon_view->prelight_ontop) {
		g_signal_connect_after(G_OBJECT (icon_view), "expose-event",
			G_CALLBACK(kiba_icon_view_expose_cb), NULL);
	}
}

/**
 * kiba_icon_view_set_prelightable:
 * @icon_view: The KibaIconView instance.
 * @prelightable: TRUE for prelightable children, otherwise FALSE.
 *
 * Sets wether the icon view children are prelightable.
 */
void kiba_icon_view_set_prelightable (KibaIconView *icon_view,
				      gboolean      prelightable) {
	GList *children;
	
	g_return_if_fail (KIBA_IS_ICON_VIEW(icon_view));
	
	if (icon_view->prelightable == prelightable)
		return;
	
	icon_view->prelightable = prelightable;
	
	if (!prelightable) {
		g_signal_handlers_disconnect_by_func (G_OBJECT(icon_view),
			G_CALLBACK(kiba_icon_view_expose_cb), NULL);
	}
	else {
		if (icon_view->prelight_ontop) {
			g_signal_connect_after(G_OBJECT (icon_view), "expose-event",
				G_CALLBACK(kiba_icon_view_expose_cb), NULL);
		}
	}
}

/**
Class Functions
**/

static void kiba_icon_view_add (GtkContainer *container,
				GtkWidget    *widget) {
	KibaIconView *icon_view = KIBA_ICON_VIEW(container);
	KibaIconViewChild *child;
	
	child = g_new0 (KibaIconViewChild, 1);
	child->allocation = widget->allocation;
	child->widget = gtk_widget_ref(widget);
	
	icon_view->children = g_list_append (icon_view->children, child);
	
	(* GTK_CONTAINER_CLASS (kiba_icon_view_parent_class)->add) (container, widget);
}

static void kiba_icon_view_remove (GtkContainer *container,
				   GtkWidget    *widget) {
	KibaIconView *icon_view = KIBA_ICON_VIEW(container);
	KibaIconViewChild *child;
	
	child = kiba_icon_view_get_child_for_widget (icon_view, widget);
	if (child) {
		icon_view->children = g_list_remove (icon_view->children, child);
		gtk_widget_unref(child->widget);
		g_free (child);
	}
	
	if (g_list_find (icon_view->active_children, widget)) {
		GtkWidget *new_active = NULL;
		GList *l;
		
		icon_view->active_children = g_list_remove (icon_view->active_children, widget);
		icon_view->active_icon_count--;
		
		if (g_list_length(icon_view->children_before) > 0) {
			new_active = g_list_last(icon_view->children_before)->data;
			
			icon_view->children_before = g_list_remove (
				icon_view->children_before, new_active);
			icon_view->icons_before_count--;
			
			icon_view->active_children = g_list_prepend (
				icon_view->active_children, new_active);
			icon_view->active_icon_count++;
		}
		else if (icon_view->icons_after_count > 0) {
			new_active = icon_view->children_after->data;
			
			icon_view->children_after = g_list_remove (
				icon_view->children_after, new_active);
			icon_view->icons_after_count--;
			
			icon_view->active_children = g_list_append (
				icon_view->active_children, new_active);
			icon_view->active_icon_count++;
		}
		
		if (new_active) {
			gtk_widget_show (new_active);
			kiba_icon_view_show_child (icon_view, new_active, KIBA_ICON_VIEW_SHOW);
		}
		
		for (l = gtk_container_get_children(GTK_CONTAINER(icon_view)); l != NULL; l = l->next) {
			GtkWidget *child = l->data;
			child->allocation.x = -1;
			child->allocation.y = -1;
		}
		kiba_icon_view_children_size_allocate (GTK_WIDGET(icon_view));
	}
	else if (g_list_find (icon_view->children_after, widget)) {
		icon_view->children_after = g_list_remove (icon_view->children_after, widget);
		icon_view->icons_after_count--;
	}
	else if (g_list_find (icon_view->children_before, widget)) {
		icon_view->children_before = g_list_remove (icon_view->children_before, widget);
		icon_view->icons_before_count--;
	}
	
	(* GTK_CONTAINER_CLASS (kiba_icon_view_parent_class)->remove) (container, widget);
}

static void kiba_icon_view_size_allocate (GtkWidget     *widget,
					  GtkAllocation *allocation) {
	widget->allocation = *allocation;
	
	kiba_icon_view_children_size_allocate (widget);
}

static void kiba_icon_view_size_request (GtkWidget      *widget,
					 GtkRequisition *requisition) {
	GList *l;
	gint columns;
	gint i;
	gint spacing;
	gint border_width;
	gint row_height, row_width;
	
	border_width = gtk_container_get_border_width (GTK_CONTAINER(widget));
	spacing = gtk_box_get_spacing (GTK_BOX(widget));
	columns = kiba_icon_view_get_columns (KIBA_ICON_VIEW(widget));
	
	requisition->width  = row_height = 0;
	requisition->height = row_width = spacing;
	
	i = 1;
	for (l = GTK_BOX(widget)->children; l != NULL; l = l->next) {
		GtkBoxChild   *child = l->data;
		GtkRequisition child_requisition;
		
		if (!g_list_find(KIBA_ICON_VIEW(widget)->active_children, child->widget))
			continue;
		
		gtk_widget_size_request (child->widget, &child_requisition);
		
		row_height = MAX(row_height, child_requisition.height + spacing);
		row_width += child_requisition.width + spacing;
		
		if (i == columns) {
			requisition->height += row_height;
			requisition->width   = MAX(row_width, requisition->width);
			
			i = 0;
			row_width = spacing;
			row_height = 0;
		}
		else if (KIBA_ICON_VIEW(widget)->fixed && !l->next) {
			row_width += (columns-i)*(child_requisition.width + spacing);
			requisition->width   = MAX(row_width, requisition->width);
		}
		
		i++;
	}
	requisition->width   = MAX(row_width, requisition->width);
	requisition->height += row_height;
	
	requisition->width  += 2*border_width;
	requisition->height += 2*border_width;
}

static void kiba_icon_view_class_init (KibaIconViewClass *class) {
	GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (class);
	GtkContainerClass *container_class = GTK_CONTAINER_CLASS (class);
	
	widget_class->size_allocate   = kiba_icon_view_size_allocate;
	widget_class->size_request    = kiba_icon_view_size_request;
	widget_class->expose_event    = kiba_icon_view_expose_event;
	
	container_class->add       = kiba_icon_view_add;
	container_class->remove    = kiba_icon_view_remove;
	
	kiba_icon_view_signals[SIG_CHILD_ACTIVATED] =
		g_signal_new ("child-activated",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaIconViewClass, child_activated),
			NULL, NULL,
			g_cclosure_marshal_VOID__OBJECT,
			G_TYPE_NONE, 1,
			G_TYPE_OBJECT);
	kiba_icon_view_signals[SIG_CHILD_SELECTED] =
		g_signal_new ("child-selected",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaIconViewClass, child_selected),
			NULL, NULL,
			g_cclosure_marshal_VOID__OBJECT,
			G_TYPE_NONE, 1,
			G_TYPE_OBJECT);
	kiba_icon_view_signals[SIG_CHILD_UNSELECTED] =
		g_signal_new ("child-unselected",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaIconViewClass, child_unselected),
			NULL, NULL,
			g_cclosure_marshal_VOID__OBJECT,
			G_TYPE_NONE, 1,
			G_TYPE_OBJECT);
}

static void kiba_icon_view_init (KibaIconView *icon_view) {
	GTK_WIDGET_SET_FLAGS (GTK_WIDGET (icon_view), GTK_NO_WINDOW);
	
	gtk_widget_add_events (GTK_WIDGET (icon_view),
		GDK_BUTTON_PRESS_MASK |
		GDK_BUTTON_RELEASE_MASK |
		GDK_ENTER_NOTIFY_MASK |
		GDK_LEAVE_NOTIFY_MASK |
		GDK_POINTER_MOTION_MASK |
		GDK_POINTER_MOTION_HINT_MASK);
	
	icon_view->columns = 4;
	icon_view->selected_child = NULL;
	icon_view->fixed = FALSE;
	icon_view->prelight_ontop = TRUE;
	icon_view->prelightable = TRUE;
	icon_view->slide_id = 0;
	icon_view->slide_step = 0;
	
	icon_view->children_before = NULL;
	icon_view->active_children = NULL;
	icon_view->children_after = NULL;
	icon_view->icons_before_count = 0;
	icon_view->active_icon_count = 0;
	icon_view->icons_after_count = 0;
	
	gtk_container_set_border_width (GTK_CONTAINER(icon_view), 0);
	gtk_box_set_spacing (GTK_BOX(icon_view), 10);
}

/**
 * kiba_icon_view_new:
 *
 * Creates and returns a KibaIconView instance.
 *
 * Returns: The KibaIconView instance created.
 */
GtkWidget* kiba_icon_view_new (void) {
	return g_object_new (KIBA_TYPE_ICON_VIEW, NULL);
}
