/*
 *  Copyright (C) 2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

/**
 * SECTION:kiba-icon
 * @short_description: Icon object on #KibaDock.
 * @see_also:#KibaWin,#KibaDock
 * @stability: Unstable
 *
 * Icon object on #KibaDock.
 */

#include "../config.h"

#include "kiba-win.h"
#include "kiba-icon.h"
#include "kiba-utils.h"
#include "kiba-macros.h"
#include "kiba-inline-pixbufs.h"
#include "kiba-svg-macro.h"
#ifdef USE_SVG
  #include <librsvg/rsvg.h>
#endif

#define FALLBACK_ICON "kiba-dock"

G_DEFINE_TYPE (KibaIcon, kiba_icon, KIBA_TYPE_OBJECT)

/**
Rendering
**/

static void kiba_icon_paint (KibaObject *object,
			     cairo_t    *cr,
			     gpointer    data) {
	cairo_surface_t *surface;
	
	g_return_if_fail (KIBA_IS_OBJECT (object));
	
	if (object->switch_surface)
		surface = object->switch_surface;
	else
		surface = object->surface_data->surface;
	g_return_if_fail (surface != NULL);
	
	cairo_set_source_surface (cr, surface, 0, 0);
	kiba_object_paint_with_prelight (object, surface, cr);
}

static void kiba_icon_surface (cairo_t          *cr,
			       cairo_surface_t  *surface,
			       gint              width,
			       gint              height,
			       void             *data) {
	KibaIcon *icon = data;
	gint size;
	GdkPixbuf *pixbuf = NULL;
#ifdef USE_SVG
	RsvgHandle *handle = NULL;
#endif
	
	size = MAX(width,height);
	
#ifdef USE_SVG
	if (kiba_load_svg (icon->file, &size, &handle)) {
		rsvg_handle_render_cairo (handle, cr);
		rsvg_handle_free (handle);
		return;
	}
#endif
	if (kiba_load_icon (icon->file, size, &pixbuf)) {
		gdk_cairo_set_source_pixbuf (cr, pixbuf, 0, 0);
		cairo_paint (cr);
		g_object_unref (pixbuf);
		return;
	}
	
	pixbuf = gdk_pixbuf_new_from_inline (-1, app_fallback_icon, FALSE, NULL);
	gdk_cairo_set_source_pixbuf (cr, pixbuf, 0, 0);
	cairo_paint (cr);
	g_object_unref (pixbuf);
	
	g_free (icon->file);
	icon->file = NULL;
}

static void kiba_icon_update_surface (KibaIcon         *icon,
				      CairoSurfaceFunc  func) {
	KibaWin *default_win;
	cairo_t *cr;
	
	KIBA_OBJECT(icon)->surface_data->width = KIBA_OBJECT(icon)->size.width;
	KIBA_OBJECT(icon)->surface_data->height = KIBA_OBJECT(icon)->size.height;
	
	default_win = kiba_win_get_default();
	cr = kiba_win_cairo_create (default_win);
	cairo_update_surface (cr, KIBA_OBJECT(icon)->surface_data, func, icon);
	cairo_destroy (cr);
}

/**
Properties
**/

/**
 * kiba_icon_set_file:
 * @icon: The KibaIcon instance.
 * @file: The file URI to set.
 *
 * Sets the icon file/name.
 *
 * Returns: TRUE if setting the file was succesfull, otherwise FALSE is returned.
 */
gboolean kiba_icon_set_file (KibaIcon   *icon,
                             const char *file) {
	g_return_val_if_fail (KIBA_IS_ICON (icon), FALSE);
	g_return_val_if_fail (file != NULL, FALSE);
	
	if (icon->file) {
		g_free (icon->file);
		icon->file = NULL;
	}
	icon->file = g_strdup(file);
	
	kiba_icon_update_surface (icon, kiba_icon_surface);
	kiba_object_queue_draw (KIBA_OBJECT(icon));
	
	return icon->file != NULL;
}

/**
 * kiba_icon_get_file:
 * @icon: The KibaIcon instance.
 *
 * Returns the icon name/file.
 *
 * Returns: The KibaIcon icon path or name.
 */
const char* kiba_icon_get_file (KibaIcon  *icon) {
	g_return_val_if_fail (KIBA_IS_ICON (icon), NULL);
	
	return icon->file;
}

/**
Class Functions
**/

static void kiba_icon_unrealize (GtkWidget *widget) {
	if (KIBA_ICON(widget)->file)
		g_free (KIBA_ICON(widget)->file);
	
	(* GTK_WIDGET_CLASS (kiba_icon_parent_class)->unrealize) (widget);
}

static void kiba_icon_realize (GtkWidget *widget) {
	KibaObject *object;
	
	object = KIBA_OBJECT(widget);
	
	(* GTK_WIDGET_CLASS (kiba_icon_parent_class)->realize) (widget);
	
	object->surface_data->width  = object->size.width;
	object->surface_data->height = object->size.height;
	
	if (KIBA_ICON(widget)->file) {
		kiba_icon_update_surface (KIBA_ICON(object), kiba_icon_surface);
	}
}

static void kiba_icon_class_init (KibaIconClass *class) {
	GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (class);
	
	widget_class->realize   = kiba_icon_realize;
	widget_class->unrealize = kiba_icon_unrealize;
}

static void kiba_icon_init (KibaIcon *icon) {
	kiba_object_set_paint_function (KIBA_OBJECT(icon), kiba_icon_paint, NULL);
	
	KIBA_OBJECT(icon)->win_state = KIBA_OBJECT_WINDOW_INVISIBLE;
	
	SET_FLAG (KIBA_OBJECT(icon)->props, KIBA_OBJECT_VISIBLE);
	SET_FLAG (KIBA_OBJECT(icon)->props, KIBA_OBJECT_MOVEABLE);
	SET_FLAG (KIBA_OBJECT(icon)->props, KIBA_OBJECT_PRELIGHTABLE);
	SET_FLAG (KIBA_OBJECT(icon)->props, KIBA_OBJECT_ZOOMABLE);
	SET_FLAG (KIBA_OBJECT(icon)->props, KIBA_OBJECT_REFLECTABLE);
}

/**
 * kiba_icon_new:
 *
 * Creates and returns a KibaIcon Instance.
 *
 * Returns: The KibaIcon Instance created.
 */
GtkWidget* kiba_icon_new (const char *group) {
	GtkWidget *widget =  g_object_new (KIBA_TYPE_ICON,
                                 "group", group, NULL);
	KIBA_ICON(widget)->file = NULL;
	gtk_container_add(GTK_CONTAINER(kiba_dock_get()), widget);
	return widget;	
}

/**
 * kiba_icon_new_from_file:
 * @file: The file URI to use.
 *
 * Creates and returns a KibaIcon from file.
 *
 * Returns: The KibaIcon Instance created.
 */
GtkWidget* kiba_icon_new_from_file (const char *group, const char *file) {
	GtkWidget *icon;
	
	icon = kiba_icon_new (group);
	KIBA_ICON(icon)->file = g_strdup (file);
	
	return icon;
}
