/*
 *  Copyright (C) 2006-2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

/**
 * SECTION:kiba-info-win
 * @short_description: Transparent toplevel window.
 * @see_also:
 * @stability: Unstable
 *
 * Transparent toplevel window.
 */

#include "../config.h"
#include <gtk/gtk.h>
#include <gdk/gdk.h>

#include "kiba-screen.h"
#include "kiba-debug.h"
#include "kiba-macros.h"
#include "kiba-prefs.h"
#include "kiba-cairo-utils.h"
#include "kiba-info-win.h"
#include "kiba-rc.h"

G_DEFINE_TYPE (KibaInfoWin, kiba_info_win, GTK_TYPE_WINDOW)

static void __gradient_changed (const char *option_id,
				  gpointer    value,
				  gpointer    data) {
	KibaInfoWin *info_win = data;
	KibaRC *rc = kiba_rc_get();
	
	info_win->gradient = rc->popup.gradient;
	info_win->numColors = rc->popup.numColors;
	gtk_widget_queue_draw(GTK_WIDGET(info_win));
}

/**
Class Functions
**/

static gboolean kiba_info_win_expose_event (GtkWidget      *widget,
				            GdkEventExpose *expose) {
	KibaInfoWin *info_win = KIBA_INFO_WIN(widget);
	cairo_t *cr = NULL;
	cairo_pattern_t *pattern = NULL;
	GtkWidget *child = NULL;
	gint border_width;
	KibaRC *rc;
	
	rc = kiba_rc_get();
	
	border_width = gtk_container_get_border_width(GTK_CONTAINER(widget));
	
	child = gtk_bin_get_child (GTK_BIN (widget));
	
	if (!GTK_WIDGET_REALIZED (widget))
		return FALSE;
	
	cr = gdk_cairo_create (widget->window);
	if (cairo_status (cr) != CAIRO_STATUS_SUCCESS)
		return FALSE;
	
	cairo_rectangle (cr,
		expose->area.x, expose->area.y,
		expose->area.width, expose->area.height);
	cairo_clip (cr);
	
	/* clear the window to be transparent */
	cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
	cairo_set_source_rgba (cr, 0, 0, 0, 0);
	cairo_paint (cr);
	
	if (rc->popup.draw) {
		cairo_set_operator (cr, CAIRO_OPERATOR_OVER);
		
		/* draw shape for the info window */
		cairo_paint_shape (cr,
			rc->popup.topleft,    rc->popup.topright,
			rc->popup.bottomleft, rc->popup.bottomright,
			border_width/2,
			border_width/2,
			widget->allocation.width - border_width/2,
			widget->allocation.height - border_width/2,
			rc->popup.corner_radius);
		
		if (rc->popup.background.style == KIBA_GRADIENT)
			pattern = cairo_get_gradient_pattern (border_width/2,
				border_width/2,
				widget->allocation.width - border_width,
				widget->allocation.height - border_width,
				info_win->gradient, info_win->numColors);
		else {
			pattern = cairo_get_png_pattern (rc->popup.background.file, 1.0);
		}
		
		cairo_set_source (cr, pattern);
		cairo_fill_preserve (cr);
		cairo_pattern_destroy (pattern);
		
		cairo_set_source_rgba (cr,
			info_win->border_color.r,
			info_win->border_color.g,
			info_win->border_color.b,
			info_win->border_color.a);
		cairo_set_line_width (cr, border_width);
		cairo_stroke (cr);
	}
	
	cairo_destroy (cr);
	
	if (child)
		gtk_container_propagate_expose (GTK_CONTAINER (widget), child,  expose);
	
	return FALSE;
}

static void kiba_info_win_screen_changed (GtkWidget *widget,
					  GdkScreen *previous_screen) {
	GdkScreen *screen;
	GdkColormap *colormap = NULL;
	GdkVisual *visual = NULL;
	
	screen = gtk_widget_get_screen (widget);
	
	colormap = gdk_screen_get_rgba_colormap (screen);
	visual   = gdk_screen_get_rgba_visual (screen);
	if (!colormap || !visual) {
		colormap = gdk_screen_get_rgb_colormap (screen);
		visual   = gdk_screen_get_rgb_visual (screen);
	}
	
	gtk_widget_set_visual (widget, visual);
	gtk_widget_set_colormap (widget, colormap);
}

static void kiba_info_win_class_init (KibaInfoWinClass *class) {
	GtkWidgetClass    *widget_class;
	
	widget_class    = GTK_WIDGET_CLASS (class);
	
	widget_class->expose_event    = kiba_info_win_expose_event;
	widget_class->screen_changed  = kiba_info_win_screen_changed;
}

static void __destroy_cb(GtkWidget *widget,
                         gpointer data) {
	kiba_prefs_remove_callback_by_data (widget);
}

static void kiba_info_win_init (KibaInfoWin *info_win) {
	KibaScreen *screen;
	KibaRC *rc;
	
	GTK_WIDGET_UNSET_FLAGS (GTK_WIDGET (info_win), GTK_NO_WINDOW);
	
	screen = kiba_screen_get();
	rc = kiba_rc_get();
	
	gtk_widget_add_events (GTK_WIDGET (info_win),
		GDK_EXPOSURE_MASK |
		GDK_FOCUS_CHANGE_MASK |
		GDK_BUTTON_PRESS_MASK |
		GDK_BUTTON_RELEASE_MASK |
		GDK_ENTER_NOTIFY_MASK |
		GDK_LEAVE_NOTIFY_MASK |
		GDK_POINTER_MOTION_MASK |
		GDK_POINTER_MOTION_HINT_MASK);
	
	info_win->border_color = rc->popup.border_color;
	
	kiba_info_win_screen_changed (GTK_WIDGET (info_win), NULL);
	gtk_widget_set_app_paintable (GTK_WIDGET (info_win), TRUE);
	gtk_window_set_decorated (GTK_WINDOW (info_win), FALSE);
	gtk_container_set_border_width (GTK_CONTAINER(info_win), rc->popup.border_width);
	
	g_signal_connect(G_OBJECT(info_win), "destroy",
		G_CALLBACK(__destroy_cb), NULL);
	
	kiba_prefs_add_callback ("rc_popup_gradient",
		__gradient_changed, info_win, TRUE);
}

/**
 * kiba_info_win_new:
 *
 * Creates and returns a KibaInfoWin Instance.
 *
 * Returns: The KibaInfoWin Instance created.
 */
GtkWidget* kiba_info_win_new (void) {
	return g_object_new (KIBA_TYPE_INFO_WIN, NULL);
}
