/*
 *  Copyright (C) 2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include "../config.h"
#include <stdlib.h>
#include <string.h>
#include <locale.h>
#include <gtk/gtk.h>

#include <libxml/parser.h>

#include "kiba-win.h"
#include "kiba-dock.h"
#include "dock-menu.h"
#include "kiba-separator.h"
#include "separator-menu.h"
#include "kiba-plugin-loader.h"
#include "prefs-dbus.h"
#include "kiba-macros.h"
#include "kiba-screen.h"
#include "kiba-debug.h"
#include "kiba-prefs.h"
#include "kiba-prefs-schema.h"
#include "kiba-globals.h"
#include "kiba-surface-factory.h"
#include "prefs-callbacks.h"

#include "surface-factory-private.h"

#ifdef USE_KDE
	const char *file_manager_global = "konqueror";
#else
#ifdef USE_XFCE
	const char *file_manager_global = "thunar";
#else
	const char *file_manager_global = "nautilus --no-desktop";
#endif
#endif

static void kiba_add_object_callbacks (GtkContainer *container,
				       GtkWidget    *widget) {
	if (!GTK_WIDGET_REALIZED(widget))
		gtk_widget_realize(widget);
	
	/* catch press signal to create context menus for separators */
	g_signal_connect (G_OBJECT(widget), "button-press-event",
		G_CALLBACK(kiba_separator_show_menu), NULL);
}

static GtkWidget* create_kiba_widget () {
	GtkWidget *window, *dock;
	GSList *dock_groups, *l;
	gint i;
	
	/* initiall the window */
	window = kiba_win_new ();
	DEBUG_INFO("Kiba Window created\n");
	
	/* initiall dock */
	dock = kiba_dock_new ();
	gtk_container_add (GTK_CONTAINER (window), dock);
	DEBUG_INFO("Kiba Dock created\n");
	
	g_signal_connect(G_OBJECT(dock), "add",
		G_CALLBACK(kiba_add_object_callbacks), NULL);
	
	dock_groups = kiba_prefs_get_list ("dock_groups");
	kiba_dock_set_group_list (KIBA_DOCK(dock), dock_groups);
	
	gtk_widget_show_all (window);
	
	kiba_prefs_callbacks_add_dock (KIBA_WIN(window), KIBA_DOCK(dock));
	
	/* adding separator objects */
	i = 1;
	for (l = dock_groups; l != NULL; l = l->next) {
		GtkWidget *object;
		
		if (!g_str_has_prefix ((char*) l->data, "separator"))
			continue;
		
		object = kiba_separator_new ((char*) l->data);
		gtk_widget_show (object);
	}
	
	return window;
}

static void kiba_screen_changed_cb (KibaScreen *screen,
				    GdkScreen  *new_screen,
				    gpointer    data) {
	GtkWidget *widget = data;
	
	gtk_widget_queue_resize (widget);
}

static void print_info_debug_messages (void) {
#ifdef HAVE_SVG
	DEBUG_INFO("Kiba was build with SVG support\n");
#else
	DEBUG_INFO("Kiba was build without SVG support\n");
#endif
#ifdef USE_GLITZ
	DEBUG_INFO("Kiba was build with GLITZ support\n");
#else
	DEBUG_INFO("Kiba was build without GLITZ support\n");
	#ifdef HAVE_GLITZ
	#ifndef CAIRO_HAS_GLITZ_SURFACE
		DEBUG_MISC("Your current cairo installation does not have glitz Support.\n");
		DEBUG_MISC("Cairo does not include glitz support by default.\n"
			"To use glitz in kiba you must recompile cairo with '--enable-glitz'.\n");
	#endif
	#endif
#endif
#ifdef USE_SDL
	DEBUG_INFO("Kiba was build with SDL support\n");
#else
	DEBUG_INFO("Kiba was build without SDL support\n");
#endif
}

int main (int argc, char *argv[]) {
	GtkWidget *widget;
	GtkWidget *dock;
	KibaPluginLoader *loader;
	char *kiba_user_dir;
	char *offset, *groups;
	char *plugin_path = NULL;
	char *config_file = NULL;
	GList *l;
	gint i;
	
	gtk_disable_setlocale();
	gtk_init (&argc, &argv);
	
#ifdef ENABLE_NLS
	bindtextdomain (PACKAGE, DATADIR"/locale");
	textdomain (PACKAGE);
	bind_textdomain_codeset (PACKAGE, "UTF-8");
#endif
	
	setlocale (LC_NUMERIC, "C");  /* use decimal point instead of comma */
	setlocale (LC_CTYPE, "");
	setlocale (LC_MESSAGES, "");
	
	/* parsing command line parameters */
	for (i = 0; i < argc; i++) {
		if (g_strrstr (argv[i], "--help") != NULL ||
		    g_strrstr (argv[i], "-h") != NULL) {
			fprintf (stderr, "Usage: kiba-dock [--verbose] [--config-file file_path] [--plugin-path plugin_path] ...\n");
			return 0;
		}
		/* enables debugging */
		else if (g_strrstr (argv[i], "--verbose") != NULL ||
			 g_strrstr (argv[i], "-v") != NULL)
			kiba_debug_init ();
		else if (i < argc + 1) {
			if (g_strrstr (argv[i], "--config-file") != NULL) {
				config_file = g_strdup (argv[i+1]);
				if (!g_file_test (config_file, G_FILE_TEST_EXISTS)) {
					DEBUG_ERROR ("Invalid Directory in --config-file argument: '%s'\n",
						config_file);
					g_free (config_file);
					config_file = NULL;
				}
			}
			else if (g_strrstr (argv[i], "--plugin-path") != NULL) {
				plugin_path = argv[i+1];
				if (!g_file_test (plugin_path, G_FILE_TEST_EXISTS)) {
					DEBUG_ERROR ("Invalid Directory in --plugin-path argument: '%s'\n",
						plugin_path);
					plugin_path = NULL;
				}
			}
		}
	}
	
	print_info_debug_messages ();
	
	/* create kiba home directory */
	kiba_user_dir = g_strdup_printf ("%s/.kiba-dock", g_get_home_dir ());
	if (!g_file_test (kiba_user_dir, G_FILE_TEST_IS_DIR)) {
		if (g_mkdir (kiba_user_dir, 7664) != 0)
			DEBUG_ERROR ("Failed to create user directory at '%s'\n", kiba_user_dir);
	}
	
#ifdef HAVE_SVG
	DEBUG_MISC("Initiall RSVG Library\n");
	rsvg_init ();
#endif
	
	/* initiall libxml */
	LIBXML_TEST_VERSION
	
	/* initiall preferences */
	if (!config_file) {
		config_file = g_strdup_printf ("%s/config.xml", kiba_user_dir);
	}
	g_free (kiba_user_dir);
	kiba_prefs_init (config_file);
	g_free (config_file);
	kiba_prefs_update_config_file(kiba_prefs_get());
	
	kiba_prefs_callbacks_add ();
	
	prefs_dbus_get();
	
	widget = create_kiba_widget ();
	dock = KIBA_WIN_CHILD(widget);
	gdk_window_focus (widget->window, TRUE);
	
	g_signal_connect (KIBA_SCREEN_DEFAULT, "screen-changed",
		G_CALLBACK(kiba_screen_changed_cb), widget);
	
	/* loading plugins */
	if (!plugin_path) {
		plugin_path = LIBDIR "/kiba-dock";
	}
	DEBUG_MISC("Loading Plugins from '%s'...\n", plugin_path);
	if (!g_file_test (plugin_path, G_FILE_TEST_IS_DIR)) {
		DEBUG_ERROR (
				"Failed to locate Plugins at '%s'\n"
				"Please install the Plugins at '" LIBDIR "/kiba-dock' "
				"or use the '--plugin-path' command line parameter.\n",
				plugin_path);
		exit(0);
	}
	
	loader = kiba_plugin_loader_get();
	if (!kiba_plugin_loader_initiall_plugins_at_path (loader, plugin_path)) {
		exit(0);
	}
	DEBUG_MISC("Plugins loaded\n\n");
	
	for (l = GTK_BOX(dock)->children; l != NULL; l = l->next) {
		GtkBoxChild *child = l->data;
		
		/* avoids scaling of objects */
		KIBA_OBJECT(child->widget)->destroy_factor = 1.0;
	}
	
	g_signal_connect (G_OBJECT(dock), "button-press-event",
		G_CALLBACK(kiba_dock_show_menu), NULL);
	
	/* show tips dialog */
// 	kiba_dialog_tips ();
	
	gtk_main ();
	
	/* save some settings wich may changed during user input */
	kiba_prefs_write_to_config_file ("edge_position",
		"char",
		kiba_layout_data[KIBA_DOCK(dock)->d.position].name);
	
	offset = g_strdup_printf ("%i", KIBA_DOCK(dock)->d.offset);
	kiba_prefs_write_to_config_file ("dock_edge_offset", "int", offset);
	g_free (offset);
	
	gint separator_groups = 0;
	for (l = GTK_BOX(dock)->children; l != NULL; l = l->next) {
		GtkBoxChild *child = l->data;
		
		if (KIBA_IS_SEPARATOR(child->widget))
			separator_groups++;
	}
	
	i = 1;
	for (l = KIBA_DOCK(dock)->groups; l != NULL; l = l->next) {
		char *group = l->data;
		
		if (!g_str_has_prefix(group, "separator"))
			continue;
		
		if (i > separator_groups) {
			l = l->prev;
			KIBA_DOCK(dock)->groups = g_list_remove (KIBA_DOCK(dock)->groups, group);
		}
		else
			sprintf (l->data, "separator%i", i);
		
		i++;
	}
	
	/* remembering the dock groups */
	groups = kiba_prefs_list_to_string(KIBA_DOCK(dock)->groups);
	kiba_prefs_write_to_config_file ("dock_groups", "list", groups);
	g_free (groups);
	
	DEBUG_MISC("Freeing Plugins..\n");
	kiba_plugins_free ();
	g_object_unref (G_OBJECT(loader));
	
	DEBUG_MISC("Destroying the kiba widget..\n");
	gtk_widget_destroy (widget);
	
	DEBUG_MISC("Cleaning up..\n");
	prefs_dbus_fini ();
	kiba_screen_destroy (kiba_screen_get());
	kiba_prefs_fini ();
	
	kiba_surface_factory_destroy();

#ifdef HAVE_SVG
	rsvg_term ();
#endif
	
	return 0;
}
