/*
 *  Copyright (C) 2006-2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

/**
 * SECTION:kiba-object
 * @short_description: Base class for objects on #KibaDock.
 * @see_also: #KibaWin,#KibaDock
 * @stability: Unstable
 *
 * Base class for objects on #KibaDock.
 */

#include "../config.h"
#include <string.h>
#include <math.h>
#include <gdk/gdkx.h>
#include <cairo-xlib.h>

#include "kiba-object.h"
#include "kiba-separator.h"
#include "kiba-win.h"
#include "kiba-prefs.h"
#include "kiba-dock.h"
#include "kiba-macros.h"
#include "kiba-screen.h"
#include "kiba-x11.h"

#ifdef USE_LIBSN
  #include "startup-notification.h"
#endif

#define OBJECT_DEFAULT_WIDTH  28
#define OBJECT_DEFAULT_HEIGHT 28
#define OBJECT_FADE_STEP      0.01f

#define KIBA_OBJECT_GET_PRIVATE(obj) \
		(G_TYPE_INSTANCE_GET_PRIVATE ((obj), KIBA_TYPE_OBJECT, KibaObjectPrivate))

G_DEFINE_TYPE (KibaObject, kiba_object, GTK_TYPE_BIN)

static const float deg_to_rad_factor = 6.283f/360.0f;

typedef struct {
	GdkWindow *event_window;
	
	char *group;    /**< dock group */
	
	gint fade_id;
	float fade_alpha;

	gint destroy_id;
	
	float scale_x; /* surface scale factor */
	float scale_y; /* surface scale factor */
	
	float gearsRotateAngle;
	float gearsAlpha;
	gint rotateGearsID;
	guint workTodo;
} KibaObjectPrivate;


enum {
	SIG_URGENT_TOGGLED,
	SIG_ACTIVE_TOGGLED,
	SIG_SWITCH_EVENT,
	SIG_KIBA_OBJECT_LAST
};

/* properties */
enum {
  _PROP_GROUP = 1,
};

static guint kiba_object_signals[SIG_KIBA_OBJECT_LAST] = { 0 };

static void kiba_object_size_request (GtkWidget      *widget,
				      GtkRequisition *requisition);
static void kiba_object_size_allocate (GtkWidget     *widget,
				       GtkAllocation *allocation);

/** Private Methods **/

/*** Helper Methods ***/

static gint __fade_timeout (gpointer data) {
	KibaObject *object = data;
	KibaObjectPrivate *priv;
	
	if (!KIBA_IS_OBJECT(object))
		return FALSE;
	
	priv = KIBA_OBJECT_GET_PRIVATE (object);
	
	kiba_object_queue_draw (object);
	
	if (object->alpha < priv->fade_alpha) {
		object->alpha += OBJECT_FADE_STEP;
		if (object->alpha > priv->fade_alpha) {
			object->alpha = priv->fade_alpha;
			priv->fade_id = 0;
			return FALSE;
		}
	}
	else {
		object->alpha -= OBJECT_FADE_STEP;
		if (object->alpha < priv->fade_alpha) {
			object->alpha = priv->fade_alpha;
			priv->fade_id = 0;
			return FALSE;
		}
	}
	
	return TRUE;
}

static gint __destroy_timeout (gpointer data) {
	KibaObject *object = data;
	KibaObjectPrivate *priv;
	GtkWidget *dock;
	float old_factor;
	
	dock = GTK_WIDGET(object)->parent;
	priv = KIBA_OBJECT_GET_PRIVATE (object);
	
	kiba_object_queue_draw (object);
	old_factor = object->destroy_factor;
	object->destroy_factor -= 0.05f;
	
	if (object->position == KIBA_OBJECT_ON_DOCK &&
	    KIBA_DOCK(dock)->scroll_offset > 0) {
		KIBA_DOCK(dock)->scroll_offset -=
			(GTK_WIDGET(object)->allocation.width -
			object->destroy_factor*GTK_WIDGET(object)->allocation.width)/4;
		KIBA_DOCK(dock)->scroll_offset = MAX(KIBA_DOCK(dock)->scroll_offset, 0);
	}
	
	gtk_widget_queue_resize_no_redraw (GTK_WIDGET(object));
	kiba_object_update_scale_factor (object);
	
	if (object->destroy_factor <= 0.0f) {
		if (GTK_WIDGET(object)->parent)
			gtk_container_remove (GTK_CONTAINER(dock), GTK_WIDGET(object));
		else
			gtk_widget_destroy (GTK_WIDGET(object));
		return FALSE;
	}
	
	return TRUE;
}

static gint __scale_timeout (gpointer data) {
	KibaObject *object = data;
	KibaObjectPrivate *priv;
	
	priv = KIBA_OBJECT_GET_PRIVATE (object);
	
	if (priv->destroy_id)
		return FALSE;
	
	object->destroy_factor += 0.05f;
	
	gtk_widget_queue_resize_no_redraw (GTK_WIDGET(object));
	
	if (object->destroy_factor >= 1.0f) {
		object->destroy_factor = 1.0f;
		kiba_object_update_scale_factor (object);
		return FALSE;
	}
	
	kiba_object_update_scale_factor (object);
	
	return TRUE;
}

static GtkAllocation __get_gear_allocation (KibaObject *object) {
	GtkAllocation ret;
	GtkWidget *widget = GTK_WIDGET(object);
	KibaDock *dock = KIBA_DOCK(widget->parent);
	const KibaLayoutData *d = &kiba_layout_data[dock->d.position];
	
	ret.width = dock->continuousProgress.size;
	ret.height = ret.width;
	ret.x = widget->allocation.x;
	ret.y = widget->allocation.y;
	
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
		ret.x -= dock->continuousProgress.size/3;
		if (d->valignment == 0)
			ret.y -= dock->continuousProgress.size/3;
		else
			ret.y += widget->allocation.height - 2*dock->continuousProgress.size/3;
	}
	else {
		ret.y -= dock->continuousProgress.size/3;
		if (d->halignment == 0)
			ret.x -= dock->continuousProgress.size/3;
		else
			ret.x += widget->allocation.width - 2*dock->continuousProgress.size/3;
	}
	
	return ret;
}

static void __gears_queue_draw (KibaObject *object) {
	GtkWidget *widget = GTK_WIDGET(object);
	GtkWidget *dock = widget->parent;
	if (!KIBA_DOCK(dock)->continuousProgress.enable)
		return;
	GdkRectangle area;
	GtkAllocation allocation = __get_gear_allocation(object);
	
	area.width = allocation.width;
	area.height = allocation.height;
	area.x = allocation.x;
	area.y = allocation.y;
	
	kiba_win_invalidate_area (dock->parent, area);
}

static gboolean __rotate_gears (gpointer data) {
	KibaObject *object = KIBA_OBJECT(data);
	KibaDock *dock = KIBA_DOCK(GTK_WIDGET(object)->parent);
	KibaObjectPrivate *priv = KIBA_OBJECT_GET_PRIVATE (object);
	
	priv->gearsRotateAngle += dock->continuousProgress.step;
	if (priv->gearsRotateAngle > 2.0f)
		priv->gearsRotateAngle = 0.0f;
	
	__gears_queue_draw(object);
	
	if (priv->workTodo > 0) {
		if (priv->gearsAlpha < 1.0f)
			priv->gearsAlpha += 0.05f;
	}
	else {
		if (priv->gearsAlpha > 0.0f)
			priv->gearsAlpha -= 0.05f;
		if (priv->gearsAlpha == 0.0f)
			DEL_FLAG(object->state, KIBA_OBJECT_WORKING);
	}
	
	return (priv->gearsAlpha != 0.0f);
}

static void __set_working (KibaObject *object,
                           gboolean    working) {
	KibaObjectPrivate *priv = KIBA_OBJECT_GET_PRIVATE (object);
	if (working) {
		SET_FLAG(object->state, KIBA_OBJECT_WORKING);
		if (priv->gearsAlpha == 0.0f)
			priv->rotateGearsID = g_timeout_add(50, __rotate_gears, object);
	}
}

static void __dock_object_size_changed_cb (KibaDock *dock,
                                           gint size,
                                           gpointer data) {
	KibaObject *object = data;
	
	if (!KIBA_IS_FIXED_OBJECT(object)) {
		object->size.width = size;
		object->size.height = size;
	}
}

/** Public functions **/

/**
 * kiba_object_set_tooltip_text:
 * @object: The KibaObject instance.
 *
 */
void kiba_object_set_tooltip_text(KibaObject *object,
                                  const char *text) {
	g_return_if_fail(KIBA_IS_OBJECT(object));
#if GTK_CHECK_VERSION(2,12,0) == TRUE
	gtk_widget_set_tooltip_text(GTK_WIDGET(object), text);
#endif
}

/**
 * kiba_object_queue_draw:
 * @object: The KibaObject instance.
 *
 * Invalidates the area on the dock used by the object.
 */
void kiba_object_queue_draw (KibaObject *object) {
	KibaObjectPrivate *priv;
	GtkWidget *widget;
	GtkWidget *win;
	GtkWidget *dock;
	GdkWindow *window;
	GdkRectangle area;
	GList *l;
	const KibaLayoutData *d;
	float width, height;
	float scale_x = 1.0f;
	float scale_y = 1.0f;
	float shadow_offset_X, shadow_offset_y;
	float offsetX = 0.0f, offsetY = 0.0f;
	float rad;
	float rotateAngle = 0.0f;
	
	g_return_if_fail (KIBA_IS_OBJECT(object));
	
	win = GTK_WIDGET(kiba_win_get_default());
	widget = GTK_WIDGET(object);
	dock = widget->parent;
	priv = KIBA_OBJECT_GET_PRIVATE (object);
	if (!GTK_WIDGET_REALIZED(widget))
		return;
	
	float alpha = object->destroy_factor * object->alpha * object->scroll_alpha;
	if (alpha == 0.0f)
		return;
	
	window = gtk_widget_get_parent_window (GTK_WIDGET(object)->parent);
	g_return_if_fail (GDK_IS_WINDOW(window));
	
	width = object->alloc.width;
	height = object->alloc.height;
	d = &kiba_layout_data[KIBA_DOCK(KIBA_WIN_CHILD(win))->d.position];
	

	if (KIBA_IS_SEPARATOR(object)) {
//TODO: separator objects, currently they paint outside their widget area
		if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
			area.width = widget->allocation.width;
			area.height = KIBA_DOCK(dock)->d.height;
			area.x = widget->allocation.x;
			area.y = dock->allocation.y + (d->valignment/2)*(dock->allocation.height - area.height);
		} else {
			area.height = widget->allocation.height;
			area.width = KIBA_DOCK(dock)->d.height;
			area.y = widget->allocation.y;
			area.x = dock->allocation.x + (d->halignment/2)*(dock->allocation.width - area.width);
		}
		kiba_win_invalidate_area (win, area);
	}
	else {
		for (l = object->custom_offset_data; l != NULL; l = l->next) {
			float customX = 0.0f, customY = 0.0f;
			KibaObjectCustomOffsetData *data = l->data;
			
			data->place_func (object, &customX, &customY, data->data);
			offsetX += customX;
			offsetY += customY;
		}
		
		for (l = object->custom_rotate_funcs; l != NULL; l = l->next) {
			float customAngle = 0.0f;
			KibaObjectCustomRotateData *data = l->data;
			
			data->func (object, &customAngle, data->data);
			rotateAngle += customAngle;
		}
		
		for (l = object->custom_scale_funcs; l != NULL; l = l->next) {
			float customX = 0.0f, customY = 0.0f;
			KibaObjectCustomScaleData *data = l->data;
			
			if (data->scale_widget)
				continue;
			
			data->scale_func (object, &customX, &customY, data->data);
			scale_x *= customX;
			scale_y *= customY;
		}
		
		area.x = object->alloc.x - (width *(scale_x-1.0f))/2;
		area.y = object->alloc.y - (height*(scale_y-1.0f))/2;
		
		area.y -= (d->valignment-1)*(offsetY);
		area.x -= (d->halignment-1)*(offsetX);
		
		if (rotateAngle != 0.0f) {
			area.height = MAX(scale_x, scale_y)*
					sqrt (height * height + width  * width);
			area.width = area.height;
			area.y -= (area.height - scale_y*height)/2;
			area.x -= (area.width - scale_x*width)/2;
		}
		else {
			area.height = scale_y*height;
			area.width  = scale_x*width;
		}
	// 	if (d->orientation == GTK_ORIENTATION_HORIZONTAL)
	// 		area.height += (d->valignment-1)*2;
	// 	else
	// 		area.width  += (d->halignment-1)*2;
		
		if (kiba_object_get_has_shadow (object)) {
			rad = deg_to_rad_factor*object->shadow_affection;
			shadow_offset_X = sin(rad)*object->shadow_offset;
			shadow_offset_y = cos(rad)*object->shadow_offset;
			
			shadow_offset_X = (shadow_offset_X*area.width)/(((float) object->size.width)*object->scale_x);
			shadow_offset_y = (shadow_offset_y*area.height)/(((float) object->size.height)*object->scale_y);
			
			if (shadow_offset_y < 0.0f) {
				area.height -= (int) shadow_offset_y+1;
				area.y += (int) shadow_offset_y-1;
			}
			else
				area.height += (int) shadow_offset_y+1;
			if (shadow_offset_X < 0.0f) {
				area.width -= (int) shadow_offset_X+1;
				area.x += (int) shadow_offset_X-1;
			}
			else
				area.width += (int) shadow_offset_X+1;
		}
	area.x -= 2;
	area.y -= 2;
	area.width += 4;
	area.height += 4;
		kiba_win_invalidate_area (win, area);
		if (kiba_object_get_has_shadow (object))
			if (shadow_offset_y < 0.0f) {
				area.y -= (int) shadow_offset_y+1;
			if (shadow_offset_X < 0.0f)
				area.x -= (int) shadow_offset_X+1;
		}
		
		if (kiba_object_get_reflectable (object)) {
			if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
				area.y = object->reflection_allocation.y + (d->valignment-1)*(offsetY +
					object->reflection_allocation.height*(scale_y/1.1 - 1.0)) -
					(object->reflection_allocation.height*(scale_y/1.1 - 1.0))/2;
			}
			else {
				area.x = object->reflection_allocation.x + (d->halignment-1)*(offsetX +
					object->reflection_allocation.width*(scale_x/1.1 - 1.0)) +
					(object->reflection_allocation.width*(scale_x/1.1 - 1.0))/2;
			}
			if (kiba_object_get_has_shadow (object)) {
				if (shadow_offset_y > 0.0f)
					area.y -= (int) shadow_offset_y-1;
				if (shadow_offset_X < 0.0f)
					area.x += (int) shadow_offset_X-1;
			}
			if (rotateAngle != 0.0f) {
				area.y -= (area.height - scale_y*height)/2;
				area.x -= (area.width - scale_x*width)/2;
			}
			kiba_win_invalidate_area (win, area);
		}
	}
}

/**
 * kiba_object_launch:
 * @object: The KibaObject Instance.
 * @command: The command to launch.
 *
 * The method executes the given command.
 * Additional it passes the command to libsn, to monitor the launch process.
 *
 * Returns: The PID.
 */
int kiba_object_launch(KibaObject *object,
                        const char *command) {
	KibaDock *dock = KIBA_DOCK(GTK_WIDGET(object)->parent);
#ifdef USE_LIBSN
	KibaSn *sn = kiba_sn_get();
	kiba_sn_launch(sn, command, G_OBJECT(object));
#endif
	return kiba_launch_programm(command);
}

/**
 * kiba_object_set_working:
 * @object: The KibaObject Instance.
 * @workTodo: TRUE if some work is Todo, otherwhise FALSE.
 *
 * Start indicating work in progress.
 */
void kiba_object_set_working (KibaObject *object,
                              gboolean    workTodo) {
	KibaObjectPrivate *priv = KIBA_OBJECT_GET_PRIVATE (object);
	
	if (workTodo)
		priv->workTodo++;
	else
		priv->workTodo--;
	
	if (priv->workTodo == 1)
		__set_working(object, TRUE);
	else {
		__set_working(object, FALSE);
		__gears_queue_draw(object);
	}
}

float kiba_object_get_destroy_factor (KibaObject *object) {
	if (!KIBA_IS_OBJECT(object))
		return 1.0f;
	
	return object->destroy_factor;
}

/**
 * kiba_object_set_alpha:
 * @object: The KibaObject instance.
 * @alpha: The alpha of KibaObject.
 *
 * Sets the alpha factor used for rendering the object surface.
 */
void kiba_object_set_alpha (KibaObject *object,
			    float      alpha) {
	KibaObjectPrivate *priv;
	
	g_return_if_fail (KIBA_IS_OBJECT(object));
	g_return_if_fail (0.0f <= alpha || alpha <= 1.0f);
	
	priv = KIBA_OBJECT_GET_PRIVATE (object);
	
	if (alpha == priv->fade_alpha)
		return;
	
	priv->fade_alpha = alpha;
	
	if (!priv->fade_id)
		priv->fade_id = g_timeout_add (20, __fade_timeout, object);
}

/**
 * kiba_object_set_icon_geometry:
 * @object: The KibaObject instance.
 * @window: The KibaWindow to set the _NET_WM_ICON property on.
 *
 * Sets the _NET_WM_ICON property on the window, for minimze/restore animations.
 */
void kiba_object_set_icon_geometry (KibaObject  *object,
				    KibaWindow  *window) {
	GtkWidget *dock;
	GtkAllocation geometry;
	gint fX, fY;
	
	g_return_if_fail (KIBA_IS_WINDOW(window));
	
	dock = GTK_WIDGET(object)->parent;
	
	fX = (int) object->alloc.x;
	fY = (int) object->alloc.y;
	
	geometry.width = GTK_WIDGET(object)->allocation.width;
	geometry.height = GTK_WIDGET(object)->allocation.height;
	geometry.x = dock->parent->allocation.x + fX;
	geometry.y = dock->parent->allocation.y + fY;

	kiba_window_set_minimize_restore_point(window, geometry.x, geometry.y, geometry.width, geometry.height);	
}

/**
 * kiba_object_show_title:
 * @widget: The KibaObject instance.
 *
 * Shows the object title.
 */
void kiba_object_show_title (GtkWidget *widget) {
	g_return_if_fail (KIBA_IS_OBJECT(widget));
	
	if (!GET_FLAG(KIBA_OBJECT(widget)->title->title->flags, KIBA_TITLE_DRAWABLE))
		return;
	
	kiba_title_object_show (KIBA_OBJECT(widget)->title);
}

/**
 * kiba_object_hide_title:
 * @widget: The KibaObject instance.
 *
 * Hides the object title.
 */
void kiba_object_hide_title (GtkWidget *widget) {
	g_return_if_fail (KIBA_IS_OBJECT(widget));
	
	kiba_title_object_hide (KIBA_OBJECT(widget)->title);
}

/**
 * kiba_object_set_urgent:
 * @object: The KibaObject instance.
 * @urgent: TRUE to set the urgent flag, otherwise FALSE to unset the urgent flag.
 *
 * Sets whether the object is urgent.
 */
void kiba_object_set_urgent (KibaObject *object,
			     gboolean    urgent) {
	g_return_if_fail (KIBA_IS_OBJECT(object));
	
	if (urgent && !GET_FLAG(object->state, KIBA_OBJECT_URGENT))
		SET_FLAG(object->state, KIBA_OBJECT_URGENT);
	else if (!urgent && GET_FLAG(object->state, KIBA_OBJECT_URGENT))
		DEL_FLAG(object->state, KIBA_OBJECT_URGENT);
	
	g_signal_emit (object,
		kiba_object_signals[SIG_URGENT_TOGGLED],
		0, urgent);
}

/**
 * kiba_object_get_urgent:
 * @object: The KibaObject instance.
 *
 * Returns whether the object is urgent.
 *
 * Returns: TRUE if the object is urgent, otherwise FALSE is returned.
 */
gboolean kiba_object_get_urgent (KibaObject *object) {
	g_return_val_if_fail (KIBA_IS_OBJECT(object), FALSE);
	
	return GET_FLAG(object->state, KIBA_OBJECT_URGENT);
}

/**
 * kiba_object_set_active:
 * @widget: The KibaObject instance.
 * @active: TRUE to activate the object, otherwise FALSE to deactivate the object.
 *
 * Sets whether the object is active.
 */
void kiba_object_set_active (GtkWidget *widget,
			     gboolean   active) {
	KibaObject *object;
	KibaWin *win;
	
	g_return_if_fail (KIBA_IS_OBJECT(widget));
	
	object = KIBA_OBJECT(widget);
	win = KIBA_WIN(widget->parent->parent);
	
	if (active && !GET_FLAG(object->state, KIBA_OBJECT_ACTIVE))
		SET_FLAG(object->state, KIBA_OBJECT_ACTIVE);
	else if (!active && GET_FLAG(object->state, KIBA_OBJECT_ACTIVE))
		DEL_FLAG(object->state, KIBA_OBJECT_ACTIVE);
	
	if (active)
		kiba_triangle_add_widget (win->triangle, widget);
	else
		kiba_triangle_remove_widget (win->triangle, widget);
	
	g_signal_emit (object,
		kiba_object_signals[SIG_ACTIVE_TOGGLED],
		0, active);
}

/**
 * kiba_object_get_active:
 * @widget: The KibaObject instance.
 *
 * Returns whether the object is active.
 *
 * Returns: TRUE if the object is active, otherwise FALSE is returned.
 */
gboolean kiba_object_get_active (GtkWidget *widget) {
	g_return_val_if_fail (KIBA_IS_OBJECT(widget), FALSE);
	
	return GET_FLAG(KIBA_OBJECT(widget)->state, KIBA_OBJECT_ACTIVE);
}

/**
 * kiba_object_set_prelight:
 * @widget: The KibaObject instance.
 *
 * Sets prelight.
 */
void kiba_object_set_prelight (GtkWidget *widget) {
	g_return_if_fail (KIBA_IS_OBJECT(widget));
	
	if (!GET_FLAG(KIBA_OBJECT(widget)->props, KIBA_OBJECT_PRELIGHTABLE))
		return;
	
	if (!GET_FLAG(KIBA_OBJECT(widget)->state, KIBA_OBJECT_PRELIGHTED))
		SET_FLAG(KIBA_OBJECT(widget)->state, KIBA_OBJECT_PRELIGHTED);
	
	kiba_object_queue_draw (KIBA_OBJECT(widget));
}

/**
 * kiba_object_unset_prelight:
 * @widget: The KibaObject instance.
 *
 * Unsets prelight.
 */
void kiba_object_unset_prelight (GtkWidget *widget) {
	g_return_if_fail (KIBA_IS_OBJECT(widget));
	
	if (GET_FLAG(KIBA_OBJECT(widget)->state, KIBA_OBJECT_PRELIGHTED))
		DEL_FLAG(KIBA_OBJECT(widget)->state, KIBA_OBJECT_PRELIGHTED);
	
	kiba_object_queue_draw (KIBA_OBJECT(widget));
}

/**
 * kiba_object_set_title:
 * @object: The KibaObject instance.
 * @title: The object title.
 *
 * Sets the object title's text.
 **/
void kiba_object_set_title (KibaObject *object,
                            const char *title) {
	g_return_if_fail (KIBA_IS_OBJECT (object));
	
	kiba_title_object_set_text (object->title, title);
}

/**
 * kiba_object_get_title:
 * @object: The KibaObject instance.
 *
 * Returns the object title.
 *
 * Returns: The object title.
 */
const char* kiba_object_get_title (KibaObject *object) {
	g_return_val_if_fail (KIBA_IS_OBJECT (object), NULL);
	
	return kiba_title_object_get_text(object->title);
}

void kiba_object_set_prelightable (KibaObject *object,
				   gboolean    prelightable) {
	g_return_if_fail (KIBA_IS_OBJECT (object));
	
	if (prelightable) {
		if (!GET_FLAG(object->props, KIBA_OBJECT_PRELIGHTABLE))
			SET_FLAG(object->props, KIBA_OBJECT_PRELIGHTABLE);
	}
	else if (GET_FLAG(object->props, KIBA_OBJECT_PRELIGHTABLE))
		DEL_FLAG(object->props, KIBA_OBJECT_PRELIGHTABLE);
}

gboolean kiba_object_get_prelightable (KibaObject *object) {
	g_return_val_if_fail (KIBA_IS_OBJECT(object), FALSE);
	
	return GET_FLAG(object->props, KIBA_OBJECT_REFLECTABLE);
}

/**
 * kiba_object_set_reflectable:
 * @object: The KibaObject instance.
 * @reflect: TRUE to enable object reflection, otherwise FALSE to disable object reflection.
 *
 * Sets whether the object is reflectable.
 */
void kiba_object_set_reflectable (KibaObject *object,
				  gboolean    reflect) {
	g_return_if_fail (KIBA_IS_OBJECT (object));
	
	kiba_object_queue_draw (object);
	
	if (reflect) {
		if (!GET_FLAG(object->props, KIBA_OBJECT_REFLECTABLE))
			SET_FLAG(object->props, KIBA_OBJECT_REFLECTABLE);
	}
	else if (GET_FLAG(object->props, KIBA_OBJECT_REFLECTABLE))
		DEL_FLAG(object->props, KIBA_OBJECT_REFLECTABLE);
	else
		return;
	
	kiba_object_queue_draw (object);
}

/**
 * kiba_object_get_reflectable:
 * @object: The KibaObject instance.
 *
 * Returns whether the object is reflectable.
 *
 * Returns: TRUE if the object has a reflection, otherwise FALSE is returned.
 */
gboolean kiba_object_get_reflectable (KibaObject *object) {
	g_return_val_if_fail (KIBA_IS_OBJECT(object), FALSE);
	
	return GET_FLAG(object->props, KIBA_OBJECT_REFLECTABLE);
}

gboolean kiba_object_get_destroying (KibaObject *object) {
	KibaObjectPrivate *priv;
	
	g_return_val_if_fail (KIBA_IS_OBJECT(object), FALSE);
	
	priv = KIBA_OBJECT_GET_PRIVATE (object);
	
	return priv->destroy_id != 0;
}

/**
 * kiba_object_set_has_shadow:
 * @object: The KibaObject instance.
 * @has_shadow: TRUE to enable the object shadow, otherwise FALSE.
 *
 * Sets whether the object has a shadow.
 */
void kiba_object_set_has_shadow (KibaObject *object,
				 gboolean    has_shadow) {
	g_return_if_fail (KIBA_IS_OBJECT (object));
	
	if (has_shadow) {
		if (!GET_FLAG(object->props, KIBA_OBJECT_HAS_SHADOW))
			SET_FLAG(object->props, KIBA_OBJECT_HAS_SHADOW);
	}
	else if (GET_FLAG(object->props, KIBA_OBJECT_HAS_SHADOW))
		DEL_FLAG(object->props, KIBA_OBJECT_HAS_SHADOW);
	else
		return;
	
	kiba_object_queue_draw (object);
}

/**
 * kiba_object_get_has_shadow:
 * @object: The KibaObject instance.
 *
 * Returns whether the object has a shadow.
 *
 * Returns: TRUE if the object has a shadow, otherwise FALSE is returned.
 */
gboolean kiba_object_get_has_shadow (KibaObject *object) {
	g_return_val_if_fail (KIBA_IS_OBJECT(object), FALSE);
	
	return GET_FLAG(object->props, KIBA_OBJECT_HAS_SHADOW);
}

//TODO: remove
void kiba_object_resize (KibaObject *object) {
	GtkAllocation allocation;
	GtkRequisition requisition;
	
	allocation = GTK_WIDGET(object)->allocation;
	
	kiba_object_size_request (GTK_WIDGET(object), &requisition);
	if (requisition.width || requisition.height) {
		/* non-empty window */
		allocation.width = requisition.width;
		allocation.height = requisition.height;
	}
	GTK_WIDGET(object)->requisition = requisition;
	kiba_object_size_allocate (GTK_WIDGET(object), &allocation);
}

/**
 * kiba_object_set_requisition_factor:
 * @object: The KibaObject instance.
 * @scale_x: The KibaObject widget width scale factor.
 * @scale_y: The KibaObject widget height scale factor.
 *
 * Sets size scale factors of the object.
 * The actual object width is (*object->size)*object->scale_x, ignoring the custom scale factors.
 */
void kiba_object_set_requisition_factor (KibaObject *object,
					 float      scale_x,
					 float      scale_y) {
	g_return_if_fail (KIBA_IS_OBJECT(object));
	
	object->scale_x = scale_x;
	object->scale_y = scale_y;
	
	kiba_object_resize (object);
}

/**
 * kiba_object_switch_surface:
 * @object: The KibaObject instance.
 *
 * Switches the object's surface, sending the "switched" signal.
 */
void kiba_object_switch_surface (KibaObject *object) {
	g_return_if_fail (KIBA_IS_OBJECT(object));
	
	g_signal_emit (object,
		kiba_object_signals[SIG_SWITCH_EVENT], 0);
}

/**
 * kiba_object_destroy:
 * @object: The KibaObject instance.
 *
 * Destroys the KibaObject, scaling the object down and unrealizing it.
 */
void kiba_object_destroy (KibaObject *object) {
	KibaObjectPrivate *priv;
	
	g_return_if_fail (KIBA_IS_OBJECT(object));
	
	priv = KIBA_OBJECT_GET_PRIVATE (object);
	
	object->paint = NULL;
	
	if (!priv->destroy_id)
		priv->destroy_id = g_timeout_add (20, __destroy_timeout, object);
	
	if (kiba_object_get_active(GTK_WIDGET(object)))
		kiba_object_set_active(GTK_WIDGET(object), FALSE);
}

/**
 * kiba_object_set_paint_function:
 * @object: The KibaObject instance.
 * @func: The default rendering method.
 * @data: The user data.
 *
 * Sets custom render function.
 */
void kiba_object_set_paint_function (KibaObject          *object,
				     KibaObjectPaintFunc  func,
				     gpointer             data) {
	g_return_if_fail (KIBA_IS_OBJECT(object));
	
	object->paint = func;
	object->paint_data = data;
}

/**
 * kiba_object_paint_reflection:
 * @object: The KibaObject instance.
 * @cr: The drawing context.
 *
 * Renders the object reflection to the context.
 */
void kiba_object_paint_reflection (KibaObject *object,
				   cairo_t    *cr) {
	KibaObjectPrivate *priv;
	KibaDock *dock;
	cairo_surface_t *surface;
	const KibaLayoutData *d;
	float alpha;
	gint unscaled_width;
	gint unscaled_height;
	float rotateAngle = 0.0f;
	float buf_alpha;
	GList *l;
	
	g_return_if_fail (KIBA_IS_OBJECT(object));
	g_return_if_fail (GTK_WIDGET(object)->parent != NULL);
	
	if (!GET_FLAG(object->props, KIBA_OBJECT_REFLECTABLE))
		return;
	
	priv = KIBA_OBJECT_GET_PRIVATE (object);
	dock = KIBA_DOCK(GTK_WIDGET(object)->parent);
	d = &kiba_layout_data[dock->d.position];
	
	unscaled_width  = (int) ((object->size.width) * object->scale_x);
	unscaled_height = (int) ((object->size.height) * object->scale_y);
	
	if (object->switch_surface)
		surface = object->switch_surface;
	else
		surface = object->surface_data->surface;
	
	g_return_if_fail (surface != NULL);
	
	alpha = object->alpha*dock->style.reflection_alpha;
	
	if (!kiba_object_scale (object, cr))
		return;
		
	for (l = object->custom_rotate_funcs; l != NULL; l = l->next) {
		float customAngle = 0.0f;
		KibaObjectCustomRotateData *data = l->data;
		
		data->func (object, &customAngle, data->data);
		rotateAngle += customAngle;
	}
	
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
		cairo_scale (cr, 1, dock->style.reflection_scale_factor);
		cairo_translate (cr, 0, (-(d->valignment-2)/2)*
			(1.0f-dock->style.reflection_scale_factor)*
			unscaled_height);
	}
	else {
		cairo_scale (cr, dock->style.reflection_scale_factor, 1);
		cairo_translate (cr, (-(d->halignment-2)/2)*
			(1.0f-dock->style.reflection_scale_factor)*
			unscaled_width, 0);
	}
	
	cairo_translate (cr,
		unscaled_width/2,
		unscaled_height/2);
	cairo_rotate (cr, M_PI * (rotateAngle + 1.0));
	cairo_translate (cr,
		-unscaled_width/2,
		-unscaled_height/2);
	
	buf_alpha = object->alpha;
	object->alpha = alpha;
	cairo_save (cr);
		kiba_object_paint_shadow (object, cr);
	cairo_restore (cr);
	cairo_save(cr);
		if (!object->paint)
			kiba_object_paint (object, cr, NULL);
		else
			object->paint (object, cr, object->paint_data);
	cairo_restore (cr);
	object->alpha = buf_alpha;
}

void kiba_object_paint_gear (KibaObject *object,
                             cairo_t *cr) {
	KibaObjectPrivate *priv = KIBA_OBJECT_GET_PRIVATE (object);
	GtkWidget *widget = GTK_WIDGET(object);
	KibaDock *dock = KIBA_DOCK(widget->parent);
	GtkAllocation allocation = __get_gear_allocation(object);
	
	cairo_translate (cr, allocation.x, allocation.y);
	
	cairo_save(cr);
		cairo_translate(cr, dock->continuousProgress.size/2, dock->continuousProgress.size/2);
		cairo_rotate(cr, priv->gearsRotateAngle*M_PI);
		cairo_translate(cr, -dock->continuousProgress.size/2, -dock->continuousProgress.size/2);
		
		cairo_set_source_surface (cr, dock->gearSurface, 0, 0);
		cairo_paint_with_alpha(cr, priv->gearsAlpha);
	cairo_restore(cr);
	
	cairo_set_source_surface (cr, dock->gearSurfaceOverlay, 0, 0);
	cairo_paint_with_alpha(cr, priv->gearsAlpha);
}

/**
 * kiba_object_paint_shadow:
 * @object: The KibaObject instance.
 * @cr: The drawing context.
 *
 * Renders the object shadow to the context.
 */
void kiba_object_paint_shadow (KibaObject *object,
			       cairo_t    *cr) {
	KibaObjectPrivate *priv;
	cairo_surface_t *surface;
	float offset_X, offset_y;
	float rad;
	float rotateAngle = 0.0f;
	GList *l;
	
	g_return_if_fail (KIBA_IS_OBJECT(object));
	g_return_if_fail (cr != NULL);
	
	if (!GET_FLAG(object->props, KIBA_OBJECT_HAS_SHADOW))
		return;
	
	priv = KIBA_OBJECT_GET_PRIVATE (object);
	
	if (object->switch_surface)
		surface = object->switch_surface;
	else
		surface = object->surface_data->surface;
	
	g_return_if_fail (surface != NULL);
		
	for (l = object->custom_rotate_funcs; l != NULL; l = l->next) {
		float customAngle = 0.0f;
		KibaObjectCustomRotateData *data = l->data;
		
		data->func (object, &customAngle, data->data);
		rotateAngle += customAngle;
	}
	
	rad = deg_to_rad_factor*object->shadow_affection;
	offset_X = sin(rad)*object->shadow_offset;
	offset_y = cos(rad)*object->shadow_offset;
	
	/* dont rotate the translation, the light is fixed ;) */
	cairo_rotate_centered (cr,
		(2.0-rotateAngle),
		GTK_WIDGET(object)->allocation.width /2,
		GTK_WIDGET(object)->allocation.height/2);
	cairo_translate (cr, offset_X, offset_y);
	cairo_rotate_centered (cr,
		rotateAngle,
		GTK_WIDGET(object)->allocation.width /2,
		GTK_WIDGET(object)->allocation.height/2);
	
	cairo_save (cr);
		cairo_set_source_rgba (cr, object->shadow_color.r,
			object->shadow_color.g, object->shadow_color.b,
			object->destroy_factor*object->alpha*object->scroll_alpha*object->shadow_color.a);
		/* using alpha mask from the object surface */
		cairo_mask_surface (cr, surface, 0, 0);
	cairo_restore (cr);
}

/**
 * kiba_object_remove_custom_rotate_function_by_data:
 * @object: The KibaObject instance.
 * @user_data: The user data.
 *
 * Removes custom rotate function.
 */
void kiba_object_remove_custom_rotate_function_by_data (KibaObject *object,
						        gpointer    user_data) {
	KibaObjectCustomRotateData *data;
	GList *l;
	
	for (l = object->custom_rotate_funcs; l != NULL; l = l->next) {
		data = l->data;
		
		if (data->data == user_data) {
			g_free (data);
			object->custom_rotate_funcs = g_list_remove (object->custom_rotate_funcs, l->data);
			
			return;
		}
	}
}

/**
 * kiba_object_add_custom_rotate_function:
 * @object: The KibaObject instance.
 * @func: The rotate method.
 * @user_data: The user data.
 *
 * Adds rotate function, for a custom rotate angle.
 */
void kiba_object_add_custom_rotate_function (KibaObject                 *object,
					     KibaObjectCustomRotateFunc  func,
					     gpointer                    user_data) {
	KibaObjectCustomRotateData *data;
	GList *l;
	
	for (l = object->custom_rotate_funcs; l != NULL; l = l->next) {
		data = l->data;
		
		if (data->func == func) {
			data->data = user_data;
			
			return;
		}
	}
	
	data = g_new0 (KibaObjectCustomRotateData, 1);
	data->func = func;
	data->data = user_data;
	
	object->custom_rotate_funcs = g_list_append (object->custom_rotate_funcs, data);
}

/**
 * kiba_object_remove_custom_offset_function_by_data:
 * @object: The KibaObject instance.
 * @user_data: The user data.
 *
 * Removes custom offset function.
 */
void kiba_object_remove_custom_offset_function_by_data (KibaObject *object,
						        gpointer    user_data) {
	KibaObjectCustomOffsetData *data;
	GList *l;
	
	for (l = object->custom_offset_data; l != NULL; l = l->next) {
		data = l->data;
		
		if (data->data == user_data) {
			g_free (data);
			object->custom_offset_data = g_list_remove (object->custom_offset_data, l->data);
			
			return;
		}
	}
}

/**
 * kiba_object_add_custom_offset_function:
 * @object: The KibaObject instance.
 * @place_func: The surface offset method.
 * @max_func: The method to get the maximal offset.
 * @user_data: The user data.
 *
 * Adds offset function, for custom surface positions.
 */
void kiba_object_add_custom_offset_function (KibaObject                  *object,
					     KibaObjectCustomOffsetFunc   place_func,
					     KibaObjectCustomMaxScaleFunc max_func,
					     gpointer                     user_data) {
	KibaObjectCustomOffsetData *data;
	GList *l;
	
	for (l = object->custom_offset_data; l != NULL; l = l->next) {
		data = l->data;
		
		if (data->place_func == place_func) {
			data->data = user_data;
			
			return;
		}
	}
	
	data = g_new0 (KibaObjectCustomOffsetData, 1);
	data->place_func = place_func;
	data->max_func = max_func;
	data->data = user_data;
	
	object->custom_offset_data = g_list_append (object->custom_offset_data, data);
}

/**
 * kiba_object_remove_custom_paint_function_by_data:
 * @object: The KibaObject instance.
 * @user_data: The user data.
 *
 * Removes custom paint function.
 */
void kiba_object_remove_custom_paint_function_by_data (KibaObject *object,
						       gpointer    user_data) {
	KibaObjectCustomPaintData *data;
	GList *l;
	
	for (l = object->custom_paint_funcs_before; l != NULL; l = l->next) {
		data = l->data;
		
		if (data->data == user_data) {
			g_free (data);
			object->custom_paint_funcs_before =
				g_list_remove (object->custom_paint_funcs_before, l->data);
			
			return;
		}
	}
	
	for (l = object->custom_paint_funcs_after; l != NULL; l = l->next) {
		data = l->data;
		
		if (data->data == user_data) {
			g_free (data);
			object->custom_paint_funcs_after =
				g_list_remove (object->custom_paint_funcs_after, l->data);
			
			return;
		}
	}
}

/**
 * kiba_object_add_custom_paint_function:
 * @object: The KibaObject instance.
 * @func: The paint method.
 * @before: TRUE to call this method before the default rendering method, otherwise FALSE.
 * @user_data: The user data.
 *
 * Adds paint function, for custom rendering ontop/below the object.
 */
void kiba_object_add_custom_paint_function (KibaObject           *object,
					    KibaObjectPaintFunc   func,
					    gboolean              before,
					    gpointer              user_data) {
	KibaObjectCustomPaintData *data;
	GList **data_list;
	GList *l;
	
	if (before)
		data_list = &object->custom_paint_funcs_before;
	else
		data_list = &object->custom_paint_funcs_after;
	
	for (l = *data_list; l != NULL; l = l->next) {
		data = l->data;
		
		if (data->func == func) {
			data->data = user_data;
			
			return;
		}
	}
	
	data = g_new0 (KibaObjectCustomPaintData, 1);
	data->func = func;
	data->data = user_data;
	
	*data_list = g_list_append (*data_list, data);
}

/**
 * kiba_object_remove_custom_scale_function_by_data:
 * @object: The KibaObject instance.
 * @user_data: The user data.
 *
 * Removes custom scale function.
 */
void kiba_object_remove_custom_scale_function_by_data (KibaObject *object,
						       gpointer    user_data) {
	KibaObjectCustomScaleData *data;
	GList *l;
	
	for (l = object->custom_scale_funcs; l != NULL; l = l->next) {
		data = l->data;
		
		if (data->data == user_data) {
			g_free (data);
			object->custom_scale_funcs = g_list_remove (object->custom_scale_funcs, l->data);
			
			return;
		}
	}
	
}

/**
 * kiba_object_add_custom_scale_function:
 * @object: The KibaObject instance.
 * @scale_func: The scale method.
 * @max_func: The method to return the maximal scale factor.
 * @scale_widget: TRUE to scale the KibaObject widget, otherwise FALSE to scale only the surface.
 * @user_data The user data.
 *
 * Adds scale function, for custom scale animations.
 */
void kiba_object_add_custom_scale_function (KibaObject                   *object,
					    KibaObjectCustomScaleFunc     scale_func,
					    KibaObjectCustomMaxScaleFunc  max_func,
					    gboolean                      scale_widget,
					    gpointer                      user_data) {
	KibaObjectCustomScaleData *data;
	GList *l;
	
	for (l = object->custom_scale_funcs; l != NULL; l = l->next) {
		data = l->data;
		
		if (data->scale_func == scale_func) {
			data->data = user_data;
			data->scale_widget = scale_widget;
			
			return;
		}
	}
	
	data = g_new0 (KibaObjectCustomScaleData, 1);
	data->scale_func = scale_func;
	data->max_func = max_func;
	data->scale_widget = scale_widget;
	data->data = user_data;
	
	object->custom_scale_funcs = g_list_append (object->custom_scale_funcs, data);
}

/**
 * kiba_object_is_on_dock:
 * @object: The KibaObject instance.
 *
 * Returns whether object is on the dock.
 *
 * Returns: TRUE if the object is on the dock, otherwise FALSE is returned.
 **/
gboolean kiba_object_is_on_dock (KibaObject *object) {
	g_return_val_if_fail (KIBA_IS_OBJECT(object), FALSE);
	
	return object->position == KIBA_OBJECT_ON_DOCK;
}

/**
 * kiba_object_update_scale_factor:
 * @object: The KibaObject instance.
 *
 * Applys custom scale factor changes to the object size.
 *
 * Returns: TRUE if the scale factor changed, otherwise FALSE is returned.
 */
gboolean kiba_object_update_scale_factor (KibaObject *object) {
	KibaObjectPrivate *priv = KIBA_OBJECT_GET_PRIVATE(object);
	GtkWidget *dock;
	float scale_x = 1.0f;
	float scale_y = 1.0f;
	GList *l;
	
	g_return_if_fail (KIBA_IS_OBJECT(object));
	
	scale_x *= object->destroy_factor;
	scale_y *= object->destroy_factor;
	
	for (l = object->custom_scale_funcs; l != NULL; l = l->next) {
		KibaObjectCustomScaleData *data = l->data;
		float func_scale_x = 1.0f;
		float func_scale_y = 1.0f;
		
		data->scale_func (object, &func_scale_x, &func_scale_y, data->data);
		
		scale_x *= func_scale_x;
		scale_y *= func_scale_y;
	}
	
	if (scale_x == priv->scale_x && scale_y == priv->scale_y)
		return FALSE;
	
	priv->scale_x = scale_x;
	priv->scale_y = scale_y;
	
	kiba_object_queue_draw (object);
	
	return TRUE;
}

/*** Rendering functions ***/

/**
 * kiba_object_scale:
 * @object: The KibaObject instance.
 * @cr: The drawing context.
 *
 * Scales the context by object's scale factor(s).
 *
 * Returns: TRUE if the scale factor is > 0.0, otherwise FALSE is returned.
 */
gboolean kiba_object_scale (KibaObject *object,
			    cairo_t    *cr) {
	KibaObjectPrivate *priv;
	float factor_x = 1.0f;
	float factor_y = 1.0f;
	GList *l;
	
	g_return_if_fail (KIBA_IS_OBJECT(object));
	g_return_if_fail (cr != NULL);
	
	priv = KIBA_OBJECT_GET_PRIVATE (object);
	
	if (priv->scale_x == 0.0f || priv->scale_y == 0.0f)
		return FALSE;
	
	for (l = object->custom_scale_funcs; l != NULL; l = l->next) {
		KibaObjectCustomScaleData *data = l->data;
		
		if (data->scale_widget)
			continue;
		
		float custom_factor_x = 1.0f;
		float custom_factor_y = 1.0f;
		
		data->scale_func (object,
			&custom_factor_x,
			&custom_factor_y,
			data->data);
		
		factor_x *= custom_factor_x;
		factor_y *= custom_factor_y;
	}
	
	cairo_translate (cr,
		((1.0-factor_x)*GTK_WIDGET(object)->allocation.width)/2,
		((1.0-factor_y)*GTK_WIDGET(object)->allocation.height)/2);
	cairo_scale (cr, priv->scale_x, priv->scale_y);
	
	return TRUE;
}

/**
 * kiba_object_rotate:
 * @object: The KibaObject instance.
 * @cr: The drawing context.
 *
 * Rotates the context by object's rotate angle.
 */
void kiba_object_rotate (KibaObject *object,
			 cairo_t    *cr) {
	KibaObjectPrivate *priv;
	gint unscaled_width;
	gint unscaled_height;
	float rotateAngle = 0.0f;
	GList *l;
	
	g_return_if_fail (KIBA_IS_OBJECT(object));
	g_return_if_fail (cr != NULL);
	
	priv = KIBA_OBJECT_GET_PRIVATE (object);
		
	for (l = object->custom_rotate_funcs; l != NULL; l = l->next) {
		float customAngle = 0.0f;
		KibaObjectCustomRotateData *data = l->data;
		
		data->func (object, &customAngle, data->data);
		rotateAngle += customAngle;
	}
	
	unscaled_width  = (int) ((object->size.width) * object->scale_x);
	unscaled_height = (int) ((object->size.height) * object->scale_y);
	
	if (rotateAngle == 0.0f)
		return;
	
	cairo_rotate_centered (cr,
		rotateAngle,
		unscaled_width/2,
		unscaled_height/2);
}

/**
 * kiba_object_paint_with_prelight:
 * @object: The KibaObject instance.
 * @surface: The surface to paint with prelight.
 * @cr: The drawing context.
 *
 * Renders the context with prelight.
 */
void kiba_object_paint_with_prelight (KibaObject       *object,
				      cairo_surface_t  *surface,
				      cairo_t          *cr) {
	KibaObjectPrivate *priv;
	KibaDock *dock;
	float alpha;
	
	g_return_if_fail (KIBA_IS_OBJECT(object));
	g_return_if_fail (GTK_WIDGET(object)->parent != NULL);
	g_return_if_fail (cr != NULL);
	
	priv = KIBA_OBJECT_GET_PRIVATE (object);
	alpha = object->destroy_factor * object->alpha * object->scroll_alpha;
	dock = KIBA_DOCK(GTK_WIDGET(object)->parent);
	
	cairo_paint_with_alpha (cr, alpha);
	
	if (!GET_FLAG(object->state, KIBA_OBJECT_PRELIGHTED))
		return;
	
	cairo_set_source_rgba (cr, dock->style.prelight_rgba.r,
		dock->style.prelight_rgba.g,
		dock->style.prelight_rgba.b,
		alpha * dock->style.prelight_rgba.a);
	cairo_mask_surface (cr, surface, 0, 0);
}

/**
 * kiba_object_paint:
 * @object: The KibaObject instance.
 * @cr: The drawing context.
 * @data: The user data.
 *
 * Default rendering method.
 */
void kiba_object_paint (KibaObject *object,
		        cairo_t    *cr,
		        gpointer    data) {
	cairo_surface_t *surface;
	
	g_return_if_fail (KIBA_IS_OBJECT (object));
	
	if (object->switch_surface)
		surface = object->switch_surface;
	else
		surface = object->surface_data->surface;
	g_return_if_fail (surface != NULL);
	
	cairo_set_source_surface (cr, surface, 0, 0);
	kiba_object_paint_with_prelight (object, surface, cr);
}

/*** Class Implementations ***/

static void kiba_object_set_property (GObject      *object,
                        guint         property_id,
                        const GValue *value,
                        GParamSpec   *pspec) {
	KibaObject *self = KIBA_OBJECT(object);
	KibaObjectPrivate *priv = KIBA_OBJECT_GET_PRIVATE (self);
	
	switch (property_id) {
		case _PROP_GROUP:
			priv->group = g_value_dup_string (value);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(object,property_id,pspec);
			break;
	}
}

static void kiba_object_get_property (GObject      *object,
                        guint         property_id,
                        GValue       *value,
                        GParamSpec   *pspec) {
	KibaObject *self = KIBA_OBJECT(object);
	KibaObjectPrivate *priv = KIBA_OBJECT_GET_PRIVATE (self);
	
	switch (property_id) {
		case _PROP_GROUP: 
			g_value_set_string (value, priv->group);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(object,property_id,pspec);
			break;
	}
}

/**
 * gtk_container_add implementation.
 **/
static void kiba_object_add (GtkContainer *object,
			     GtkWidget    *widget) {
	GtkBin *bin;
	
	g_return_if_fail (GTK_IS_WIDGET (widget));
	
	bin = GTK_BIN (object);
	
	g_return_if_fail (bin->child == NULL);
	
	gtk_container_set_focus_child (object, widget);
	
	gtk_widget_set_parent (widget, GTK_WIDGET (bin));
	bin->child = widget;
	
	if (GTK_WIDGET_REALIZED (widget->parent))
		gtk_widget_realize (widget);
	
	if (GTK_WIDGET_VISIBLE (widget->parent) && GTK_WIDGET_VISIBLE (widget)) {
		if (GTK_WIDGET_MAPPED (widget->parent))
			gtk_widget_map (widget);
		
		gtk_widget_queue_resize (widget);
	}
}

/**
 * gtk_container_remove implementation.
 **/
static void kiba_object_remove (GtkContainer *object,
			        GtkWidget    *widget) {
	GtkBin *bin;
	gboolean widget_was_visible;
	
	g_return_if_fail (GTK_IS_WIDGET (widget));
	
	bin = GTK_BIN (object);
	g_return_if_fail (bin->child == widget);
	
	widget_was_visible = GTK_WIDGET_VISIBLE (widget);
	
	gtk_widget_unparent (widget);
	bin->child = NULL;
	
	/* queue resize regardless of GTK_WIDGET_VISIBLE (container),
	* since that's what is needed by toplevels, which derive from GtkBin.
	*/
	if (widget_was_visible)
		gtk_widget_queue_resize (GTK_WIDGET (object));
}

/**
 * gtk_widget_size_request implementation.
 **/
static void kiba_object_size_request (GtkWidget      *widget,
				      GtkRequisition *requisition) {
	KibaObject *object = KIBA_OBJECT(widget);
	GtkWidget *child = gtk_bin_get_child (GTK_BIN(widget));
	GList *l;
	float scale_x = 1.0f;
	float scale_y = 1.0f;
	
	/* using the child allocation, if there is a child */
	if (child) {
		gtk_widget_size_request (child, requisition);
		return;
	}
	
	scale_x = object->scale_x;
	scale_y = object->scale_y;
	requisition->width  = object->size.width;
	requisition->height = object->size.height;
	
	for (l = object->custom_scale_funcs; l != NULL; l = l->next) {
		KibaObjectCustomScaleData *data = l->data;
		
		if (!data->scale_widget)
			continue;
		
		float custom_factor_x = 1.0f;
		float custom_factor_y = 1.0f;
		
		data->scale_func (object,
			&custom_factor_x,
			&custom_factor_y,
			data->data);
		
		scale_x *= custom_factor_x;
		scale_y *= custom_factor_y;
	}
	
	requisition->width  = (int) requisition->width*scale_x;
	requisition->height = (int) requisition->height*scale_y;
}

/**
 * gtk_widget_size_allocat implementation.
 **/
static void kiba_object_size_allocate (GtkWidget     *widget,
				       GtkAllocation *allocation) {
	KibaObjectPrivate *priv = KIBA_OBJECT_GET_PRIVATE (widget);
	KibaObject *object = KIBA_OBJECT(widget);
	GtkWidget *child;
	const KibaLayoutData *d;
	GList *l;
	
	if (widget->allocation.width != allocation->width ||
	    widget->allocation.height != allocation->height ||
	    widget->allocation.y != allocation->y ||
	    widget->allocation.x != allocation->x) {
		kiba_object_queue_draw (object);
		
		widget->allocation = *allocation;
		
		kiba_object_queue_draw (object);
	}
	
	child = gtk_bin_get_child (GTK_BIN(widget));
	if (child) {
		GtkAllocation child_allocation;
		
		child_allocation = widget->allocation;
		child_allocation.x = 0;
		child_allocation.y = 0;
		gtk_widget_size_allocate (child, &child_allocation);
	}
	
	if (!GTK_WIDGET_REALIZED (widget))
		return;
	
	gdk_window_move_resize (priv->event_window,
			widget->allocation.x,
			widget->allocation.y,
			widget->allocation.width,
			widget->allocation.height);
}

/**
 * gtk_widget_unmap implementation.
 **/
static void kiba_object_unmap (GtkWidget *widget) {
	KibaObjectPrivate *priv = KIBA_OBJECT_GET_PRIVATE (widget);
	GtkBin *bin;
	
	GTK_WIDGET_UNSET_FLAGS (widget, GTK_MAPPED);
	bin = GTK_BIN (widget);
	
	gdk_window_hide (priv->event_window);
	
	if (bin->child && GTK_WIDGET_MAPPED (bin->child))
		gtk_widget_unmap (bin->child);
}

/**
 * gtk_widget_map implementation.
 **/
static void kiba_object_map (GtkWidget *widget) {
	KibaObjectPrivate *priv = KIBA_OBJECT_GET_PRIVATE (widget);
	GtkBin *bin;
	
	GTK_WIDGET_SET_FLAGS (widget, GTK_MAPPED);
	bin = GTK_BIN (widget);
	
	gdk_window_show (priv->event_window);
	
	if (bin->child && GTK_WIDGET_VISIBLE (bin->child) && !GTK_WIDGET_MAPPED (bin->child))
		gtk_widget_map (bin->child);
}

/**
 * gtk_widget_unrealize implementation.
 **/
static void kiba_object_unrealize (GtkWidget *widget) {
	KibaObject *object = KIBA_OBJECT(widget);
	KibaObjectPrivate *priv = KIBA_OBJECT_GET_PRIVATE (widget);
	GList *l;
	
	gdk_window_set_user_data (priv->event_window, NULL);
	gdk_window_destroy (priv->event_window);
	priv->event_window = NULL;
	
	if (priv->fade_id)
		g_source_remove (priv->fade_id);
	if (priv->destroy_id)
		g_source_remove (priv->destroy_id);
	
	if (object->surface_data) {
		if (object->surface_data->surface)
			cairo_surface_destroy (object->surface_data->surface);
		g_free (object->surface_data);
	}
	
	for (l = object->custom_scale_funcs; l != NULL; l = object->custom_scale_funcs) {
		KibaObjectCustomScaleData *data = l->data;
		
		g_free (data);
		object->custom_scale_funcs = g_list_remove (object->custom_scale_funcs, l->data);
	}
	g_list_free (object->custom_scale_funcs);
	
	for (l = object->custom_offset_data; l != NULL; l = object->custom_offset_data) {
		KibaObjectCustomOffsetData *data = l->data;
		
		g_free (data);
		object->custom_offset_data = g_list_remove (object->custom_offset_data, l->data);
	}
	g_list_free (object->custom_offset_data);
	
	for (l = object->custom_paint_funcs_after; l != NULL; l = object->custom_paint_funcs_after) {
		KibaObjectCustomPaintData *data = l->data;
		
		g_free (data);
		object->custom_paint_funcs_after = g_list_remove (object->custom_paint_funcs_after, l->data);
	}
	g_list_free (object->custom_paint_funcs_after);
	
	for (l = object->custom_paint_funcs_before; l != NULL; l = object->custom_paint_funcs_before) {
		KibaObjectCustomPaintData *data = l->data;
		
		g_free (data);
		object->custom_paint_funcs_before =
			g_list_remove (object->custom_paint_funcs_before, l->data);
	}
	g_list_free (object->custom_paint_funcs_before);
	
	if (GTK_WIDGET_REALIZED(widget->parent)) {
		kiba_progress_object_remove (KIBA_OBJECT(widget)->progress);
		kiba_title_object_remove (KIBA_OBJECT(widget)->title);
	}
}

/**
 * gtk_widget_realize implementation.
 **/
static void kiba_object_realize (GtkWidget *widget) {
	KibaObjectPrivate *priv = KIBA_OBJECT_GET_PRIVATE (widget);
	KibaObject *object = KIBA_OBJECT(widget);
	GdkWindowAttr attributes;
	gint attributes_mask;
	gint border_width;
	GtkRequisition requisition;
	
	GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);
	
	border_width = GTK_CONTAINER (widget)->border_width;
	
	gtk_widget_size_request (widget, &requisition);
	
	if (widget->allocation.x == -1 && widget->allocation.y == -1 &&
	    widget->allocation.width == 1 && widget->allocation.height == 1) {
		GtkAllocation allocation = { 0, 0, 200, 200 };
		
		if (requisition.width || requisition.height) {
			/* non-empty window */
			allocation.width = requisition.width;
			allocation.height = requisition.height;
		}
		gtk_widget_size_allocate (widget, &allocation);
	}
	
	widget->window = gtk_widget_get_parent_window (widget);
	g_object_ref (widget->window); /* increase the reference count of the window */
	
	attributes.x = widget->allocation.x;
	attributes.y = widget->allocation.y;
	attributes.width = widget->allocation.width;
	attributes.height = widget->allocation.height;
	attributes.event_mask = gtk_widget_get_events (widget)
				| GDK_BUTTON_MOTION_MASK
				| GDK_BUTTON_PRESS_MASK
				| GDK_BUTTON_RELEASE_MASK
				| GDK_ENTER_NOTIFY_MASK
				| GDK_LEAVE_NOTIFY_MASK;
	attributes_mask = GDK_WA_X | GDK_WA_Y;
	
	attributes.window_type = GDK_WINDOW_CHILD;
	attributes.wclass = GDK_INPUT_ONLY;
	
	priv->event_window = gdk_window_new (widget->window, &attributes, attributes_mask);
	gdk_window_set_user_data (priv->event_window, widget); /* catching the window events on the widget */
	gdk_window_register_dnd (priv->event_window); /* allowing dnd an the window */
	gdk_window_set_type_hint (priv->event_window, GDK_WINDOW_TYPE_HINT_DOCK);
	
	widget->style = gtk_style_attach (widget->style, widget->window);
	
	KIBA_OBJECT(widget)->title = kiba_title_object_add (widget);
	KIBA_OBJECT(widget)->progress = kiba_progress_object_add(widget);
	
	object->surface_data = g_new0 (CairoSurfaceData, 1);
	object->surface_data->width  = requisition.width;
	object->surface_data->height = requisition.height;
	object->surface_data->surface = NULL;
	
	priv->fade_id = g_timeout_add (20, __fade_timeout, object);
	g_timeout_add (20, __scale_timeout, object);
}

/*** Class init ***/

static void kiba_object_class_init (KibaObjectClass *class) {
	GObjectClass *gobject_class = G_OBJECT_CLASS (class);
	GtkContainerClass *container_class = GTK_CONTAINER_CLASS (class);
	GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (class);
	
	g_type_class_add_private (class, sizeof (KibaObjectPrivate));

	gobject_class->set_property = kiba_object_set_property;
	gobject_class->get_property = kiba_object_get_property;
	
	/* GtkWidget method implementations */
	widget_class->realize = kiba_object_realize;
	widget_class->unrealize = kiba_object_unrealize;
	widget_class->map = kiba_object_map;
	widget_class->unmap = kiba_object_unmap;
	widget_class->size_request = kiba_object_size_request;
	widget_class->size_allocate = kiba_object_size_allocate;
	
	/* GtkContainer method implementations */
	container_class->add = kiba_object_add;
	container_class->remove = kiba_object_remove;
	
	g_object_class_install_property 
		(gobject_class,
		 _PROP_GROUP,
		 g_param_spec_string ("group",
		 "Group",
		 "Group on the Dock",
		 NULL,
		 G_PARAM_CONSTRUCT|G_PARAM_READWRITE));
	
	/* KibaObject signals */
	kiba_object_signals[SIG_SWITCH_EVENT] =
		g_signal_new ("switch-event",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaObjectClass, switch_event),
			NULL, NULL,
			g_cclosure_marshal_VOID__VOID,
			G_TYPE_NONE, 0);
	kiba_object_signals[SIG_URGENT_TOGGLED] =
		g_signal_new ("urgent-toggled",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaObjectClass, urgent_toggled),
			NULL, NULL,
			g_cclosure_marshal_VOID__BOOLEAN,
			G_TYPE_NONE, 1,
			G_TYPE_BOOLEAN);
	kiba_object_signals[SIG_ACTIVE_TOGGLED] =
		g_signal_new ("active-toggled",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaObjectClass, active_toggled),
			NULL, NULL,
			g_cclosure_marshal_VOID__INT,
			G_TYPE_NONE, 1,
			G_TYPE_INT);
}

static void kiba_object_init (KibaObject *object) {
	static const GtkTargetEntry targets[] = {
		{ "text/plain", 0, 0 }
	};
	
	KibaObjectPrivate *priv;
	KibaDock *dock = kiba_dock_get();
	
	GTK_WIDGET_SET_FLAGS (GTK_WIDGET (object), GTK_NO_WINDOW);
	
	priv = KIBA_OBJECT_GET_PRIVATE (object);
	
	gtk_widget_add_events (GTK_WIDGET (object),
		GDK_BUTTON_PRESS_MASK |
		GDK_BUTTON_RELEASE_MASK |
		GDK_ENTER_NOTIFY_MASK |
		GDK_LEAVE_NOTIFY_MASK |
		GDK_POINTER_MOTION_MASK |
		GDK_POINTER_MOTION_HINT_MASK);
	
	object->position = KIBA_OBJECT_ON_DOCK;
	object->props = 0;
	object->win_state = KIBA_OBJECT_WINDOW_INVISIBLE;
	object->paint = kiba_object_paint;
	object->surface_data = NULL;
	object->switch_surface = NULL;
	object->alpha = 0.1f;
	object->scale_x = 1.0f;
	object->scale_y = 1.0f;
	object->alloc = (KibaAllocation) {0.0f, 0.0f, 0.0f, 0.0f};
	object->size = (GtkRequisition) {dock->d.object_size, dock->d.object_size};
	object->scroll_alpha = 1.0f;
	object->destroy_factor = 0.01f;
	object->title = NULL;
	object->progress = NULL;
	object->custom_paint_funcs_after = NULL;
	object->custom_paint_funcs_before = NULL;
	object->shadow_offset = 5;
	
	object->custom_scale_funcs = NULL;
	
	priv->gearsAlpha = 0.0f;
	priv->gearsRotateAngle = 0.0f;
	priv->workTodo = 0;
	priv->fade_alpha = 1.0f;
	priv->scale_x = 1.0f;
	priv->scale_y = 1.0f;
	
	object->shadow_color = (KibaRGBA) { 0.0f, 0.0f, 0.0f, 0.6f };
	object->shadow_affection = 45;
	
	gtk_drag_dest_set (GTK_WIDGET (object),
			GTK_DEST_DEFAULT_ALL,
			targets, G_N_ELEMENTS (targets),
			GDK_ACTION_MOVE | GDK_ACTION_COPY | GDK_ACTION_LINK);
	gtk_drag_dest_add_uri_targets (GTK_WIDGET (object));
	
	g_signal_connect(G_OBJECT(dock), "object_size_changed",
		G_CALLBACK(__dock_object_size_changed_cb), object);
}

/**
 * kiba_object_new:
 * @group: The group on the dock.
 *
 * Creates and returns a KibaObject container.
 *
 * Returns: The KibaObject instance created.
 */
GtkWidget* kiba_object_new (const char *group) {
	GtkWidget *object = g_object_new (KIBA_TYPE_OBJECT,
                                 "group", group, NULL);
	gtk_container_add (GTK_CONTAINER(kiba_dock_get()), object);
	
	return object;
}
