/*
 *  Copyright (C) 2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

/**
 * SECTION:kiba-plugin-loader
 * @short_description: Loads modules.
 * @see_also:#KibaPlugin
 * @stability: Unstable
 *
 * Loads modules.
 */

#include <string.h>

#include "kiba-plugin-loader.h"
#include "kiba-plugin.h"
#include "kiba-prefs.h"
#include "kiba-dialog.h"

G_DEFINE_TYPE (KibaPluginLoader, kiba_plugin_loader, G_TYPE_OBJECT)

enum {
	SIG_PLUGIN_INITIALLED = 0,
	SIG_PLUGIN_LOADER_LAST
};

static guint plugin_loader_signals[SIG_PLUGIN_LOADER_LAST] = { 0 };

static KibaPluginLoader *plugin_loader_instance = NULL;

/**
Helper Functions
**/

static void __plugin_toggled_cb (const char *option_id,
			       gpointer    value,
			       gpointer    data) {
	KibaPlugin *plugin = data;
	
	plugin->enabled = (gboolean)(GPOINTER_TO_INT (value));
	
	if (!plugin->loaded && plugin->enabled)
		g_type_module_use (G_TYPE_MODULE(plugin));
	else if (plugin->loaded && !plugin->enabled) {
		/* we need to destroy the plugin types before calling g_type_module_unuse,
		   to decrease the module counter, otherwise kiba_plugin_unload would not be called */
		if (plugin->vtable->unload)
			plugin->vtable->unload (plugin);
		g_type_module_unuse (G_TYPE_MODULE(plugin));
	}
}

/**
Public Functions
**/

/**
 * kiba_plugin_loader_initiall_plugins_at_path:
 * @loader: The KibaPluginLoader instance.
 * @plugin_path: The path to load plugins from.
 *
 * Loads all plugins from plugin_path.
 *
 * Returns: TRUE on success.
 */
gboolean kiba_plugin_loader_initiall_plugins_at_path (KibaPluginLoader *loader,
						      const char       *plugin_path) {
	GDir    *dir;
	GError  *error = NULL;
	char    *file_name;
	
	dir = g_dir_open (plugin_path, 0, &error);
	if (error != NULL) {
		kiba_dialog_error ("Failed to open Plugin Directory '%s': %s\n",
			plugin_path, error->message);
		g_error_free (error);
		return FALSE;
	}
	
	while ((file_name = (char *) g_dir_read_name (dir)) != NULL) {
		static PrefMember plugin_enable_options[] = {
			{ NULL, "FALSE", KIBA_BOOLEAN }
		};
		
		KibaPlugin *plugin;
		const char *name;
		
		if (!g_str_has_suffix (file_name, ".so")) continue;
		
		char library_path[strlen(plugin_path)+strlen(file_name)+1];
		sprintf (library_path, "%s/%s", plugin_path, file_name);
		
		plugin = kiba_plugin_new (library_path);
		if (!plugin) continue;
		
		name = G_TYPE_MODULE(plugin)->name;
		char enable_key[strlen(name) + 8];
		
		g_signal_emit (loader,
			plugin_loader_signals[SIG_PLUGIN_INITIALLED],
			0, plugin);
		
		sprintf (enable_key, "%s_enable", name);
		plugin_enable_options->id = g_strdup (enable_key);
		kiba_prefs_add_options (plugin_enable_options, G_N_ELEMENTS(plugin_enable_options));
		plugin->enabled = kiba_prefs_get_bool (enable_key);
		kiba_prefs_add_callback (enable_key, __plugin_toggled_cb, plugin, FALSE);
		
		if (plugin->enabled)
			g_type_module_use (G_TYPE_MODULE(plugin));
	}
	g_dir_close (dir);
	
	return TRUE;
}

/**
Class Functions
**/

static void kiba_plugin_loader_class_init (KibaPluginLoaderClass *class) {
	plugin_loader_signals[SIG_PLUGIN_INITIALLED] =
		g_signal_new ("initialled",
				G_OBJECT_CLASS_TYPE (class),
				G_SIGNAL_RUN_LAST,
				G_STRUCT_OFFSET (KibaPluginLoaderClass, initialled),
				NULL, NULL,
				g_cclosure_marshal_VOID__OBJECT,
				G_TYPE_NONE, 1,
				G_TYPE_OBJECT);
}

static void kiba_plugin_loader_init (KibaPluginLoader *plugin) {
	
}

static KibaPluginLoader* kiba_plugin_loader_new (void) {
	return g_object_new (KIBA_TYPE_PLUGIN_LOADER, NULL);
}

/**
 * kiba_plugin_loader_get:
 *
 * Creates/returns the KibaPluginLoader instance.
 *
 * Returns: The default KibaPluginLoader instance.
 */
KibaPluginLoader* kiba_plugin_loader_get (void) {
	if (!plugin_loader_instance)
		plugin_loader_instance = kiba_plugin_loader_new();
	
	return plugin_loader_instance;
}
