/*
 *  Copyright (C) 2006-2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

/**
 * SECTION:kiba-plugin
 * @short_description: Module Type.
 * @see_also:#KibaPluginLoader
 * @stability: Unstable
 *
 * Module Type.
 */

#include "../config.h"
#include <math.h>
#include <string.h>
#include <dlfcn.h>

#include "kiba-plugin.h"
#include "kiba-types.h"
#include "kiba-debug.h"

#define KIBA_PLUGIN_GET_PRIVATE(obj) \
		(G_TYPE_INSTANCE_GET_PRIVATE ((obj), KIBA_TYPE_PLUGIN, KibaPluginPrivate))

G_DEFINE_TYPE (KibaPlugin, kiba_plugin, G_TYPE_TYPE_MODULE)

typedef struct _KibaPluginPrivate KibaPluginPrivate;

struct _KibaPluginPrivate {
	void     *handle;   /* opened library, or NULL */
};

enum {
	SIG_UNLOADED = 0,
	SIG_LOADED,
	SIG_PLUGIN_LAST
};

static guint plugin_signals[SIG_PLUGIN_LAST] = { 0 };

static GList *plugins = NULL;
static GList *loaded_plugins = NULL;

/**
 * kiba_plugin_get_for_id:
 * @id: The ID of the Plugin to return.
 *
 * Returns a pointer to a loaded KibaPlugin.
 *
 * Returns: The KibaPlugin or NULL if no plugin with @id was loaded.
 */
KibaPlugin* kiba_plugin_get_for_id (const char *id) {
	GList *l;
	
	for (l = loaded_plugins; l != NULL; l = l->next) {
		KibaPlugin* plugin = l->data;
		
		if (strcmp (plugin->vtable->name, id) == 0)
			return plugin;
	}
	
	return NULL;
}

/**
 * kiba_plugin_get_loaded:
 *
 * Returns loaded KibaPlugin's.
 *
 * Returns: The loaded KibaPlugin's.
 */
GList* kiba_plugin_get_loaded (void) {
	return loaded_plugins;
}

/**
 * kiba_plugins_free:
 *
 * Unloads and destroys all plugins.
 */
void kiba_plugins_free () {
	GList *l;
	
	for (l = plugins; l != NULL; l = plugins) {
		KibaPlugin *plugin = l->data;
		
		if (!KIBA_IS_PLUGIN (plugin))
			continue;
		
		if (plugin->loaded) {
			g_signal_emit (plugin, plugin_signals[SIG_UNLOADED], 0);
			/* we need to destroy the plugin types before calling g_type_module_unuse,
			to decrease the module counter, otherwise kiba_plugin_unload would not be called */
			if (plugin->vtable->unload)
				plugin->vtable->unload (plugin);
			g_type_module_unuse (G_TYPE_MODULE(plugin));
		}
		
		g_free (plugin->lib_path);
		
		plugins = g_list_remove (plugins, plugin);
	}
	g_list_free (plugins);
	g_list_free (loaded_plugins);
}

/**
 * kiba_plugin_for_all:
 * @for_all_func: The Function to launch for every KibaPlugin.
 * @data: The UserData for the Function.
 *
 * Launches function for all loaded KibaPlugin's.
 */
void kiba_plugin_for_all (GFunc  for_all_func,
			  void  *data) {
	GList *l;
	
	for (l = loaded_plugins; l != NULL; l = l->next) {
		KibaPlugin* plugin = l->data;
		
		for_all_func (plugin, data);
	}
}

/**
Loading/Unloading
**/

gboolean kiba_plugin_register (KibaPlugin *plugin) {
	g_return_val_if_fail(KIBA_IS_PLUGIN(plugin), FALSE);

	/* If this plugin has been registered already then exit */
	if (g_list_find(plugins, plugin))
		return TRUE;
	
	if (plugin->vtable->load == NULL || plugin->vtable->unload == NULL) {
		DEBUG_ERROR ("'%s' is not loadable: "
			"The plugin does not implement all required functions\n", plugin->lib_path);
		return FALSE;
	}
	
	return TRUE;
}

/**
Class Functions
**/

static void kiba_plugin_unload (GTypeModule *module) {
	KibaPlugin *plugin;
	KibaPluginPrivate *priv;
	
	plugin = KIBA_PLUGIN(module);
	priv = KIBA_PLUGIN_GET_PRIVATE (plugin);
	
	loaded_plugins = g_list_remove (loaded_plugins, plugin);
	
	kiba_prefs_remove_options (plugin->options);
	
	dlclose (priv->handle);
	priv->handle = NULL;
	plugin->loaded = FALSE;
	
	plugin->options = NULL;
	plugin->option_elements = 0;
	plugin->vtable = NULL;
	
	DEBUG_MISC("'%s' Plugin unloaded\n", G_TYPE_MODULE(plugin)->name);
}

static gboolean kiba_plugin_load (GTypeModule *module) {
	typedef gboolean (*PluginInitFunc) (KibaPlugin*);
	
	KibaPlugin *plugin = KIBA_PLUGIN(module);
	KibaPluginPrivate *priv;
	PluginInitFunc plugin_init_func;
	void *handle = NULL;
	char *dlsym_error = NULL;
	
	plugin_init_func = NULL;
	
	/* Load The library */
	handle = dlopen (plugin->lib_path, RTLD_LAZY);
	if (!handle) {
		DEBUG_ERROR("failed to open library '%s': %s\n",
			plugin->lib_path, dlerror());
		return FALSE;
	}
	
	priv = KIBA_PLUGIN_GET_PRIVATE (plugin);
	
	dlerror();
	plugin_init_func = (PluginInitFunc) dlsym(handle, "kiba_init_plugin");
	dlsym_error = dlerror();
	if (dlsym_error) {
		DEBUG_ERROR ("'%s' is not usable because the "
					"'kiba_init_plugin' symbol could not be "
					"found.  Does the plugin call the "
					"KIBA_INIT_PLUGIN() macro?\n", plugin->lib_path);
		dlclose (handle);
		return FALSE;
	}
	
	if (!plugin_init_func (plugin)) {
		dlclose (handle);
		DEBUG_ERROR("failed to register Plugin '%s'\n", plugin->lib_path);
		return FALSE;
	}
	
	priv->handle = handle;
	
	/* Initiall Plugin Options */
	if (plugin->options)
		kiba_prefs_add_options (plugin->options, plugin->option_elements);
	
	/* Load Plugin */
	plugin->loaded = plugin->vtable->load (plugin);
	if (!plugin->loaded) {
		if (plugin->options)
			kiba_prefs_remove_options (plugin->options);
		dlclose (handle);
		DEBUG_ERROR ("'%s' is not loadable\n", plugin->lib_path);
		return FALSE;
	}
	
	loaded_plugins = g_list_append (loaded_plugins, plugin);
	
	g_signal_emit (plugin, plugin_signals[SIG_LOADED], 0);
	
	DEBUG_MISC("'%s' Plugin loaded\n", G_TYPE_MODULE(plugin)->name);
	
	return TRUE;
}

static void kiba_plugin_class_init (KibaPluginClass *class) {
	GObjectClass *object_class = G_OBJECT_CLASS (class);
	GTypeModuleClass *module_class = G_TYPE_MODULE_CLASS (class);
	
	module_class->load = kiba_plugin_load;
	module_class->unload = kiba_plugin_unload;
	
	g_type_class_add_private (object_class, sizeof (KibaPluginPrivate));
	
	plugin_signals[SIG_UNLOADED] =
		g_signal_new ("unloaded",
				G_OBJECT_CLASS_TYPE (class),
				G_SIGNAL_RUN_LAST,
				G_STRUCT_OFFSET (KibaPluginClass, unloaded),
				NULL, NULL,
				g_cclosure_marshal_VOID__VOID,
				G_TYPE_NONE, 0,
				G_TYPE_INVALID);
	plugin_signals[SIG_LOADED] =
		g_signal_new ("loaded",
				G_OBJECT_CLASS_TYPE (class),
				G_SIGNAL_RUN_LAST,
				G_STRUCT_OFFSET (KibaPluginClass, loaded),
				NULL, NULL,
				g_cclosure_marshal_VOID__VOID,
				G_TYPE_NONE, 0,
				G_TYPE_INVALID);
}

static void kiba_plugin_init (KibaPlugin *plugin) {
	KibaPluginPrivate  *priv = KIBA_PLUGIN_GET_PRIVATE (plugin);
	
	plugin->loaded = FALSE;
	priv->handle = NULL;
}

/**
 * kiba_plugin_new:
 * @library_path: The path to the Plugin Library.
 *
 * Creates and returns a new KibaPlugin instance.
 *
 * Returns: The KibaPlugin instance created.
 */
KibaPlugin* kiba_plugin_new (char *library_path) {
	KibaPlugin  *plugin;
	gchar *so_name;
	
	g_return_val_if_fail(library_path != NULL, NULL);
	
	if (!g_file_test(library_path, G_FILE_TEST_EXISTS))
		return NULL;
	
	DEBUG_MISC("Found plugin library at '%s'\n", library_path);
	
	plugin = g_object_new (KIBA_TYPE_PLUGIN, NULL);
	plugin->lib_path = g_strdup (library_path);
	
	so_name = g_strrstr (library_path, "/") + 4;
	
	char name[strlen (so_name)];
	sprintf (name, "%s", so_name);
	name[strlen (so_name)-3] = '\0';
	g_type_module_set_name (G_TYPE_MODULE(plugin), name);
	
	plugins = g_list_append (plugins, plugin);
	
	return plugin;
}
