/*
 *  Copyright (C) 2006-2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author(s) : Daniel    Bessler    <danielb@kiba-dock.org>
 *              Torbjörn Svangård  <trptorbjorn@gmail.com>
 *
*/

/**
 * SECTION:kiba-prefs
 * @short_description: Kiba-Dock preferences.
 * @see_also:#kiba-prefs-xml
 * @stability: Unstable
 *
 * Kiba-Dock preferences.
 */

#include "../config.h"

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <libintl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <glib/gstdio.h>
#include <glib/gprintf.h>
#include <gtk/gtk.h>
#include <gdk/gdkx.h>

#include <libxml/parser.h>
#include <libxml/xmlwriter.h>

#include "kiba-prefs.h"
#include "kiba-prefs-schema.h"
#include "kiba-utils.h"
#include "kiba-debug.h"
#include "prefs-dbus.h"

#define PLUGIN_SCHEMAS DATADIR"/kiba-dock/config_schemas/plugins"
#define DOCK_SCHEMAS   DATADIR"/kiba-dock/config_schemas/dock"

typedef struct {
	KibaPrefCallback func;
	gpointer data;
}KibaOptionCallback;

static KibaPrefs *kiba_prefs = NULL;

/*** Private methods ***/

static void __free_options(GList *options) {
	GList *l;
	for (l = options; l != NULL; l = options) {
		KibaOption *optionData = l->data;
		options = g_list_remove(options, optionData);
		g_free(optionData->id);
		g_free(optionData->type);
		if (optionData->value)
			g_free(optionData->value);
		g_free(optionData);
	}
	g_list_free(options);
}

static void __set_option_from_string (KibaPrefOption *option,
				    const char     *new_value) {
	switch (option->member.type) {
		case KIBA_DOUBLE:
		case KIBA_STRING: {
			if (option->value != NULL)
				g_free(option->value);
			option->value = (gpointer) (g_strdup(new_value));
			break;
		}
		case KIBA_INT: {
			option->value = GINT_TO_POINTER(atoi (new_value));
			break;
		}
		case KIBA_BOOLEAN: {
			option->value = GINT_TO_POINTER(strcmp (new_value, "TRUE") == 0);
			break;
		}
		case KIBA_LIST: {
			char *value_buf;
			
			if (option->value) {
				g_slist_free(option->value);
				option->value = NULL;
			}
			
			value_buf = (char*) new_value;
			while (strlen(value_buf)) {
				int i;
				char *value;
				
				value = g_strdup (value_buf);
				for (i = 1; i < strlen(value)-1; i++) {
					if (value[i] == ',') {
						value[i] = '\0';
						break;
					}
				}
				value_buf += strlen(value);
				if (!strlen(value_buf))
					value[strlen(value)-1] = '\0';
				
				option->value = (gpointer) g_slist_append ((GSList*) option->value, g_strdup(value+1));
				g_free(value);
			}
			break;
		}
	}
}

/** XML coinfig parsing/writing **/

/* read options from xml config file */
static GList* __get_options (KibaPrefs *settings) {
	xmlDoc *doc;
	xmlNode *root_element;
	xmlNode *settingNode;
	GList *ret;
	
	if (!g_file_test(settings->config_file, G_FILE_TEST_IS_REGULAR)) {
		return NULL;
	}
	
	/*parse the file and get the DOM */
	doc = xmlReadFile(settings->config_file, NULL, 0);
	if (doc == NULL) {
		DEBUG_ERROR("could not parse config file '%s'", settings->config_file);
		return NULL;
	}
	
	ret = NULL;
	
	root_element = xmlDocGetRootElement(doc);
	for (settingNode = root_element->children; settingNode; settingNode = settingNode->next) {
		if (settingNode->type != XML_ELEMENT_NODE) continue;
		struct _xmlAttr *attr;
		KibaOption *option;
		char *id = NULL, *type = NULL;
		
		/* get the option id */
		for (attr = settingNode->properties; attr; attr = attr->next) {
			if (strcmp(attr->name, "id") == 0)
				id = g_strdup(attr->children->content);
			else if (strcmp(attr->name, "type") == 0)
				type = g_strdup(attr->children->content);
		}
		if (!id) {
			if (type) g_free(type);
			continue;
		} else if (!type)
			continue;
		
		option = g_new0(KibaOption, 1);
		option->type = type;
		option->id = id;
		if (settingNode->children && settingNode->children->content)
			option->value = g_strdup(settingNode->children->content);
		else
			option->value = g_strdup("(null)");
		
		ret = g_list_append(ret, option);
	}
	
	xmlFreeDoc(doc);
	
	return ret;
}

/* write options to the xml config file */
void __write_options (KibaPrefs *settings) {
	xmlTextWriter *writer;
	GList *l;
	
	writer = xmlNewTextWriterFilename(settings->config_file, 0);
	
	xmlTextWriterStartDocument(writer, NULL, NULL, NULL);
		/* start root element */
		xmlTextWriterStartElement(writer, BAD_CAST "config");
		/* add options */
		for (l = settings->options; l != NULL; l = l->next) {
			KibaOption *optionData = l->data;
			xmlTextWriterWriteString(writer, "\n\t");
			/* add option node */
			xmlTextWriterStartElement(writer, BAD_CAST "option");
				xmlTextWriterWriteAttribute(writer, BAD_CAST "id", BAD_CAST optionData->id);
				xmlTextWriterWriteAttribute(writer, BAD_CAST "type", BAD_CAST optionData->type);
				if (optionData->value)
					xmlTextWriterWriteRaw(writer, BAD_CAST optionData->value);
				else
					xmlTextWriterWriteRaw(writer, BAD_CAST "(null)");
			xmlTextWriterEndElement(writer);
		}
		xmlTextWriterWriteString(writer, "\n");
	/* end root element and write the file */
	xmlTextWriterEndDocument(writer);
	
	/* clean up */
	xmlFreeTextWriter(writer);
}

/*** Public methods ***/

/**
 * kiba_prefs_edge_style_from_string:
 * @string: The corner style string.
 *
 * Returns a edge style from string.
 *
 * Returns: The corner style.
 */
gint kiba_prefs_edge_style_from_string (const char *string) {
	const struct { KibaCornerShape style; const char *string; } corner_data[]= {
		{ KIBA_ROUNDED, "rounded" },
		{ KIBA_MITERED, "mitered" },
		{ KIBA_SQUARED, "squared" },
		{ KIBA_NOTCHED, "notched" }
	};
	
	gint i;
	
	g_return_val_if_fail (string != NULL, 0);
	
	for (i = 0; i < G_N_ELEMENTS(corner_data); i++) {
		if (strcmp (corner_data[i].string, string) == 0)
			return corner_data[i].style;
	}
	
	return KIBA_SQUARED;
}

/**
 * kiba_prefs_background_style_from_string:
 * @string: The style string.
 *
 * Returns a background style from string.
 *
 * Returns: The style.
 */
gint kiba_prefs_background_style_from_string (const char *string) {
	const struct { KibaBackgroundStyle style; const char *string; } bg_data[]= {
		{ KIBA_LINEAR, "linear" },
		{ KIBA_RADIAL, "radial" },
		{ KIBA_GRADIENT, "gradient" }
	};
	
	gint i;
	
	g_return_val_if_fail (string != NULL, 0);
	
	for (i = 0; i < G_N_ELEMENTS(bg_data); i++) {
		if (strcmp (bg_data[i].string, string) == 0)
			return bg_data[i].style;
	}
	
	return KIBA_NONE;
}

/**
 * kiba_prefs_get_type_from_string:
 * @string: The type string.
 *
 * Returns the option type from string.
 *
 * Returns: The type.
 */
KibaType kiba_prefs_get_type_from_string (const char *string) {
	gint i;
	
	g_return_val_if_fail (string != NULL, 0);
	
	for (i = 0; i < G_N_ELEMENTS (kiba_type_data); i++) {
		if (strcmp (kiba_type_data[i].pref_type, string) == 0)
			return i;
	}
	
	return KIBA_INVALID;
}

/**
 * kiba_prefs_get_position_from_string:
 * @string: The position string.
 *
 * Returns a position from string.
 *
 * Returns: The position.
 */
KibaPosition kiba_prefs_get_position_from_string (const char *string) {
	gint i;
	
	g_return_val_if_fail (string != NULL, 0);
	
	for (i = 0; i < G_N_ELEMENTS (kiba_layout_data); i++) {
		if (strcmp (kiba_layout_data[i].name, string) == 0)
			return i;
	}
	
	return KIBA_POSITION_BOTTOM_EDGE;
}

/**
 * kiba_prefs_rgba_from_string:
 * @string: The color string.
 *
 * Returns a Color from string.
 *
 * Returns: The color.
 */
KibaRGBA kiba_prefs_rgba_from_string (const char *string) {
	KibaRGBA rgba;
	int red, green, blue, opacity;
	
	sscanf (string, "#%02X%02X%02X%02X", &red, &green, &blue, &opacity);
	rgba.r = (float) red   / 256.0;
	rgba.g = (float) green / 256.0;
	rgba.b = (float) blue  / 256.0;
	rgba.a = (float) opacity  / 256.0;
	
	return rgba;
}

/**
 * kiba_prefs_get_member_for_id:
 * @options:  An array of PrefMember's.
 * @array_elemets: Number of Elements in the Option Struct.
 * @setting_id: The Pref Name.
 *
 * Gets a initialled PrefMember for a Pref Name.
 *
 * Returns: The PrefMember associated with the Pref Name.
 */
const PrefMember* kiba_prefs_get_member_for_id (const PrefMember options[],
						int              array_elemets,
						const char      *setting_id) {
	int i = 0;
	
	for (i = 0; i < array_elemets; i++) {
		if (strcmp (setting_id, options[i].id) == 0)
			return &options[i];
	}
	
	return NULL;
}

/**
 * kiba_prefs_get_pointer:
 * @id: The option id.
 *
 * Returns value pointer from the database.
 *
 * Returns: The pointer to the database entry, or NULL.
 */
gpointer kiba_prefs_get_pointer (const char *id) {
	KibaPrefs *prefs;
	GList     *l;
	
	g_return_val_if_fail (id != NULL, NULL);
	
	prefs = kiba_prefs_get ();
	
	for (l = prefs->groups; l != NULL; l = l->next) {
		KibaPrefGroup *group = l->data;
		GList         *li;
		
		for (li = group->options; li != NULL; li = li->next) {
			KibaPrefOption *option = li->data;
			
			if (option->member.type == KIBA_INT && strcmp (option->member.id, id) == 0)
				return option->value;
		}
	}
	
	DEBUG_WARNING ("Pointer for '%s' not found", id);
	
	return NULL;
}

/**
 * kiba_prefs_get_int:
 * @id: The Option Name to watch for.
 *
 * Gets Int Value from Kiba Preferences Database.
 *
 * Returns: The Value, or -1 if none Setting was found.
 */
gint kiba_prefs_get_int (const char *id) {
	KibaPrefs *prefs;
	GList     *l;
	
	g_return_val_if_fail (id != NULL, 0);
	
	prefs = kiba_prefs_get ();
	
	for (l = prefs->groups; l != NULL; l = l->next) {
		KibaPrefGroup *group = l->data;
		GList         *li;
		
		for (li = group->options; li != NULL; li = li->next) {
			KibaPrefOption *option = li->data;
			
			if (option->member.type == KIBA_INT && strcmp (option->member.id, id) == 0)
				return GPOINTER_TO_INT (option->value);
		}
	}
	
	DEBUG_WARNING ("Int Option '%s' not found", id);
	
	return -1;
}

/**
 * kiba_prefs_get_bool:
 * @id: The Option Name to watch for.
 *
 * Gets Bool Value from Kiba Preferences Database.
 *
 * Returns: The Value, or FALSE if none Setting was found.
 */
gboolean kiba_prefs_get_bool (const char *id) {
	KibaPrefs *prefs;
	GList     *l;
	
	g_return_val_if_fail (id != NULL, 0);
	
	prefs = kiba_prefs_get ();
	
	for (l = prefs->groups; l != NULL; l = l->next) {
		KibaPrefGroup *group = l->data;
		GList         *li;
		
		for (li = group->options; li != NULL; li = li->next) {
			KibaPrefOption *option = li->data;
			
			if (option->member.type == KIBA_BOOLEAN &&
			    strcmp (option->member.id, id) == 0)
				return (gboolean)(GPOINTER_TO_INT (option->value));
		}
	}
	
	DEBUG_WARNING ("Bool Option '%s' not found", id);
	
	return FALSE;
}

/**
 * kiba_prefs_get_float:
 * @id: The Option Name to watch for.
 *
 * Gets Double Value from Kiba Preferences Database.
 *
 * Returns: The Value, or 0.0 if none Setting was found.
 */
float kiba_prefs_get_float (const char *id) {
	KibaPrefs *prefs;
	GList     *l;
	
	g_return_val_if_fail (id != NULL, 0.0);
	
	prefs = kiba_prefs_get ();
	
	for (l = prefs->groups; l != NULL; l = l->next) {
		KibaPrefGroup *group = l->data;
		GList *li;
		
		for (li = group->options; li != NULL; li = li->next) {
			KibaPrefOption *option = li->data;
			
			if (option->member.type == KIBA_DOUBLE &&
			    strcmp (option->member.id, id) == 0)
				return g_ascii_strtod((char*) (option->value), NULL);
		}
	}
	
	DEBUG_WARNING ("Float Option '%s' not found", id);
	
	return 0.0;
}

/**
 * kiba_prefs_get_string:
 * @id: The Option Name to watch for.
 *
 * Gets String Value from Kiba Preferences Database.
 *
 * Returns: The Value, or NULL if none Setting was found.
 */
gchar* kiba_prefs_get_string (const char *id) {
	KibaPrefs *prefs;
	GList     *l;
	
	g_return_val_if_fail (id != NULL, NULL);
	
	prefs = kiba_prefs_get ();
	
	for (l = prefs->groups; l != NULL; l = l->next) {
		KibaPrefGroup *group = l->data;
		GList         *li;
		
		for (li = group->options; li != NULL; li = li->next) {
			KibaPrefOption *option = li->data;
			
			if (option->member.type == KIBA_STRING &&
			    strcmp (option->member.id, id) == 0)
				return (char*) option->value;
		}
	}
	
	DEBUG_WARNING ("String Option '%s' not found", id);
	
	return NULL;
}

/**
 * kiba_prefs_get_list:
 * @id: The Option Name to watch for.
 *
 * Gets GList Value from Kiba Preferences Database.
 *
 * Returns: GList, or NULL if none Setting was found.
 */
GSList* kiba_prefs_get_list (const char *id) {
	KibaPrefs *prefs;
	GList     *l;
	
	g_return_val_if_fail (id != NULL, NULL);
	
	prefs = kiba_prefs_get ();
	
	for (l = prefs->groups; l != NULL; l = l->next) {
		KibaPrefGroup *group = l->data;
		GList         *li;
		
		for (li = group->options; li != NULL; li = li->next) {
			KibaPrefOption *option = li->data;
			
			if (option->member.type == KIBA_LIST &&
			    strcmp (option->member.id, id) == 0)
				return (GSList*) option->value;
		}
	}
	
	DEBUG_WARNING ("List Option '%s' not found", id);
	
	return NULL;
}

void kiba_prefs_remove_list_item (const char *list_id,
				  char       *rem_item) {
	KibaPrefs *prefs;
	GList     *l;
	
	g_return_if_fail (rem_item != NULL);
	
	prefs = kiba_prefs_get ();
	
	for (l = prefs->groups; l != NULL; l = l->next) {
		KibaPrefGroup *group = l->data;
		GList *li;
		
		for (li = group->options; li != NULL; li = li->next) {
			KibaPrefOption *option = li->data;
			
			if (option->member.type == KIBA_LIST &&
			    strcmp (option->member.id, list_id) == 0) {
				option->value = g_list_remove (option->value, rem_item);
				return;
			}
		}
	}
}

/**
 * kiba_prefs_append_to_list:
 * @list_id: The list Name.
 * @new_item: The New Item to append.
 *
 * Appends Item to list and write List-Option to Config File.
 */
void kiba_prefs_append_to_list (const char *list_name,
				const char *new_item) {
	GSList  *list = kiba_prefs_get_list (list_name);
	GSList  *l;
	char    *value = g_strdup_printf ("[%s", new_item);
	char    *val_buf;
	
	for (l = list; l != NULL; l = l->next) {
		char *item = l->data;
		
		if (strcmp (item, new_item) == 0)
			return;
	}
	
	for (l = list; l != NULL; l = l->next) {
		char *item = l->data;
		char *buf  = NULL;
		
		if (!item || !g_strrstr (item, ";"))
			continue;
		
		buf = g_strdup (value);
		g_free (value);
		value = NULL;
		value = g_strdup_printf ("%s,%s", buf, item);
	}
	
	val_buf = value;
	value = g_strdup_printf ("%s]", val_buf);
	
	kiba_prefs_write_to_config_file (list_name, "list", value);
	g_free (value);
	g_free (val_buf);
}

/**
 * kiba_prefs_remove_from_list:
 * @list_id: The list Name.
 * @remove_item: The Item to remove.
 *
 * Removes Item from list and write List-Option to Config File.
 */
void kiba_prefs_remove_from_list (const char *list_name,
				  const char *remove_item) {
	GSList *l;
	char *value = g_strdup ("[");
	char *val_buf;
	GSList *list;
	
	list = kiba_prefs_get_list (list_name);
	
	for (l = g_slist_copy(list); l != NULL; l = l->next) {
		char *item = l->data;
		
		if (strcmp (item, remove_item) == 0) {
			list = g_slist_remove (list, item);
			kiba_prefs_remove_list_item (list_name, item);
		}
	}
	g_slist_free (l);
	
	for (l = list; l != NULL; l = l->next) {
		char *item = l->data;
		char *buf  = NULL;
		
		if (!g_strrstr (item, ";"))
			continue;
		
		buf = g_strdup (value);
		g_free (value);
		value = NULL;
		
		if (strcmp (buf, "[") != 0)
			value = g_strdup_printf ("%s,%s", buf, item);
		else
			value = g_strdup_printf ("%s%s", buf, item);
	}
	
	val_buf = value;
	value = g_strdup_printf ("%s]", val_buf);
	
	kiba_prefs_write_to_config_file (list_name, "list", value);
	g_free (val_buf);
	g_free (value);
}

GSList* kiba_prefs_get_list_from_string (const char *data) {
	GSList *ret = NULL;
	char   *value, *buf;
	
	value = g_strndup (data, strlen (data) - 1);
	buf   = g_strdup (g_strrstr (value, ","));
	if (!buf)
		ret = g_slist_append (ret, g_strdup (value+1));
	
	while (buf) {
		char *written, *missing;
	
		ret = g_slist_append (ret, g_strdup (buf+1));
		
		written = g_strrstr (value, buf);
		if (!written)
			break;
		
		missing = g_strndup (value, strlen (value) - strlen (written));
		g_free (buf);
		buf = NULL;
		buf = g_strdup (g_strrstr (missing, ","));
		
		if (!buf)
			ret = g_slist_append (ret, g_strdup (missing+1));
		
		g_free (missing);
	}
	
	g_free (value);
	g_free (buf);
	
	return ret;
}

char* kiba_prefs_list_to_string(GList *list) {
	GList *l;
	char *ret;
	char *buf;
	
	if (!list) return g_strdup("[]");
	
	ret = g_strdup_printf("[%s", (char*) list->data);
	for (l = list->next; l != NULL; l = l->next) {
		buf = ret;
		ret = g_strdup_printf("%s,%s", buf, (char*) l->data);
		g_free (buf);
	}
	buf = ret;
	ret = g_strdup_printf("%s]", buf);
	g_free (buf);
	
	return ret;
}

/**
 * kiba_prefs_set_option:
 * @id: The Pref Name.
 * @type: The Pref Type.
 * @value: The Pref Value.
 *
 * Sets New Value for Preferences Member and call the changed_callback.
 */
void kiba_prefs_set_option (const char *id,
			    const char *type,
			    const char *value) {
	GList     *l;
	KibaPrefs *prefs;
	
	g_return_if_fail (id != NULL);
	g_return_if_fail (type != NULL);
	g_return_if_fail (value != NULL);
	
	prefs = kiba_prefs_get();
	
	for (l = prefs->options; l != NULL; l = l->next) {
		KibaOption *optionData = l->data;
		if (strcmp(id, optionData->id) == 0 &&
		    strcmp(type, optionData->type) == 0) {
			g_free(optionData->value);
			optionData->value = g_strdup(value);
			break;
		}
	}
	
	for (l = prefs->groups; l != NULL; l = l->next) {
		KibaPrefGroup *group = l->data;
		GList         *li;
		
		for (li = group->options; li != NULL; li = li->next) {
			KibaPrefOption *option= li->data;
			
			if (strcmp (option->member.id, id) == 0) {
				GList *callbacks;
				
				__set_option_from_string (option, value);
				
				callbacks = option->callbacks;
				while (callbacks) {
					KibaOptionCallback *callback = callbacks->data;
					callbacks = callbacks->next;
					
					callback->func (id, option->value, callback->data);
				}

				return;
			}
		}
	}
}

/**
 * kiba_prefs_add_callback:
 * @id: The setting id.
 * @callback: The setting changed callback.
 * @data: The user data.
 * @launch: TRUE to call the callback on adding, otherwise FALSE is returned.
 *
 * Adds value changed callback for option.
 */
void kiba_prefs_add_callback (const char       *id,
			      KibaPrefCallback  callback,
			      gpointer          data,
			      gboolean          launch) {
	KibaPrefs *prefs;
	GList     *l;
	
	g_return_if_fail (id != NULL);
	
	prefs = kiba_prefs_get ();
	
	for (l = prefs->groups; l != NULL; l = l->next) {
		KibaPrefGroup *group = l->data;
		GList         *li;
		
		for (li = group->options; li != NULL; li = li->next) {
			KibaPrefOption *option = li->data;
			
			if (strcmp (option->member.id, id) == 0) {
				KibaOptionCallback *cb;
				
				cb = g_new0 (KibaOptionCallback, 1);
				cb->func = callback;
				cb->data = data;
				
				if (launch)
					callback (id, option->value, data);
				option->callbacks = g_list_append (option->callbacks, cb);
				return;
			}
		}
	}
}

/**
 * kiba_prefs_remove_callback_by_data:
 * @data: The user data.
 *
 * Removes value changed callback.
 */
void kiba_prefs_remove_callback_by_data (gpointer data) {
	KibaPrefs *prefs;
	GList     *l;
	
	g_return_if_fail (data != NULL);
	
	prefs = kiba_prefs_get ();
	
	for (l = prefs->groups; l != NULL; l = l->next) {
		KibaPrefGroup *group = l->data;
		GList *li;
		
		for (li = group->options; li != NULL; li = li->next) {
			KibaPrefOption *option = li->data;
			GList *cbs;
			
			for (cbs = option->callbacks; cbs != NULL; cbs = cbs->next) {
				KibaOptionCallback *cb = cbs->data;
				
				if (cb->data == data) {
					if (cbs->prev)
						cbs = cbs->prev;
					else
						cbs = option->callbacks;
					option->callbacks = g_list_remove (option->callbacks, cb);
					g_free (cb);
					cb = NULL;
				}
			}
		}
	}
}

/**
 * kiba_prefs_remove_options:
 * @options: An array of PrefMember's to remove.
 *
 * Removes previously added Options from the Preferences Database.
 */
void kiba_prefs_remove_options (PrefMember *options) {
	KibaPrefs *prefs;
	GList *l;
	
	g_return_if_fail (options != NULL);
	
	prefs = kiba_prefs_get ();
	
	for (l = prefs->groups; l != NULL; l = l->next) {
		KibaPrefGroup *group = l->data;
	
		if (options == group->member) {
			GList *li;
			
			prefs->groups = g_list_remove (prefs->groups, group);
			
			for (li = group->options; li != NULL; li = group->options) {
				KibaPrefOption *option = li->data;
				
				group->options = g_list_remove (group->options, option);
				switch (option->member.type) {
					case KIBA_DOUBLE:
					case KIBA_STRING:
						g_free(option->value);
						break;
					case KIBA_LIST:
						g_slist_free(option->value);
						break;
					case KIBA_INT:
					case KIBA_BOOLEAN:
					default:
						break;
				}
				
				g_free (option);
				option = NULL;
			}
			g_list_free (group->options);
			group->options = NULL;
			g_free (group);
			group = NULL;
			break;
		}
	}
}

/**
 * kiba_prefs_add_options:
 * @options: An array of PrefMember's.
 * @option_elements: Element number of Option Struct.
 *
 * Initialls Options in the Preferences Database.
 */
void kiba_prefs_add_options (PrefMember *options,
			     gint        option_elements) {
	KibaPrefs *settings;
	KibaPrefGroup *group;
	GList *l;
	gint i = 0;
	
	g_return_if_fail (options != NULL);
	
	settings = kiba_prefs_get ();
	
	group = g_new0 (KibaPrefGroup, 1);
	group->member = options;
	group->member_elements = option_elements;
	group->options = NULL;
	
	for (i = 0; i < option_elements; i++) {
		KibaPrefOption *option;
		KibaOption *optionData = NULL;
		
		for (l = settings->options; l != NULL; l = l->next) {
			KibaOption *foo = l->data;
			if (strcmp(foo->id, options[i].id) != 0)
				continue;
			if (kiba_prefs_get_type_from_string(foo->type) != options[i].type)
				continue;
			optionData = foo;
		}
		if (!optionData) {
			DEBUG_WARNING("Invalid Option '%s' added to preferences database!\n", options[i].id);
			continue;
		}
		
		option = g_new0 (KibaPrefOption, 1);
		option->member = options[i];
		option->callbacks = NULL;
		option->value = NULL;
		__set_option_from_string(option, optionData->value);
		
		group->options = g_list_append (group->options, option);
	}
	
	settings->groups = g_list_append (settings->groups, group);
}

/**
 * kiba_prefs_write_to_config_file:
 * @id: The Pref Name.
 * @type: The Pref Type.
 * @value: The Pref Value.
 *
 * Writes Option to Config File.
 *
 * Returns: TRUE on success, otherwise FALSE.
 */
gboolean kiba_prefs_write_to_config_file (const char *id,
					  const char *type,
					  const char *value) {
	KibaPrefs *settings = kiba_prefs_get();
	GList *l;
	gboolean isNewOption;
	
	g_return_if_fail(id != NULL);
	g_return_if_fail(type != NULL);
	g_return_if_fail(value != NULL);
	
	/* get options */
	isNewOption = TRUE;
	for (l = settings->options; l != NULL; l = l->next) {
		KibaOption *optionData = l->data;
		if (strcmp(id, optionData->id) == 0 &&
		    strcmp(type, optionData->type) == 0) {
			g_free(optionData->value);
			optionData->value = g_strdup(value);
			isNewOption = FALSE;
			break;
		}
	}
	if (isNewOption) {
		KibaOption *newOption = g_new0(KibaOption, 1);
		newOption->id = g_strdup(id);
		newOption->type = g_strdup(type);
		newOption->value = g_strdup(value);
		settings->options = g_list_append(settings->options, newOption);
	}
	
	__write_options(settings);
}

/**
 * kiba_prefs_get_option_from_file:
 *
 * Gets String Value from Kiba Config File.
 * @id: The Option Name to watch for.
 * @type: The Type of the Option.
 *
 * Returns: The Value as String, or NULL if none Setting was found.
 */
char* kiba_prefs_get_option_from_file (const char *id,
				       const char *type) {
	KibaPrefs *prefs = kiba_prefs_get ();
	GList *l;
	char *ret = NULL;
	
	for (l = prefs->options; l != NULL; l = l->next) {
		KibaOption *optionData = l->data;
		if (strcmp (optionData->id, id) == 0 && strcmp (optionData->type, type) == 0) {
			ret = g_strdup(optionData->value);
			break;
		}
	}
	
	return ret;
}

void kiba_prefs_update_config_file(KibaPrefs *prefs) {
	KibaSchema *schema;
	GList *schemaOptions, *l;
	gboolean configChanged = FALSE;
	
	schema = kiba_prefs_schema_get();
	schemaOptions = kiba_prefs_schema_get_options(schema);
	
	//TODO: remove options from config that do not appear in the schema.
	// but currently there are some used options that do not appear in the schema (for example 'dock_groups')
	
	/* add new options from schema */
	for (l = schemaOptions; l != NULL; l = l->next) {
		KibaOption *schemaOption = l->data;
		gboolean newOption = TRUE;
		GList *li;
		
		/* check if $schemaOption is allready added to the config */
		for (li = prefs->options; li != NULL; li = li->next) {
			KibaOption *option = li->data;
			
			if (strcmp (schemaOption->id, option->id) == 0 &&
			    strcmp (schemaOption->type, option->type) == 0) {
				newOption = FALSE;
				break;
			}
		}
		if (!newOption) continue;
		
		/* adding $schemaOption to the config file */
		KibaOption *configOption = g_new0(KibaOption, 1);
		configOption->id = g_strdup(schemaOption->id);
		configOption->type = g_strdup(schemaOption->type);
		configOption->value = g_strdup(schemaOption->value);
		prefs->options = g_list_append(prefs->options, configOption);
		
		configChanged = TRUE;
	}
	
	if (configChanged)
		__write_options(prefs);
	
	kiba_prefs_schema_fini(schema);
}

/*** Initi/Fini ***/

/**
 * kiba_prefs_fini:
 *
 * Frees allocated Memory.
 */
void kiba_prefs_fini (void) {
	g_free (kiba_prefs->config_file);
	__free_options(kiba_prefs->options);
	g_free (kiba_prefs);
}

/**
 * kiba_prefs_init:
 * @config_file: The config file to use.
 *
 * Initialls the preferences database.
 */
void kiba_prefs_init (const char *config_file) {
	if (kiba_prefs)
		return;
	
	kiba_prefs = g_new0 (KibaPrefs, 1);
	kiba_prefs->config_file = g_strdup (config_file);
	kiba_prefs->groups = NULL;
	kiba_prefs->options = __get_options(kiba_prefs);
}

/**
 * kiba_prefs_get:
 *
 * Creates and returns a new KibaPrefs object.
 *
 * Returns: The KibaPrefs object created.
 */
KibaPrefs* kiba_prefs_get () {
	if (!kiba_prefs) {
		return NULL;
	}
	
	return kiba_prefs;
}
