/*
 *  Copyright (C) 2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

/**
 * SECTION:kiba-progress
 * @short_description: Shows progress on #KibaObject's.
 * @see_also:#KibaObject
 * @stability: Unstable
 *
 * Shows progress on #KibaObject's.
 */

#include "../config.h"
#include <math.h>

#include "kiba-win.h"
#include "kiba-dock.h"
#include "kiba-progress.h"
#include "kiba-macros.h"

#define PROGRESS_SIZE         20
#define PROGRESS_STEP         2
#define PROGRESS_STOP_VALUE   100
#define PROGRESS_BORDER_SIZE  2

#define KIBA_PROGRESS_GET_PRIVATE(obj) \
		(G_TYPE_INSTANCE_GET_PRIVATE ((obj), KIBA_TYPE_PROGRESS, KibaProgressPrivate))

G_DEFINE_TYPE (KibaProgress, kiba_progress, G_TYPE_OBJECT)

typedef struct _KibaProgressPrivate KibaProgressPrivate;

struct _KibaProgressPrivate {
	GList *progress_objects;
};

static void kiba_progress_object_set_position (KibaProgressObject *object) {
	KibaDock *dock;
	const KibaLayoutData *d;
	GdkRectangle new_position;
	gint fX, fY;
	
	dock = KIBA_DOCK(KIBA_WIN_CHILD(object->progress->win));
	if (!dock)
		d = &kiba_layout_data[KIBA_POSITION_BOTTOM_EDGE];
	else
		d = &kiba_layout_data[dock->d.position];
	
	fX = object->widget->allocation.x;
	fY = object->widget->allocation.y;
	
	if (d->orientation == GTK_ORIENTATION_HORIZONTAL) {
		new_position.x = fX +
			(object->widget->allocation.width - object->area.width) / 2 +
			object->widget->allocation.width/2;
		new_position.y = (d->valignment/2) * (fY +
			(object->widget->allocation.height - object->area.height) / 2 +
			object->widget->allocation.height/2);
	}
	else {
		new_position.y = fY +
			(object->widget->allocation.height - object->area.height) / 2 +
			object->widget->allocation.height/2;
		new_position.x = (d->halignment/2) * (fX +
			(object->widget->allocation.width - object->area.width) / 2 +
			object->widget->allocation.width/2);
	}
	
	if (new_position.x != object->area.x || new_position.y != object->area.y) {
		if (KIBA_OBJECT(object->widget)->position == KIBA_OBJECT_ON_DOCK)
			kiba_win_invalidate_area (object->progress->win, object->area);
		
		object->area.x = new_position.x;
		object->area.y = new_position.y;
		
		if (KIBA_OBJECT(object->widget)->position == KIBA_OBJECT_ON_DOCK)
			kiba_win_invalidate_area (object->progress->win, object->area);
	}
}

static gint animate_value_changes (gpointer data) {
	KibaProgressObject *object = data;
	
	if (object->value > object->current_value) {
		object->current_value += PROGRESS_STEP;
		if (object->value < object->current_value)
			object->current_value = object->value;
	}
	else if (object->value < object->current_value) {
		object->current_value -= PROGRESS_STEP;
		if (object->value > object->current_value)
			object->current_value = object->value;
	}
	
	if (KIBA_OBJECT(object->widget)->position == KIBA_OBJECT_ON_DOCK)
		kiba_win_invalidate_area (object->progress->win, object->area);
	
	if (object->current_value >= PROGRESS_STOP_VALUE) {
		object->current_value = 0;
		object->value = 0;
	}
	
	if (object->current_value == object->value) {
		object->value_change_id = 0;
		return FALSE;
	}
	
	return TRUE;
}

/**
Rendering
**/

static void kiba_progress_draw_surface (cairo_t          *cr,
					cairo_surface_t  *surface,
					gint              width,
					gint              height,
					void             *data) {
	KibaProgress *progress = data;
	gint radius;
	
	g_return_if_fail (KIBA_IS_PROGRESS(progress));
	
	radius = PROGRESS_SIZE/2;
	
	cairo_arc (cr, radius, radius, radius - (PROGRESS_BORDER_SIZE/2), 0, 2*M_PI);
	cairo_close_path (cr);
	cairo_set_source_rgba (cr,
		progress->style.bg_rgba.r,
		progress->style.bg_rgba.g,
		progress->style.bg_rgba.b,
		progress->style.bg_rgba.a);
	cairo_fill_preserve (cr);
	
	cairo_set_source_rgba (cr,
		progress->style.rgba.r,
		progress->style.rgba.g,
		progress->style.rgba.b, 1);
	cairo_set_line_width (cr, PROGRESS_BORDER_SIZE);
	cairo_stroke (cr);
}

static void kiba_progress_update_surface (KibaProgress *progress) {
	cairo_t *cr;
	
	cr = kiba_win_cairo_create (KIBA_WIN(progress->win));
	cairo_update_surface_at_size (cr, &progress->surface,
		PROGRESS_SIZE, PROGRESS_SIZE,
		kiba_progress_draw_surface, progress);
	cairo_destroy (cr);
}

static void kiba_progress_object_draw (KibaProgressObject *object,
				       cairo_t            *cr) {
	gint radius;
	
	if (KIBA_OBJECT(object->widget)->position != KIBA_OBJECT_ON_DOCK)
		return;
	
	radius = PROGRESS_SIZE/2;
	
	cairo_translate (cr, object->area.x, object->area.y);
	
	cairo_set_source_surface (cr, object->progress->surface, 0, 0);
	cairo_paint (cr);
	
	//FIXME: this causes heavy cpu load, i guess cairo does not like arcs much
	{
		cairo_move_to (cr, radius, radius);
		cairo_arc (cr,
			radius,
			radius,
			radius,
			M_PI/2,
			(90 + 360*object->current_value/PROGRESS_STOP_VALUE) * (M_PI/180));
			
		cairo_set_source_rgba (cr,
			object->progress->style.rgba.r,
			object->progress->style.rgba.g,
			object->progress->style.rgba.b,
			object->progress->style.rgba.a);
		cairo_fill (cr);
	}
}

/**
 * kiba_progress_draw:
 * @progress: The KibaProgress Instance.
 * @cr: The Drawing Context.
 *
 * Draw the progress Surface-Buffer to a Drawing context.
 */
void kiba_progress_draw (KibaProgress *progress,
			 cairo_t      *cr) {
	KibaProgressPrivate *priv;
	GList *l;
	
	g_return_if_fail (KIBA_IS_PROGRESS(progress));
	
	if (!GET_FLAG(progress->flags, KIBA_PROGRESS_DRAWABLE))
		return;
	
#ifdef USE_AKAMARU
	if (kiba_dock_akamaru_enabled(KIBA_DOCK(gtk_bin_get_child(GTK_BIN(progress->win)))))
		return;
#endif
	
	priv = KIBA_PROGRESS_GET_PRIVATE (progress);
	
	for (l = priv->progress_objects; l != NULL; l = l->next) {
		KibaProgressObject *object = l->data;
		
		if (kiba_progress_object_is_running(object)) {
			cairo_save (cr);
				kiba_progress_object_draw (object, cr);
			cairo_restore (cr);
		}
	}
}

/**
Public Functions
**/

/**
 * kiba_progress_set_style:
 * @progress: The KibaProgress instance.
 * @style: The style to set.
 *
 * Sets progress background style.
 */
void kiba_progress_set_style (KibaProgress      *progress,
			      KibaProgressStyle  style) {
	KibaProgressPrivate *priv;
	GList *l;
	
	g_return_if_fail (KIBA_IS_PROGRESS(progress));
	
	priv = KIBA_PROGRESS_GET_PRIVATE (progress);
	progress->style = style;
	
	kiba_progress_update_surface (progress);
	
	for (l = priv->progress_objects; l != NULL; l = l->next) {
		KibaProgressObject *object = l->data;
		
		if (kiba_progress_object_is_running(object) &&
		    KIBA_OBJECT(object->widget)->position == KIBA_OBJECT_ON_DOCK)
			kiba_win_invalidate_area (progress->win, object->area);
	}
}

/**
 * kiba_progress_set_flag:
 * @progress: The KibaProgress instance.
 * @flag: The flag to set.
 *
 * Sets progress flag.
 */
void kiba_progress_set_flag (KibaProgress *progress,
			     gint          flag) {
	KibaProgressPrivate *priv;
	GList *l;
	
	g_return_if_fail (KIBA_IS_PROGRESS(progress));
	
	if (GET_FLAG(progress->flags, flag))
		return;
	
	priv = KIBA_PROGRESS_GET_PRIVATE (progress);
	
	SET_FLAG(progress->flags, flag);
	
	switch (flag) {
		case KIBA_PROGRESS_DRAWABLE:
			for (l = priv->progress_objects; l != NULL; l = l->next) {
				KibaProgressObject *object = l->data;
				
				if (kiba_progress_object_is_running(object) &&
				    KIBA_OBJECT(object->widget)->position == KIBA_OBJECT_ON_DOCK)
					kiba_win_invalidate_area (progress->win, object->area);
			}
			break;
		default:
			break;
	}
}

/**
 * kiba_progress_unset_flag:
 * @progress: The KibaProgress instance.
 * @flag: The flag to unset.
 *
 * Unsets progress flag.
 */
void kiba_progress_unset_flag (KibaProgress *progress,
			       gint          flag) {
	KibaProgressPrivate *priv;
	GList *l;
	
	g_return_if_fail (KIBA_IS_PROGRESS(progress));
	
	if (!GET_FLAG(progress->flags, flag))
		return;
	
	priv = KIBA_PROGRESS_GET_PRIVATE (progress);
	
	DEL_FLAG(progress->flags, flag);
	
	switch (flag) {
		case KIBA_PROGRESS_DRAWABLE:
			for (l = priv->progress_objects; l != NULL; l = l->next) {
				KibaProgressObject *object = l->data;
				
				if (kiba_progress_object_is_running(object) &&
				    KIBA_OBJECT(object->widget)->position == KIBA_OBJECT_ON_DOCK)
					kiba_win_invalidate_area (progress->win, object->area);
			}
			break;
		default:
			break;
	}
}

/**
KibaProgressObject
**/

static void progress_object_parent_set_cb (GtkWidget *widget,
					   GtkObject *old_parent,
					   gpointer   user_data) {
	KibaProgressObject *object = user_data;
	KibaProgressPrivate *priv;
	
	if (!widget->parent)
		return;
	
	object->progress = KIBA_WIN(widget->parent->parent)->progress;
	priv = KIBA_PROGRESS_GET_PRIVATE(object->progress);
	priv->progress_objects = g_list_append (priv->progress_objects, object);
}

static void progress_object_size_allocate_cb (GtkWidget     *widget,
					      GtkAllocation *allocation,
					      gpointer       user_data) {
	KibaProgressObject *object = user_data;
	
	kiba_progress_object_set_position (object);
}

/**
 * kiba_progress_object_add:
 * @widget: The KibaProgressObject instance.
 *
 * Adds progress object.
 *
 * Returns: The KibaProgressObject created.
 */
KibaProgressObject* kiba_progress_object_add (GtkWidget *widget) {
	KibaProgressObject *object;
	KibaProgressPrivate *priv;
	
	g_return_val_if_fail (widget != NULL, NULL);
	
	object = g_new0 (KibaProgressObject, 1);
	object->widget = widget;
	object->area.width  = PROGRESS_SIZE;
	object->area.height = PROGRESS_SIZE;
	object->value = 0;
	object->current_value = 0;
	object->value_change_id = 0;
	
	if (widget->parent) {
		object->progress = KIBA_WIN(widget->parent->parent)->progress;
		priv = KIBA_PROGRESS_GET_PRIVATE(object->progress);
		priv->progress_objects = g_list_append (priv->progress_objects, object);
	}
	else
		object->progress = NULL;
	
	kiba_progress_object_set_position (object);
	
	g_signal_connect (G_OBJECT(widget), "parent-set",
		G_CALLBACK(progress_object_parent_set_cb), object);
	g_signal_connect (G_OBJECT(widget), "size-allocate",
		G_CALLBACK(progress_object_size_allocate_cb), object);
	g_signal_connect_after (G_OBJECT(widget), "size-allocate",
		G_CALLBACK(progress_object_size_allocate_cb), object);
	
	return object;
}

/**
 * kiba_progress_object_remove:
 * @object: The KibaProgressObject instance.
 *
 * Removes progress object.
 */
void kiba_progress_object_remove (KibaProgressObject *object) {
	KibaProgressPrivate *priv;
	GList *l;
	
	g_return_if_fail (object != NULL);
	
	if (KIBA_IS_PROGRESS(object->progress)) {
		priv = KIBA_PROGRESS_GET_PRIVATE(object->progress);
		
		for (l = priv->progress_objects; l != NULL; l = l->next) {
			if (l->data != object)
				continue;
			
			priv->progress_objects = g_list_remove (priv->progress_objects, l->data);
			break;
		}
	}
	
	g_signal_handlers_disconnect_by_func (G_OBJECT(object->widget),
		G_CALLBACK(progress_object_parent_set_cb), object);
	g_signal_handlers_disconnect_by_func (G_OBJECT(object->widget),
		G_CALLBACK(progress_object_size_allocate_cb), object);
	
	g_free (object);
}

/**
 * kiba_progress_object_is_running:
 * @object: The KibaProgressObject instance.
 *
 * Return if a progress for a KibaProgress Instance is running.
 *
 * Returns: TRUE  if a progress is running, otherwise FALSE is returned.
 */
gboolean kiba_progress_object_is_running (KibaProgressObject *object) {
	g_return_val_if_fail (object != NULL, FALSE);
	
	return object->current_value != 0;
}

/**
 * kiba_progress_object_set_info:
 * @object: The KibaProgressObject instance.
 * @value: The new progress value (must be <= 100).
 * @message: The new Progress message or NULL.
 *
 * Set progress info for a KibaProgressObject instance.
 */
void kiba_progress_object_set_value (KibaProgressObject *object,
				    gint                value) {
	g_return_if_fail (object != NULL);
	g_return_if_fail (value <= PROGRESS_STOP_VALUE);
	
	object->value = value;
	
	if (!object->value_change_id)
		object->value_change_id = g_timeout_add (20, animate_value_changes, object);
}

/**
Class Functions
**/

static void kiba_progress_class_init (KibaProgressClass *class) {
	g_type_class_add_private (class, sizeof (KibaProgressPrivate));
}

static void kiba_progress_init (KibaProgress *progress) {
	KibaProgressPrivate *priv;
	
	priv = KIBA_PROGRESS_GET_PRIVATE(progress);
	
	SET_FLAG(progress->flags, KIBA_PROGRESS_DRAWABLE);
	
	progress->surface = NULL;
	progress->style = (KibaProgressStyle) {
		{ 0.0f, 0.0f, 0.0f, 0.7f }, { 1.0f, 1.0f, 1.0f, 0.8f }
	};
	
	priv->progress_objects = NULL;
}

/**
 * kiba_progress_new:
 * @win: The KibaWin instancet.
 *
 * Create a KibaProgress Instance.
 *
 * Returns: The KibaProgress Instance created.
 */
KibaProgress* kiba_progress_new (GtkWidget *win) {
	KibaProgress *progress = g_object_new (KIBA_TYPE_PROGRESS, NULL);
	
	progress->win = win;
	
	return progress;
}

/**
 * kiba_progress_destroy:
 * @progress: The KibaProgress instance.
 *
 * Free allocated memory for a KibaProgress instance.
 */
void kiba_progress_destroy (KibaProgress *progress) {
	KibaProgressPrivate *priv;
	KibaProgressObject *object;
	GList *l;
	
	g_return_if_fail (KIBA_IS_PROGRESS(progress));
	
	priv = KIBA_PROGRESS_GET_PRIVATE (progress);
	
	for (l = priv->progress_objects; l != NULL; l = priv->progress_objects) {
		object = l->data;
		priv->progress_objects = g_list_remove (priv->progress_objects, l->data);
	}
	g_list_free (priv->progress_objects);
	if (progress->surface)
		cairo_surface_destroy (progress->surface);
	g_object_unref (progress);
	progress = NULL;
}
