/*
 *  Copyright (C) 2008 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#define SN_API_NOT_YET_FROZEN
#include <libsn/sn.h>

#include "marshallers.h"
#include "startup-notification.h"

#define KIBA_SN_GET_PRIVATE(obj) \
		(G_TYPE_INSTANCE_GET_PRIVATE ((obj), KIBA_TYPE_SN, KibaSnPrivate))

G_DEFINE_TYPE (KibaSn, kiba_sn, G_TYPE_OBJECT)

typedef struct _KibaSnPrivate KibaSnPrivate;

typedef struct {
	SnLauncherContext *context;
	GObject *object;
	const char *id;
}KibaSnLaunchData;

struct _KibaSnPrivate {
	SnDisplay *display;
	SnMonitorContext *monitorContext;
	GList *launchData;
	gint errorTrapDepth;
	gint processEventCount;
};

enum {
	SIG_LAUNCH_STARTED = 0,
	SIG_LAUNCH_FINISHED,
	SIG_KIBA_SN_LAST
};

static KibaSn *instance = NULL;

static guint kiba_sn_signals[SIG_KIBA_SN_LAST] = { 0 };

static pthread_mutex_t event_mutex = PTHREAD_MUTEX_INITIALIZER;

/*** Private Methods ***/

static void error_trap_push (SnDisplay *display,
                             Display   *xdisplay) {
	KibaSnPrivate *priv = KIBA_SN_GET_PRIVATE(instance);
	priv->errorTrapDepth++;
}

static void error_trap_pop (SnDisplay *display,
                            Display   *xdisplay) {
	KibaSnPrivate *priv = KIBA_SN_GET_PRIVATE(instance);
	if (priv->errorTrapDepth == 0) {
		fprintf (stderr, "Error trap underflow!\n");
		return;
	}
	
	XSync (xdisplay, False); /* get all errors out of the queue */
	priv->errorTrapDepth--;
}

static gpointer eventThreadStart (gpointer data) {
	KibaSn *sn = data;
	KibaSnPrivate *priv = KIBA_SN_GET_PRIVATE(sn);
	Display *xdisplay = sn_display_get_x_display(priv->display);
	while (priv->processEventCount > 0) {
		XEvent xevent;
		XNextEvent (xdisplay, &xevent);
		/* process the event in libsn, this may invoke monitor_event_func */
		sn_display_process_event (priv->display, &xevent);
	}
}

static void kiba_sn_watch_events(KibaSn *sn) {
	KibaSnPrivate *priv = KIBA_SN_GET_PRIVATE(sn);
	priv->processEventCount++;
	/* start the thread, if this was the first request for event watching */
	if (priv->processEventCount == 1) {
		pthread_attr_t thread_attributes;
		pthread_t event_thread;
		
		pthread_attr_init (&thread_attributes);
		pthread_attr_setdetachstate (&thread_attributes,
						PTHREAD_CREATE_DETACHED);
#ifdef _POSIX_THREAD_ATTR_STACKSIZE
		pthread_attr_setstacksize (&thread_attributes, 128*1024);
#endif
		pthread_create (&event_thread, &thread_attributes,
				eventThreadStart, sn);
	}
}

static void kiba_sn_unwatch_events(KibaSn *sn) {
	KibaSnPrivate *priv = KIBA_SN_GET_PRIVATE(sn);
	priv->processEventCount--;
}

static void monitor_event_func (SnMonitorEvent *event,
                                void            *user_data) {
	KibaSn *sn = instance;
	KibaSnPrivate *priv = KIBA_SN_GET_PRIVATE(sn);
	SnMonitorContext *context;
	SnStartupSequence *sequence;
	SnLauncherContext *launcherContext;
	KibaSnLaunchData *data;
	const char *id;
	GList *l;
	
	context = sn_monitor_event_get_context (event);
	sequence = sn_monitor_event_get_startup_sequence (event);
	
	/* get the launch data associated to this event */
	id = sn_startup_sequence_get_id (sequence);
	for (l = priv->launchData; l != NULL; l =l->next) {
		KibaSnLaunchData *launcherData = l->data;
		if (strcmp (id, launcherData->id) == 0) {
			data = launcherData;
			break;
		}
	}
	if (data == NULL)
		return;
	
	switch (sn_monitor_event_get_type (event)) {
		case SN_MONITOR_EVENT_INITIATED: {
			g_signal_emit (sn,
				kiba_sn_signals[SIG_LAUNCH_STARTED],
				0, data->object);
			break;
		}
		case SN_MONITOR_EVENT_COMPLETED:
		case SN_MONITOR_EVENT_CANCELED: {
			gboolean complete = (sn_monitor_event_get_type (event) == SN_MONITOR_EVENT_COMPLETED);
			g_signal_emit (sn,
				kiba_sn_signals[SIG_LAUNCH_FINISHED],
				0, complete, data->object);
			
			/* cleaning up */
			pthread_mutex_lock (&event_mutex);
			kiba_sn_unwatch_events(sn);
			sn_launcher_context_unref (data->context);
			priv->launchData = g_list_remove(priv->launchData, data);
			g_free(data);
			pthread_mutex_unlock (&event_mutex);
			break;
		}
		default:
			break;
	}
}

/*** Public Methods ***/

/**
 * kiba_startup_notification_launch:
 * @sn: The KibaSn Instance.
 * @command: The command to monitor.
 * @object: Custom user data to pass to the signal handler.
 *
 * Monitors the next time @command is launched.
 * The KibaSn instance will emit a "launch-finished" signal then.
 */
void kiba_sn_launch (KibaSn *sn,
                     const char *command,
                     GObject *object) {
	KibaSnPrivate *priv = KIBA_SN_GET_PRIVATE(sn);
	Display *xdisplay = sn_display_get_x_display(priv->display);
	KibaSnLaunchData *data = g_new0(KibaSnLaunchData, 1);
//TODO: timeout
	
	data->object = object;
	data->context = sn_launcher_context_new (priv->display, DefaultScreen (xdisplay));
	sn_launcher_context_initiate (data->context,
									"kiba-launcher",
									command,
									CurrentTime); /*<< TODO: get the real time, CurrentTime is allways 0 here */
	/* remember the id */
	data->id = sn_launcher_context_get_startup_id(data->context);
	priv->launchData = g_list_append(priv->launchData, data);
	kiba_sn_watch_events(sn);
}

/**
Class Functions
**/

static void kiba_sn_class_init (KibaSnClass *class) {
	g_type_class_add_private (class, sizeof (KibaSnPrivate));
	
	kiba_sn_signals[SIG_LAUNCH_STARTED] =
		g_signal_new ("launch-started",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaSnClass, launch_started),
			NULL, NULL,
			g_cclosure_marshal_VOID__OBJECT,
			G_TYPE_NONE, 1,
			G_TYPE_OBJECT);
	kiba_sn_signals[SIG_LAUNCH_FINISHED] =
		g_signal_new ("launch-finished",
			G_OBJECT_CLASS_TYPE (class),
			G_SIGNAL_RUN_LAST,
			G_STRUCT_OFFSET (KibaSnClass, launch_finished),
			NULL, NULL,
			_kiba_marshal_VOID__BOOLEAN_OBJECT,
			G_TYPE_NONE, 2,
			G_TYPE_BOOLEAN, G_TYPE_OBJECT);
}

static void kiba_sn_init (KibaSn *sn) {
	KibaSnPrivate *priv = KIBA_SN_GET_PRIVATE(sn);
	Display *xdisplay;
	SnLauncherContext *context;
	SnMonitorContext *monitorContext;
	
	priv->errorTrapDepth = 0;
	priv->processEventCount = 0;
	priv->display = NULL;
	priv->launchData = NULL;
	
	xdisplay = XOpenDisplay (NULL);
	if (xdisplay == NULL) {
		fprintf (stderr, "Could not open display\n");
		return;
	}
	
	if (getenv ("LIBSN_SYNC") != NULL)
		XSynchronize (xdisplay, True);
	
	priv->display = sn_display_new (xdisplay,
						error_trap_push,
						error_trap_pop);
	
	/* We have to select for property events on at least one
	* root window (but not all as INITIATE messages go to
	* all root windows)
	*/
	XSelectInput (xdisplay, DefaultRootWindow (xdisplay),
					PropertyChangeMask);
	priv->monitorContext = sn_monitor_context_new (priv->display, DefaultScreen (xdisplay),
										monitor_event_func,
										NULL, NULL);
}

/**
 * kiba_sn_get:
 *
 * Create a KibaSn Instance.
 *
 * Returns: The KibaSn Instance created.
 */
KibaSn* kiba_sn_get (void) {
	if (instance == NULL)
		instance = g_object_new (KIBA_TYPE_SN, NULL);
	return instance;
}

void kiba_sn_destroy (KibaSn *sn) {
	KibaSnPrivate *priv = KIBA_SN_GET_PRIVATE(sn);
	GList *l;
	
	if (sn = instance)
		instance = NULL;
	
	for (l = priv->launchData; l != NULL; l = priv->launchData) {
		KibaSnLaunchData *data = l->data;
		sn_launcher_context_unref (data->context);
		priv->launchData = g_list_remove(priv->launchData, data);
		g_free(data);
	}
	g_list_free(priv->launchData);
	
	sn_monitor_context_unref (priv->monitorContext);
	sn_display_unref (priv->display);
	
	g_object_unref(G_OBJECT(sn));
}
