/*
 *  Copyright (C) 2006-2007 Daniel Bessler <danielb@kiba-dock.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 *
 *  Author : Daniel Bessler <danielb@kiba-dock.org>
*/

/**
 * SECTION:kiba-title
 * @short_description: Title for #KibaObject's.
 * @see_also:#KibaObject
 * @stability: Unstable
 *
 * Title for #KibaObject's.
 */

#include <math.h>
#include <string.h>
#include <glib.h>

#include "kiba-win.h"
#include "kiba-title.h"
#include "kiba-cairo-utils.h"
#include "kiba-debug.h"
#include "kiba-macros.h"

#define BACKGROUND_STEPSIZE   0.02f
#define FADE_FPS              50
#define TITLE_OBJECT_DISTANCE 9

#define KIBA_TITLE_GET_PRIVATE(obj) \
	(G_TYPE_INSTANCE_GET_PRIVATE ((obj), KIBA_TYPE_TITLE, KibaTitlePrivate))

G_DEFINE_TYPE (KibaTitle, kiba_title, G_TYPE_OBJECT)

typedef struct {
	GList *title_objects;
	KibaTitleObject *active_object;
	
	GdkRectangle area;
	float fade_alpha;
	gint scale_id;
}KibaTitlePrivate;

/**
Helper Functions
**/

static void kiba_title_set_position (KibaTitle *title) {
	KibaTitlePrivate *priv;
	KibaDock *dock;
	GtkWidget *object;
	const KibaLayoutData *d;
	float fX, fY;
	
	priv = KIBA_TITLE_GET_PRIVATE(title);
	dock = KIBA_DOCK(KIBA_WIN_CHILD(title->win));
	d = &kiba_layout_data[dock->d.position];
	object = priv->active_object->widget;
	
	fX = KIBA_OBJECT(object)->alloc.x;
	fY = KIBA_OBJECT(object)->alloc.y;
	
	priv->area.x = fX;
	priv->area.y = fY;
	if (d->orientation == GTK_ORIENTATION_VERTICAL) {
		priv->area.x += (d->halignment == 0 ? object->allocation.width :
			-priv->area.width) - (d->halignment - 1) * TITLE_OBJECT_DISTANCE;
		priv->area.y += -(priv->area.height - object->allocation.height)/2;
	}
	else {
		priv->area.x += -(priv->area.width - object->allocation.width)/2;
		priv->area.y += (d->valignment == 0 ? object->allocation.height :
			-priv->area.height) - (d->valignment - 1) * TITLE_OBJECT_DISTANCE;
	}
	
	priv->area.x = MAX(0, priv->area.x);
	priv->area.x = MIN(title->win->allocation.width - priv->area.width, priv->area.x);
	priv->area.y = MAX(0, priv->area.y);
	priv->area.y = MIN(title->win->allocation.height - priv->area.height, priv->area.y);
}

#ifdef PANGO_VERSION_CHECK
#if PANGO_VERSION_CHECK(1,16,0)
static void kiba_title_object_set_pango_gravity (KibaTitleObject *object) {
	PangoContext *context;
	KibaDock *dock;
	const KibaLayoutData *d;
	
	g_return_if_fail (PANGO_IS_LAYOUT(object->pango_layout));
	
	dock = KIBA_DOCK(KIBA_WIN_CHILD(object->title->win));
	d = &kiba_layout_data[dock->d.position];
	
	if (d->orientation != GTK_ORIENTATION_VERTICAL)
		return;
	
	context = pango_layout_get_context (object->pango_layout);
	pango_context_set_base_gravity (context,
		d->halignment == 0 ? PANGO_GRAVITY_EAST : PANGO_GRAVITY_WEST);
	pango_context_set_gravity_hint (context, PANGO_GRAVITY_HINT_STRONG);
}
#endif
#endif

static gint kiba_title_fade_timeout (gpointer data) {
	KibaTitle *title = data;
	KibaTitlePrivate *priv;
	KibaDock *dock;
	
	priv = KIBA_TITLE_GET_PRIVATE(title);
	dock = KIBA_DOCK(KIBA_WIN_CHILD(title->win));
	
	kiba_win_invalidate_area (title->win, priv->area);
	
	if (title->state == TITLE_HIDING) {
		priv->fade_alpha -= 1.0f / title->hide_delay;
		
		if (priv->fade_alpha <= 0.0f) {
			title->state = TITLE_HIDDEN;
			priv->active_object = NULL;
			priv->scale_id = 0;
			return FALSE;
		}
	}
	else {
		priv->scale_id = 0;
		return FALSE;
	}
	
	return TRUE;
}

/**
Rendering
**/

/**
 * kiba_title_draw:
 *
 * Renders the title to the drawing context.
 */
void kiba_title_draw (KibaTitle            *title,
		      const KibaLayoutData *d,
		      cairo_t              *cr) {
	KibaTitlePrivate *priv;
	KibaTitleObject *object;
	float fade_factor;
	gint i;
	
	g_return_if_fail (KIBA_IS_TITLE(title));
	
	priv = KIBA_TITLE_GET_PRIVATE(title);
	object = priv->active_object;
	
	if (!object || !object->pango_layout || !GET_FLAG(title->flags, KIBA_TITLE_DRAWABLE))
		return;
	
	fade_factor = priv->fade_alpha < 0.8f ? priv->fade_alpha + 0.2f : 1.0f;
	
	cairo_translate (cr, priv->area.x, priv->area.y);
	
	if (GET_FLAG(title->flags, KIBA_TITLE_SCALING)) {
		cairo_translate (cr, -((fade_factor-1.0f) * priv->area.width) / 2.0,
				-((fade_factor-1.0f) * priv->area.height) / 2.0);
		cairo_scale (cr, fade_factor, fade_factor);
	}
	
	/* draw background */
	cairo_save (cr);
		cairo_scale (cr, (float) priv->area.width / object->surface_data.width,
			 (float) priv->area.height / object->surface_data.height);
		
		cairo_set_source_surface (cr, object->surface_data.surface, 0.0, 0.0);
		cairo_paint_with_alpha (cr, priv->fade_alpha);
	cairo_restore (cr);
	
	/* rotate the title to be vertical */
	if (d->orientation == GTK_ORIENTATION_VERTICAL) {
		cairo_rotate (cr, (d->halignment + 1.0f)*M_PI/2.0f);
		if (d->halignment == 2)
			cairo_translate (cr, -priv->area.height, 0);
		else
			cairo_translate (cr, 0, -priv->area.width);
	}
	
	cairo_translate (cr, title->style.border_width + 5,
			     title->style.border_width + 1);
	
	/* layout the shadow */
	/* TODO: maybe scaling the context would be better then showing the layout 4 times, but it seems to be ok here */
	cairo_set_source_rgba (cr,
			priv->fade_alpha*title->style.shadow_rgba.r,
			priv->fade_alpha*title->style.shadow_rgba.g,
			priv->fade_alpha*title->style.shadow_rgba.b,
			priv->fade_alpha*title->style.shadow_rgba.a);
	for (i = 0; i < 4; i++) {
		cairo_save (cr);
			cairo_translate (cr, (1 - (i&2)), (1 - 2*(i&1)));
			pango_cairo_show_layout (cr, object->pango_layout);
		cairo_restore (cr);
	}
		
	cairo_set_source_rgba (cr, priv->fade_alpha*title->style.rgba.r,
		priv->fade_alpha*title->style.rgba.g,
		priv->fade_alpha*title->style.rgba.b,
		priv->fade_alpha*title->style.rgba.a);
	cairo_save (cr);
		pango_cairo_show_layout (cr, object->pango_layout);
	cairo_restore (cr);
	
	/* draw highlight */
/*
	if (GET_FLAG(title->flags, KIBA_TITLE_HIGHLIGHTING)) {
		cairo_set_operator (cr, CAIRO_OPERATOR_ATOP);
		
		pango_cairo_layout_path (cr, title->pango_layout);
		pat = cairo_pattern_create_linear (title_allocation.width * title_highlight_offset, 0.0,
				title_allocation.width * (title_highlight_offset + 0.7), 0.0);
		cairo_pattern_add_color_stop_rgba (pat, 0.9, 1.0, 1.0, 1.0, 0.0);
		cairo_pattern_add_color_stop_rgba (pat, 0.3,
			title_fade_alpha*title->style.highlight_rgba.r,
			title_fade_alpha*title->style.highlight_rgba.g,
			title_fade_alpha*title->style.highlight_rgba.b,
			title_fade_alpha*title->style.highlight_rgba.a);
		cairo_pattern_add_color_stop_rgba (pat, 0.0, 1.0, 1.0, 1.0, 0.0);
		cairo_set_source (cr, pat);
		cairo_fill_preserve (cr);
		cairo_pattern_destroy (pat);
	}
*/
}

static void kiba_title_object_draw_surface (cairo_t          *cr,
					    cairo_surface_t  *surface,
					    gint              width,
					    gint              height,
					    void             *data) {
	KibaTitleObject *object = data;
	cairo_pattern_t *pattern = NULL;
	
	cairo_paint_shape (cr,
		object->title->style.border_style,
		object->title->style.border_style,
		object->title->style.border_style,
		object->title->style.border_style,
		object->title->style.border_width/2,
		object->title->style.border_width/2,
		object->surface_data.width - object->title->style.border_width/2,
		object->surface_data.height - object->title->style.border_width/2,
		5);
	
	if (object->title->style.background.style == KIBA_GRADIENT)
		pattern = cairo_get_gradient_pattern (object->title->style.border_width/2,
				object->title->style.border_width/2,
				object->surface_data.width - object->title->style.border_width/2,
				object->surface_data.height - object->title->style.border_width/2,
				object->title->style.gradient, object->title->style.numColors);
	else
		pattern = cairo_get_png_pattern (object->title->style.background.file, 1.0);
	if (pattern == NULL)
		return;
	
	cairo_set_source (cr, pattern);
	cairo_fill_preserve (cr);
	cairo_pattern_destroy (pattern);
	
	cairo_set_source_rgba (cr,
		object->title->style.border_rgba.r,
		object->title->style.border_rgba.g,
		object->title->style.border_rgba.b,
		object->title->style.border_rgba.a);
	cairo_set_line_width (cr, object->title->style.border_width);
	cairo_stroke (cr);
}

static void kiba_title_object_update_surface (KibaTitleObject *object) {
	KibaDock *dock;
	cairo_t *cr;
	const KibaLayoutData *d;
	
	dock = KIBA_DOCK(KIBA_WIN_CHILD(object->title->win));
	d = &kiba_layout_data[dock->d.position];
	
	struct {
		gint  *w,  *h;
	} magic[2] = {
		{ &object->surface_data.width,  &object->surface_data.height },
		{ &object->surface_data.height, &object->surface_data.width }
	};
	
	if (!PANGO_IS_LAYOUT(object->pango_layout)) {
		*magic[d->orientation].w = 100;
		*magic[d->orientation].h = 40;
	}
	else {
		PangoRectangle  log;
		
		pango_layout_get_pixel_extents (object->pango_layout, NULL, &log);
		
		*magic[d->orientation].w = log.width + 10 + 2*object->title->style.border_width;
		*magic[d->orientation].h = log.height + 2 + 2*object->title->style.border_width;
	}
	
	cr = kiba_win_cairo_create (KIBA_WIN(object->title->win));
	cairo_update_surface (cr, &object->surface_data, kiba_title_object_draw_surface, object);
	cairo_destroy (cr);
}

/**
KibaTitleObjects
**/

static void title_object_parent_set_cb (GtkWidget *widget,
					GtkObject *old_parent,
					gpointer   user_data) {
	KibaTitleObject *object = user_data;
	KibaTitlePrivate *priv;
	
	if (!widget->parent)
		return;
	
	object->title = KIBA_WIN(widget->parent->parent)->title;
	priv = KIBA_TITLE_GET_PRIVATE(object->title);
	priv->title_objects = g_list_append (priv->title_objects, object);
}

static void title_object_size_allocate_cb (GtkWidget     *widget,
					      GtkAllocation *allocation,
					      gpointer       user_data) {
	KibaTitleObject *object = user_data;
	KibaTitlePrivate *priv;
	GdkRectangle old_area;
	
	if (!kiba_title_object_is_shown(object))
		return;
	
	priv = KIBA_TITLE_GET_PRIVATE(object->title);
	
	old_area = priv->area;
	
	kiba_title_set_position (object->title);
	
	if (old_area.x == priv->area.x && old_area.y == priv->area.y)
		return;
	
	kiba_win_invalidate_area (object->title->win, old_area);
	kiba_win_invalidate_area (object->title->win, priv->area);
}

/**
 * kiba_title_object_add:
 *
 * Adds title object.
 */
KibaTitleObject* kiba_title_object_add (GtkWidget *widget) {
	KibaTitleObject *object;
	
	g_return_val_if_fail (widget != NULL, NULL);
	
	object = g_new0 (KibaTitleObject, 1);
	object->widget = widget;
	object->pango_layout = NULL;
	object->text = NULL;
	object->surface_data.surface = NULL;
	object->surface_data.width = 40;
	object->surface_data.height = 40;
	
	if (widget->parent) {
		KibaTitlePrivate *priv;
		object->title = KIBA_WIN(widget->parent->parent)->title;
		priv = KIBA_TITLE_GET_PRIVATE(object->title);
		priv->title_objects = g_list_append (priv->title_objects, object);
	}
	else
		object->title = NULL;
	
	g_signal_connect (G_OBJECT(widget), "parent-set",
		G_CALLBACK(title_object_parent_set_cb), object);
	g_signal_connect_after (G_OBJECT(widget), "size-allocate",
		G_CALLBACK(title_object_size_allocate_cb), object);
	
	return object;
}

/**
 * kiba_title_object_remove:
 *
 * Removes title object.
 */
void kiba_title_object_remove (KibaTitleObject *object) {
	KibaTitlePrivate *priv;
	GList *l;
	
	g_return_if_fail (object != NULL);
	
	priv = KIBA_TITLE_GET_PRIVATE(object->title);
	
	for (l = priv->title_objects; l != NULL; l = l->next) {
		if (l->data == object) {
			priv->title_objects = g_list_remove (priv->title_objects, l->data);
			break;
		}
	}
	
	g_signal_handlers_disconnect_by_func (G_OBJECT(object->widget),
		  G_CALLBACK(title_object_size_allocate_cb), object);
	g_signal_handlers_disconnect_by_func (G_OBJECT(object->widget),
		G_CALLBACK(title_object_parent_set_cb), object);
	
	if (object->surface_data.surface)
		cairo_surface_destroy (object->surface_data.surface);
	if (object->pango_layout)
		g_object_unref (object->pango_layout);
	if (object->text)
		g_free (object->text);
	g_free (object);
}

/**
 * kiba_title_object_show:
 *
 * Shows title object.
 */
void kiba_title_object_show (KibaTitleObject *object) {
	KibaTitlePrivate *priv;
	KibaDock *dock;
	const KibaLayoutData *d;
	
	g_return_if_fail (object != NULL);
	
	priv = KIBA_TITLE_GET_PRIVATE(object->title);
	dock = KIBA_DOCK(KIBA_WIN_CHILD(object->title->win));
	d = &kiba_layout_data[dock->d.position];
	
	if (kiba_title_is_shown(object->title))
		kiba_win_invalidate_area (object->title->win, priv->area);
	
	priv->active_object = object;
	object->title->state = TITLE_SHOWING;
	
	if (priv->scale_id) {
		g_source_remove(priv->scale_id);
		priv->scale_id = 0;
	}
	priv->fade_alpha = 1.0f;
	object->title->state = TITLE_SHOWN;
	
	if (object->pango_layout) {
		PangoRectangle  log;
		
		pango_layout_get_pixel_extents (object->pango_layout, NULL, &log);
		
		if (d->orientation == GTK_ORIENTATION_VERTICAL) {
			priv->area.height  = log.width + 10 + 2*object->title->style.border_width;
			priv->area.width = log.height + 2 + 2*object->title->style.border_width;
		}
		else {
			priv->area.width  = log.width + 10 + 2*object->title->style.border_width;
			priv->area.height = log.height + 2 + 2*object->title->style.border_width;
		}
	}
	
	kiba_title_set_position (object->title);
	kiba_win_invalidate_area (object->title->win, priv->area);
}

/**
 * kiba_title_object_hide:
 *
 * Hides title object.
 */
void kiba_title_object_hide (KibaTitleObject *object) {
	KibaTitlePrivate *priv;
	
	g_return_if_fail (object != NULL);
	
	priv = KIBA_TITLE_GET_PRIVATE(object->title);
	
	if (object != priv->active_object) {
		return;
	}
	
	if (!priv->scale_id)
		priv->scale_id = g_timeout_add (1000/FADE_FPS, kiba_title_fade_timeout, object->title);
	object->title->state = TITLE_HIDING;
}

/**
 * kiba_title_object_is_shown:
 *
 * Returns wether the title is shown.
 */
gboolean kiba_title_object_is_shown (KibaTitleObject *object) {
	KibaTitlePrivate *priv;
	
	g_return_val_if_fail (object != NULL, FALSE);
	
	priv = KIBA_TITLE_GET_PRIVATE(object->title);
	
	return (object == priv->active_object);
}

/**
 * kiba_title_object_set_text:
 *
 * Sets title text.
 */
void kiba_title_object_set_text (KibaTitleObject *object,
				 const char *text) {
	KibaTitlePrivate *priv;
	KibaDock *dock;
	const KibaLayoutData *d;
	gint diff;
	cairo_t *cr = NULL;
	PangoRectangle log;
	
	g_return_if_fail (object != NULL);
	g_return_if_fail (text != NULL);
	
	priv = KIBA_TITLE_GET_PRIVATE(object->title);
	dock = KIBA_DOCK(KIBA_WIN_CHILD(object->title->win));
	d = &kiba_layout_data[dock->d.position];
	
	if (object->text) {
		if (strcmp (object->text, text) == 0)
			return;
		diff = strlen (object->text) - strlen (text);
		g_free (object->text);
		object->text = NULL;
	}
	else
		diff = -5;
	object->text = g_strdup (text);
	
	
	cr = kiba_win_cairo_create (KIBA_WIN(object->title->win));
	cairo_update_pango_layout (object->text,
		object->title->font,
		cr,
		&object->pango_layout);
	cairo_destroy (cr);
#ifdef PANGO_VERSION_CHECK
#if PANGO_VERSION_CHECK(1,16,0)
	if (PANGO_IS_LAYOUT(object->pango_layout))
		kiba_title_object_set_pango_gravity (object);
#endif
#endif
	if (diff < -4 || diff > 4)
		kiba_title_object_update_surface (object);

	if (!kiba_title_object_is_shown (object))
		return;
	
	kiba_win_invalidate_area (object->title->win, priv->area);
	
	pango_layout_get_pixel_extents (object->pango_layout, NULL, &log);
	
	if (d->orientation == GTK_ORIENTATION_VERTICAL) {
		priv->area.height  = log.width + 10 + 2*object->title->style.border_width;
		priv->area.width = log.height + 2 + 2*object->title->style.border_width;
	}
	else {
		priv->area.width  = log.width + 10 + 2*object->title->style.border_width;
		priv->area.height = log.height + 2 + 2*object->title->style.border_width;
	}
	kiba_title_set_position (object->title);
	
	kiba_win_invalidate_area (object->title->win, priv->area);
}

/**
 * kiba_title_object_get_text:
 *
 * Returns title text.
 */
char* kiba_title_object_get_text (KibaTitleObject *object) {
	g_return_val_if_fail (object != NULL, NULL);
	
	return object->text;
}

/**
Public Functions
**/

/**
 * kiba_title_set_font:
 *
 * Sets title font.
 */
void kiba_title_set_font (KibaTitle  *title,
			  char       *font) {
	KibaTitlePrivate *priv;
	KibaDock *dock;
	GList *l;
	cairo_t *cr = NULL;
	PangoRectangle log;
	const KibaLayoutData *d;
	
	g_return_if_fail (KIBA_IS_TITLE(title));
	
	priv = KIBA_TITLE_GET_PRIVATE(title);
	dock = KIBA_DOCK(KIBA_WIN_CHILD(title->win));
	d = &kiba_layout_data[dock->d.position];
	
	title->font = font;
	
	/* update pango layouts for the new font */
	cr = kiba_win_cairo_create (KIBA_WIN(title->win));
	for (l = priv->title_objects; l != NULL; l = l->next) {
		KibaTitleObject *object = l->data;
		
		cairo_update_pango_layout (object->text,
			title->font,
			cr,
			&object->pango_layout);
#ifdef PANGO_VERSION_CHECK
#if PANGO_VERSION_CHECK(1,16,0)
		if (PANGO_IS_LAYOUT(object->pango_layout))
			kiba_title_object_set_pango_gravity (object);
#endif
#endif
	}
	cairo_destroy (cr);
	
	if (!priv->active_object)
		return;
	
	kiba_win_invalidate_area (title->win, priv->area);
	
	pango_layout_get_pixel_extents (priv->active_object->pango_layout, NULL, &log);
		
	if (d->orientation == GTK_ORIENTATION_VERTICAL) {
		priv->area.height = log.width + 10 + 2*title->style.border_width;
		priv->area.width = log.height + 2 + 2*title->style.border_width;
	}
	else {
		priv->area.width  = log.width + 10 + 2*title->style.border_width;
		priv->area.height = log.height + 2 + 2*title->style.border_width;
	}
	kiba_title_set_position (title);
	
	kiba_win_invalidate_area (title->win, priv->area);
}

/**
 * kiba_title_set_flag:
 *
 * Sets title flag.
 */
void kiba_title_set_flag (KibaTitle *title,
			  gint       flag) {
	g_return_if_fail (KIBA_IS_TITLE(title));
	
	if (GET_FLAG(title->flags, flag))
		return;
	
	SET_FLAG(title->flags, flag);
	
	switch (flag) {
		default:
			break;
	}
}

/**
 * kiba_title_unset_flag:
 *
 * Unsets title flag.
 */
void kiba_title_unset_flag (KibaTitle *title,
			    gint       flag) {
	g_return_if_fail (KIBA_IS_TITLE(title));
	
	if (!GET_FLAG(title->flags, flag))
		return;
	
	DEL_FLAG(title->flags, flag);
	
	switch (flag) {
		default:
			break;
	}
}

/**
 * kiba_title_set_style:
 *
 * Sets the background style.
 */
void kiba_title_set_style (KibaTitle      *title,
			   KibaTitleStyle  style) {
	KibaTitlePrivate *priv;
	GList *l;
	
	g_return_if_fail (KIBA_IS_TITLE(title));
	
	priv = KIBA_TITLE_GET_PRIVATE(title);
	
	title->style = style;
	
	for (l = priv->title_objects; l != NULL; l = l->next) {
		KibaTitleObject *object = l->data;
		
		kiba_title_object_update_surface (object);
	}
	
	if (kiba_title_is_shown (title))
		kiba_win_invalidate_area (title->win, priv->area);
}

/**
 * kiba_title_is_shown:
 *
 * Returns wether the title is shown.
 */
gboolean kiba_title_is_shown (KibaTitle *title) {
	KibaTitlePrivate *priv;
	
	g_return_val_if_fail (KIBA_IS_TITLE(title), FALSE);
	
	priv = KIBA_TITLE_GET_PRIVATE(title);
	
	return priv->active_object != NULL;
}

/**
Callbacks
**/

static void kiba_title_orientation_cb (KibaDock      *dock,
				       KibaPosition   position,
				       gpointer       user_data) {
	KibaTitle *title = user_data;
	KibaTitlePrivate *priv;
	GList *l;
	cairo_t *cr = NULL;
	PangoRectangle log;
	const KibaLayoutData *d;
	
	priv = KIBA_TITLE_GET_PRIVATE(title);
	d = &kiba_layout_data[dock->d.position];
	
	/* update pango layouts for the new font */
	cr = kiba_win_cairo_create (KIBA_WIN(title->win));
	for (l = priv->title_objects; l != NULL; l = l->next) {
		KibaTitleObject *object = l->data;
		
		if (!object->text)
			continue;
		
		cairo_update_pango_layout (object->text,
			title->font,
			cr,
			&object->pango_layout);
#ifdef PANGO_VERSION_CHECK
#if PANGO_VERSION_CHECK(1,16,0)
		if (d->orientation == GTK_ORIENTATION_VERTICAL) {
			PangoContext *context;
			
			context = pango_layout_get_context (object->pango_layout);
			pango_context_set_base_gravity (context,
				d->halignment == 0 ? PANGO_GRAVITY_EAST : PANGO_GRAVITY_WEST);
			pango_context_set_gravity_hint (context, PANGO_GRAVITY_HINT_STRONG);
		}
#endif
#endif
		kiba_title_object_update_surface (object);
	}
	cairo_destroy (cr);
	
	if (!priv->active_object)
		return;
	
	kiba_win_invalidate_area (title->win, priv->area);
	
	pango_layout_get_pixel_extents (priv->active_object->pango_layout, NULL, &log);
		
	if (d->orientation == GTK_ORIENTATION_VERTICAL) {
		priv->area.height = log.width + 10 + 2*title->style.border_width;
		priv->area.width = log.height + 2 + 2*title->style.border_width;
	}
	else {
		priv->area.width  = log.width + 10 + 2*title->style.border_width;
		priv->area.height = log.height + 2 + 2*title->style.border_width;
	}
	kiba_title_set_position (title);
	
	kiba_win_invalidate_area (title->win, priv->area);
}

/**
Class Functions
**/

static void kiba_title_class_init (KibaTitleClass *class) {
	g_type_class_add_private (class, sizeof (KibaTitlePrivate));
}

static void kiba_title_init (KibaTitle *title) {
	KibaTitlePrivate *priv;
	
	priv = KIBA_TITLE_GET_PRIVATE(title);
	
	SET_FLAG(title->flags, KIBA_TITLE_DRAWABLE);
	
	KibaRGBA *gradient = g_new0(KibaRGBA, 1);
	gradient[0] = (KibaRGBA) {1.0,1.0,1.0,1.0};
	title->style = (KibaTitleStyle) {
		KIBA_GRADIENT, KIBA_ROUNDED,
		2.0f, 45.0f, 5.0f,
		{ 1.0f, 1.0f, 1.0f, 0.8f },
		{ 0.8f, 0.8f, 0.8f, 0.8f }, { 0.0f, 0.0f, 0.0f, 0.8f },
		gradient, 1
	};
	title->font = "Arial regular 12";
	title->hide_delay = 50;
	title->state = TITLE_HIDDEN;
	
	priv->title_objects = NULL;
	priv->active_object = NULL;
	priv->area = (GdkRectangle) { 0, 0, 0, 0 };
	priv->fade_alpha = 0.0f;
	priv->scale_id = 0;
}

/**
 * kiba_title_new:
 *
 * Creates and returns a new KibaTitle.
 *
 * Returns: The KibaTitle object created.
 */
KibaTitle* kiba_title_new (GtkWidget *win) {
	KibaTitle *title = g_object_new (KIBA_TYPE_TITLE, NULL);
	
	g_return_val_if_fail (KIBA_IS_WIN(win), NULL);
	
	title->win = win;
	
	g_signal_connect (G_OBJECT(KIBA_WIN_CHILD(win)), "position-changed",
		G_CALLBACK(kiba_title_orientation_cb), title);
	
	return title;
}

/**
 * kiba_title_destroy:
 *
 * Destroys the KibaTitle.
 */
void kiba_title_destroy (KibaTitle *title) {
	KibaTitlePrivate *priv;
	
	g_return_if_fail (KIBA_IS_TITLE(title));
	
	priv = KIBA_TITLE_GET_PRIVATE(title);
	
	g_list_free (priv->title_objects);
	priv->title_objects = NULL;
	/*g_object_unref (title);
	title = NULL;*/
}
